/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.squarespace.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************

   Fotoxx image edit - functions for navigation of image files

**************************************************************************/


namespace navi 
{
   #define thumbnail_cachesize 10000                                       //  max thumbnails cached in memory
                                                                           //  supported image file types
   #define imagetypes ".jpeg .jpg .png .tif .tiff .bmp .gif .svg .xpm .tga .ico" 

   #define TEXTWIN GTK_TEXT_WINDOW_TEXT                                    //  GDK window of GTK text view
   #define SCROLLWIN GTK_SCROLLED_WINDOW
   #define NEVER GTK_POLICY_NEVER
   #define ALWAYS GTK_POLICY_ALWAYS

   GtkWidget      *wing = 0, *vboxx, *scrwing, *drwing;                    //  image gallery windows
   GtkWidget      *pwing = 0;                                              //  parent window
   GtkAdjustment  *scrollbar;

   #define     thumbxx 6                                                   //  thumbx array size
   int         thumbx[6] = { 512, 360, 256, 180, 128, 90 };                //  thumbnail image step sizes

   char        *galleryname = 0;                                           //  image directory or file list name
   int         gallerytype = 0;                                            //  1/2/3 = directory/file-list/metadata
   int         nfiles = 0;                                                 //  total file count (incl. subdirks)
   int         nimages = 0;                                                //  image file count
   char        **flist = 0;                                                //  image file list
   char        **mdlist = 0;                                               //  corresp. metadata

   typedef void ufunc(int Nth, int button);                                //  callback function for clicked image
   ufunc       *userfunc;

   int         xwinW = 1000, xwinH = 700;                                  //  gallery window initial size
   int         xwinX, xwinY;                                               //  gallery window initial position
   int         thumbsize = thumbfilesize;                                  //  thumbnail image <= thumbnail file
   int         thumbW, thumbH;                                             //  gallery window thumbnail cell size
   int         xrows, xcols;                                               //  gallery window thumbnail rows, cols
   int         margin = 5;                                                 //  cell margin from left and top edge
   int         genthumbs = 0;                                              //  count newly generated thumbnails
   int         maxscroll;                                                  //  max. scroll position
   int         fpresent = 0;                                               //  force gallery window to z-top
   int         targfileposn = 0;                                           //  scroll-to file position (Nth)

   //  private functions
   int    gallery_paint(GtkWidget *, cairo_t *);                           //  gallery window paint function
   int    gallery_paint_metadata(GtkWidget *drwing, cairo_t *cr);          //  same, for metadata report
   void   draw_text(cairo_t *cr, char *text, int x, int y, int ww);        //  draw text in gallery window
   void   gallery_destroy();                                               //  gallery window destroy event function
   void   menufuncx(GtkWidget *win, cchar *menu);                          //  function for gallery window buttons
   void   mouse_xevent(GtkWidget *, GdkEventButton *, void *);             //  gallery window mouse event function
	int    KBxpress(GtkWidget *, GdkEventKey *, void *);                    //  gallery window key press event func.
	int    KBxrelease(GtkWidget *, GdkEventKey *, void *);                  //  gallery window key release event
   char * image_navigate(cchar *filez, cchar *action, int Nth = 0);        //  image file list setup and navigate
   int    image_fcomp(cchar *file1, cchar *file2);                         //  file name compare (special sort)
}


/**************************************************************************

   public function to create/update image gallery (thumbnail window)       //  overhauled

   Make window scrolling window of thumbnails for a list of files
   Handle window buttons (up row, down page, open file, etc.)
   Call ufunc() when thumbnail image is clicked

   char * image_gallery(cchar *filez, cchar *action, int Nth, 
                         void ufunc(int Nth, int button), GtkWidget *parent)

   filez: image file or directory of image files or file with list of image files

   action:  init:    filez = initial file or directory 
            initF:   filez = file with list of image files to use
            sort:    sort the file list, directories first, ignore case
            insert:  insert filez into file list at position Nth (0 to last+1)
            delete:  delete Nth file in list
            find:    return Nth file (0 base) or null if Nth > last
            paint1:  create or refresh gallery window, anchor = Nth file
            paint2:  refresh gallery window if present, anchor = Nth file
            close    close gallery window
   
   Nth: file to return (action = find) or file to scroll-to (action = paint)

   void ufunc(int Nth, int button):
      - returns Nth of clicked thumbnail (0 to last)
      - returns -1 if gallery window is closed
      - returns -2 if key F1 is pressed (for context help)
      - button is mouse button used on clicked thumbnail
      
   parent: optional parent window
           if present, gallery window will overlay the parent window
           (window placement can be undone by the user)

   Returned values:
      Nth: filespec,    others: null
      The returned file belongs to caller and is subject for zfree().

***************************************************************************/

char * image_gallery(cchar *filez, cchar *action, int Nth, 
                     void ufunc(int Nth, int button), GtkWidget *parent)
{
   using namespace navi;

   GtkWidget      *tbarx;

   zthreadcrash();                                                         //  thread usage not allowed

   if (ufunc) userfunc = ufunc;                                            //  save callback function
   if (parent) pwing = parent;                                             //  save parent window

   if (strstr("init initF sort insert delete find",action))
      return image_navigate(filez,action,Nth);                             //  create or navigate image file list
      
   if (strEqu(action,"close")) {
      if (wing) gtk_widget_destroy(wing);
      return 0;
   }
   
   if (! strnEqu(action,"paint",5))                                        //  must be paint1 or paint2
      zappcrash("image_gallery action: %s",action);

   if (strEqu(action,"paint2") && ! wing) return 0;                        //  refresh but window not active
   if (strEqu(action,"paint1")) fpresent++;                                //  bring new window to z-top
   
   if (Nth >= 0) targfileposn = Nth;                                       //  scroll-to file position
   else if (filez) targfileposn = image_gallery_position(filez,0);         //  or use filez if present
   if (targfileposn > nfiles-1) targfileposn = nfiles-1;
   
   if (wing) {                                                             //  repaint existing gallery window
      gtk_widget_queue_draw(drwing);
      return 0;
   }

   wing = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create new gallery window
   
   vboxx = gtk_vbox_new(0,0);                                              //  vertical packing box
   gtk_container_add(GTK_CONTAINER(wing),vboxx);                           //  add to main window

   tbarx = create_toolbar(vboxx,24);                                       //  add toolbar and buttons 
   gtk_toolbar_set_style(GTK_TOOLBAR(tbarx),GTK_TOOLBAR_BOTH);

   add_toolbar_button(tbarx, ZTX("bigger"), ZTX("increase thumbnail size"), "gtk-zoom-in", menufuncx);
   add_toolbar_button(tbarx, ZTX("smaller"), ZTX("reduce thumbnail size"), "gtk-zoom-out", menufuncx);
   add_toolbar_button(tbarx, ZTX("parent"), ZTX("parent directory"), "folder.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("first page"), ZTX("jump to first file"), "first-page.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("prev page"), ZTX("previous page"), "prev-page.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("prev row"), ZTX("previous row"), "prev-row.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("next row"), ZTX("next row"), "next-row.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("next page"), ZTX("ttip::next page"), "next-page.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("last page"), ZTX("jump to last file"), "last-page.png", menufuncx);
   add_toolbar_button(tbarx, ZTX("close"), ZTX("close image gallery"), "gtk-close", menufuncx);

   gtk_toolbar_set_style(GTK_TOOLBAR(tbarx),GTK_TOOLBAR_BOTH);             //  set toolbar style         v.12.01
   if (strEqu(tbar_style,"icons"))
      gtk_toolbar_set_style(GTK_TOOLBAR(tbarx),GTK_TOOLBAR_ICONS);
   if (strEqu(tbar_style,"text"))
      gtk_toolbar_set_style(GTK_TOOLBAR(tbarx),GTK_TOOLBAR_TEXT);

   scrwing = gtk_scrolled_window_new(0,0);                                          //  create scrolled window
   gtk_scrolled_window_set_policy(SCROLLWIN(scrwing),ALWAYS,ALWAYS);                //  vertical scroll bar
   scrollbar = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrwing));
   gtk_container_add(GTK_CONTAINER(vboxx),scrwing);                                 //  add to main window
   drwing = gtk_drawing_area_new();                                                 //  create drawing window
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrwing),drwing);      //  add to scrolled window

   if (pwing) {
      gtk_window_get_size(GTK_WINDOW(pwing),&xwinW,&xwinH);                //  overlay parent window 
      gtk_window_get_position(GTK_WINDOW(pwing),&xwinX,&xwinY);
      gtk_window_set_default_size(GTK_WINDOW(wing),xwinW,xwinH);
   }

   gtk_widget_show_all(wing);                                              //  show all widgets
   gtk_window_move(GTK_WINDOW(wing),xwinX,xwinY);                          //  moved here             v.12.02

   G_SIGNAL(wing,"destroy",gallery_destroy,0);                             //  connect window events
   G_SIGNAL(drwing,"draw",gallery_paint,0);
   gtk_widget_add_events(drwing,GDK_BUTTON_PRESS_MASK);                    //  connect mouse events
   G_SIGNAL(drwing,"button-press-event",mouse_xevent,0);
   G_SIGNAL(wing,"key-press-event",KBxpress,0);                           	//  connect KB events
   G_SIGNAL(wing,"key-release-event",KBxrelease,0);

   return 0;
}


//  private function
//  paint gallery window - draw all thumbnail images that can fit

int navi::gallery_paint(GtkWidget *drwing, cairo_t *cr)
{
   using namespace navi;

   GdkPixbuf   *pxbT;
   double      x1, y1, x2, y2;
   int         ii, nrows, row, col;
   int         row1, row2;
   int         ww, hh, scrollp;
   int         drwingW, drwingH;
   int         thumx, thumy;
   char        *pp, *fname, p0;
   char        wintitle[200];
   
   snprintf(wintitle,199,"%s  %d files",galleryname,nfiles);               //  window title: gallery name and file count
   gtk_window_set_title(GTK_WINDOW(wing),wintitle);

   if (fpresent) gtk_window_present(GTK_WINDOW(wing));                     //  bring window to top
   fpresent = 0;

   if (gallerytype == 3) {                                                 //  metadata report
      gallery_paint_metadata(drwing,cr);
      return 1;
   }
   
   thumbW = thumbsize + 10;                                                //  thumbnail cell size + file name text
   thumbH = thumbsize + 30;

   if (! thumbsize) {
      thumbW = 400;                                                        //  zero, list view
      thumbH = 20;
   }

   xwinW = gtk_widget_get_allocated_width(scrwing);                        //  drawing window size
   xwinH = gtk_widget_get_allocated_height(scrwing);

   xrows = int(0.1 + 1.0 * xwinH / thumbH);                                //  get thumbnail rows and cols that
   xcols = int(0.1 + 1.0 * xwinW / thumbW);                                //    (almost) fit in window 
   if (xrows < 1) xrows = 1;
   if (xcols < 1) xcols = 1;
   nrows = (nfiles+xcols-1) / xcols;                                       //  thumbnail rows, 1 or more
   if (nrows < 1) nrows = 1;

   drwingW = xcols * thumbW + margin + 10;                                 //  layout size for entire gallery
   drwingH = (nrows + 1) * thumbH;                                         //  last row
   if (drwingH < xwinH) drwingH = xwinH;                                   //  at least current size
   
   gtk_widget_get_size_request(drwing,&ww,&hh);                            //  current size
   if (ww != drwingW || hh != drwingH)
      gtk_widget_set_size_request(drwing,drwingW,drwingH);                 //  needs to change

   maxscroll = (nrows - xrows) * thumbH;                                   //  max. layout scroll position
   if (maxscroll < 0) maxscroll = 0;

   gtk_adjustment_set_step_increment(scrollbar,thumbH);                    //  scrollbar works in row steps
   gtk_adjustment_set_page_increment(scrollbar,thumbH * xrows);            //  and in page steps

   if (targfileposn >= 0) {                                                //  target file position (Nth)
      scrollp = targfileposn / xcols * thumbH;                             //  set scroll position to target file
      if (scrollp > maxscroll) scrollp = maxscroll;
      gtk_adjustment_set_value(scrollbar,scrollp);                         //  will cause re-entrance
      if (gtk_adjustment_get_value(scrollbar) != scrollp) return 1;        //  scrollp OK in next call
      targfileposn = -1;
   }
   
   cairo_clip_extents(cr,&x1,&y1,&x2,&y2);                                 //  window region to paint
   row1 = y1 / thumbH;                                                     //  (< full screen if scrolling)
   row2 = y2 / thumbH;
   
   scrollp = gtk_adjustment_get_value(scrollbar);                          //  ignore redundant gtk3 draw events
   if (row1 < scrollp / thumbH) return 1;
   if (row2 > (scrollp + xwinH) / thumbH) return 1;

   for (row = row1; row <= row2; row++)                                    //  draw file thumbnails
   {
      for (col = 0; col < xcols; col++)                                    //  draw all columns in row
      {
         ii = row * xcols + col;                                           //  next file
         if (ii >= nfiles) return 1;                                       //  exit 2 nested loops

         p0 = *flist[ii];                                                  //  replace possible ! with /
         *flist[ii] = '/';

         fname = flist[ii];                                                //  filespec
         pp = strrchr(fname,'/');                                          //  get file name only
         if (pp) fname = pp + 1;

         thumx = col * thumbW + margin;                                    //  upper left corner in drawing area
         thumy = row * thumbH + margin;

         draw_text(cr,fname,thumx,thumy,thumbW-5);                         //  draw file name

         if (thumbsize) {                                                  //  zero >> list view
            pxbT = image_thumbnail(flist[ii],thumbsize);                   //  get thumbnail
            if (pxbT) {
               thumy = thumy + 16;                                         //  thumbnail 16 pixels down from text
               gdk_cairo_set_source_pixbuf(cr,pxbT,thumx,thumy);
               cairo_paint(cr);
               g_object_unref(pxbT);
            }
         }

         *flist[ii] = p0;                                                  //  restore !
      }
   }

   return 1;
}


//  private function
//  paint metadata report - draw thumbnail images + metadata

int navi::gallery_paint_metadata(GtkWidget *drwing, cairo_t *cr)
{
   using namespace navi;

   GdkPixbuf   *pxbT;
   double      x1, y1, x2, y2;
   int         ii, nrows, row, col;
   int         row1, row2;
   int         ww, hh, scrollp;
   int         drwingW, drwingH;
   int         thumx, thumy, textww;
   char        p0;
   
   if (thumbsize < 64) thumbsize = 64;
   thumbW = thumbsize + 10;                                                //  thumbnail layout size
   thumbH = thumbsize + 20;

   xwinW = gtk_widget_get_allocated_width(scrwing);                        //  drawing window size
   xwinH = gtk_widget_get_allocated_height(scrwing);

   xrows = int(0.1 + 1.0 * xwinH / thumbH);                                //  get thumbnail rows fitting in window
   if (xrows < 1) xrows = 1;
   xcols = 1;                                                              //  force cols = 1
   nrows = nfiles;                                                         //  thumbnail rows

   drwingW = xwinW;                                                        //  layout size for entire file list
   if (drwingW < 800) drwingW = 800;
   drwingH = (nrows + 1) * thumbH;                                         //  last row
   if (drwingH < xwinH) drwingH = xwinH;

   gtk_widget_get_size_request(drwing,&ww,&hh);                            //  current size
   if (ww != drwingW || hh != drwingH)
      gtk_widget_set_size_request(drwing,drwingW,drwingH);                 //  needs to change

   maxscroll = (nrows - xrows) * thumbH;                                   //  max. layout scroll position
   if (maxscroll < 0) maxscroll = 0;

   gtk_adjustment_set_step_increment(scrollbar,thumbH);                    //  scrollbar works in row steps
   gtk_adjustment_set_page_increment(scrollbar,thumbH * xrows);            //  and in page steps

   if (targfileposn >= 0) {                                                //  target file position (Nth)
      scrollp = targfileposn / xcols * thumbH;                             //  set scroll position to target file
      if (scrollp > maxscroll) scrollp = maxscroll;
      gtk_adjustment_set_value(scrollbar,scrollp);                         //  will cause re-entrance
      if (gtk_adjustment_get_value(scrollbar) != scrollp) return 1;        //  scrollp OK in next call
      targfileposn = -1;
   }
   
   cairo_clip_extents(cr,&x1,&y1,&x2,&y2);                                 //  window region to paint
   row1 = y1 / thumbH;                                                     //  (< full screen if scrolling)
   row2 = y2 / thumbH;
   
   scrollp = gtk_adjustment_get_value(scrollbar);                          //  ignore redundant gtk3 draw events
   if (row1 < scrollp / thumbH) return 1;
   if (row2 > (scrollp + xwinH) / thumbH) return 1;

   textww = drwingW - thumbW - 2 * margin;                                 //  space for text right of thumbnail

   for (row = row1; row <= row2; row++)                                    //  draw file thumbnails
   {
      for (col = 0; col < xcols; col++)
      {
         ii = row * xcols + col;                                           //  next file
         if (ii >= nfiles) return 1;                                       //  exit 2 nested loops

         p0 = *flist[ii];                                                  //  replace possible ! with /
         *flist[ii] = '/';

         thumx = col * thumbW + margin;                                    //  upper left corner in window space
         thumy = row * thumbH + margin;
         
         pxbT = image_thumbnail(flist[ii],thumbsize);                      //  get thumbnail
         if (pxbT) {
            gdk_cairo_set_source_pixbuf(cr,pxbT,thumx,thumy);
            cairo_paint(cr);
            g_object_unref(pxbT);
         }

         draw_text(cr,flist[ii],thumbW+margin,thumy,textww);               //  write filespec to right of thumbnail
         
         if (mdlist && mdlist[ii])                                         //  write metadata if present
            draw_text(cr,mdlist[ii],thumbW+margin,thumy+20,textww);
         
         *flist[ii] = p0;                                                  //  restore !
      }
   }

   return 1;
}


//  private function
//  write text for thumbnail limited by width of thumbnail

void navi::draw_text(cairo_t *cr, char *text, int px, int py, int ww)
{
   using namespace navi;

   static PangoFontDescription   *pfont = 0;
   static PangoLayout            *playout = 0;
   
   int            nn;
   static int     thumbsize2 = 0;
   static char    thumbfont[12] = "";
   
   if (thumbsize != thumbsize2) {                                          //  scale file name font to thumbnail size
      thumbsize2 = thumbsize;
      nn = 7 + thumbsize / 128;                                            //  font size from 7 to 10
      if (thumbsize == 0) nn = 9;                                          //  list view, use 9
      sprintf(thumbfont,"sans %d",nn);
      pfont = pango_font_description_from_string(thumbfont);
      playout = pango_cairo_create_layout(cr);                             //  gtk3
      pango_layout_set_font_description(playout,pfont);
   }

   pango_layout_set_width(playout,ww*PANGO_SCALE);                         //  limit width to avail. space  v.12.01
   pango_layout_set_ellipsize(playout,PANGO_ELLIPSIZE_END);
   pango_layout_set_text(playout,text,-1);

   cairo_move_to(cr,px,py);                                                //  gtk3
   cairo_set_source_rgb(cr,0,0,0);
   pango_cairo_show_layout(cr,playout);
   return;
}


//  private function
//  gallery window destroy event - mark window not active.

void navi::gallery_destroy()
{
   using namespace navi;
   
   wing = 0;                                                               //  no window
   if (userfunc) userfunc(-1,0);                                           //  tell caller
   return;
}


//  private function - menu function for gallery window
//    - scroll window as requested
//    - jump to new file or folder as requested

void navi::menufuncx(GtkWidget *win, cchar *menu)
{
   using namespace navi;

   int         ii, yn, Gwarn, scrollp;
   char        *filez, *pp;
      
   if (strEqu(menu,ZTX("close"))) {                                        //  close image gallery window
      gtk_widget_destroy(wing);                                            //  destroy event function calls userfunc(-1)
      return;
   }

   if (strEqu(menu,ZTX("parent"))) {
      Gwarn = 0;
      if (navi::gallerytype > 1) Gwarn = 1;                                //  warn if discarding search or
      pp = navi::galleryname;                                              //    collection type gallery
      if (pp && strEqu(pp,"Recent Files")) Gwarn = 0;                      //  v.12.01
      if (Gwarn) {                                                         //  if gallery not from a directory,
         yn = zmessageYN(wing,Bdiscard,navi::galleryname);                 //    warn user, gallery will be discarded
         if (! yn) return;                                                 //  do not discard
      }
      if (galleryname && gallerytype == 1) 
         filez = strdupz(galleryname,0,"navi");
      else filez = strdupz(curr_dirk,0,"navi");                            //  v.12.01
      pp = strrchr(filez,'/');                                             //  go up to '/' and stop
      if (pp && pp > filez) *pp = 0;
      else filez = strdupz("/",0,"navi");                                  //  v.12.01
      image_navigate(filez,"init");                                        //  get new file list
      zfree(filez);
      targfileposn = 0;
      gtk_widget_queue_draw(drwing);
      return;
   }

   scrollp = gtk_adjustment_get_value(scrollbar);                          //  current scroll position

   if (strEqu(menu,ZTX("bigger")))  {                                      //  next bigger thumbnail size
      for (ii = 0; ii < thumbxx; ii++) 
         if (thumbsize == thumbx[ii]) break;
      if (ii == 0) return;
      thumbsize = thumbx[ii-1];
      targfileposn = scrollp / thumbH * xcols;                             //  keep top row position
      gtk_widget_queue_draw(drwing);
      return;
   }

   if (strEqu(menu,ZTX("smaller")))  {                                     //  next smaller
      for (ii = 0; ii < thumbxx; ii++) 
         if (thumbsize == thumbx[ii]) break;
      if (ii >= thumbxx-1) thumbsize = 0;                                  //  no thumbs, list view
      else  thumbsize = thumbx[ii+1];
      if (thumbsize < 128 && gallerytype == 3)                             //  min. for metadata report   v.12.01
         thumbsize = 128;
      targfileposn = scrollp / thumbH * xcols;                             //  keep top row position
      gtk_widget_queue_draw(drwing);
      return;
   }

   if (strEqu(menu,ZTX("prev row"))) scrollp -= thumbH;
   if (strEqu(menu,ZTX("next row"))) scrollp += thumbH;
   if (strEqu(menu,ZTX("prev page"))) scrollp -= thumbH * xrows;
   if (strEqu(menu,ZTX("next page"))) scrollp += thumbH * xrows;
   if (strEqu(menu,ZTX("first page"))) scrollp = 0;
   if (strEqu(menu,ZTX("last page"))) scrollp = maxscroll;

   if (scrollp < 0) scrollp = 0;                                           //  enforce limits
   if (scrollp > maxscroll) scrollp = maxscroll;
   targfileposn = scrollp / thumbH * xcols;
   gtk_widget_queue_draw(drwing);
   return;
}


//  private function
//  mouse event function for gallery window - get selected thumbnail and file
//  user function receives clicked file, which is subject for zfree()

void navi::mouse_xevent(GtkWidget *, GdkEventButton *event, void *)
{
   using namespace navi;

   int            mousex, mousey, mousebutt;
   int            row, col, nrows, ii, err;
   char           *filez;
   struct stat    statb;
   
   if (! nfiles) return;                                                   //  empty window
   
   mousex = int(event->x);
   mousey = int(event->y);
   mousebutt = event->button;

   row = (mousey - margin) / thumbH;                                       //  find selected row, col
   col = (mousex - margin) / thumbW;

   nrows = 1 + (nfiles-1) / xcols;                                         //  total thumbnail rows, 1 or more
   if (col < 0 || col >= xcols) return;
   if (row < 0 || row >= nrows) return;

   ii = xcols * row + col;
   if (ii >= nfiles) return;

   filez = strdupz(flist[ii],0,"navi");                                    //  selected file
   *filez = '/';
   
   err = stat(filez,&statb);
   if (err) {                                                              //  file is gone?
      zfree(filez);
      return;
   }

   if (S_ISDIR(statb.st_mode)) {                                           //  if directory, go there
      image_navigate(filez,"init");
      gtk_widget_queue_draw(drwing);
      zfree(filez);
      return;
   }
   
   if (userfunc) userfunc(ii,mousebutt);                                   //  clicked file position to user
   return;
}


//  private function
//  KB event function - respond to keyboard navigation keys
//  key definitions: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int navi::KBxpress(GtkWidget *win, GdkEventKey *event, void *)             //  prevent propagation of key-press
{                                                                          //    events to toolbar buttons 
   return 1;
}

int navi::KBxrelease(GtkWidget *win, GdkEventKey *event, void *)
{
   using namespace navi;

   int      KBkey;
   
   KBkey = event->keyval;

   if (KBkey == GDK_KEY_plus) menufuncx(win,ZTX("bigger"));                //  +/- = bigger/smaller thumbnails
   if (KBkey == GDK_KEY_equal) menufuncx(win,ZTX("bigger"));
   if (KBkey == GDK_KEY_minus) menufuncx(win,ZTX("smaller"));
   if (KBkey == GDK_KEY_KP_Add) menufuncx(win,ZTX("bigger"));              //  keypad +/- also 
   if (KBkey == GDK_KEY_KP_Subtract) menufuncx(win,ZTX("smaller"));

   if (KBkey == GDK_KEY_Left) menufuncx(win,ZTX("prev page"));             //  left arrow = previous page
   if (KBkey == GDK_KEY_Right) menufuncx(win,ZTX("next page"));            //  right arrow = next page
   if (KBkey == GDK_KEY_Up) menufuncx(win,ZTX("prev row"));                //  up arrow = previous row
   if (KBkey == GDK_KEY_Down) menufuncx(win,ZTX("next row"));              //  down arrow = next row
   
   if (KBkey == GDK_KEY_Home) menufuncx(win,ZTX("first page"));            //  keys added 
   if (KBkey == GDK_KEY_End) menufuncx(win,ZTX("last page"));
   if (KBkey == GDK_KEY_Page_Up) menufuncx(win,ZTX("prev page"));
   if (KBkey == GDK_KEY_Page_Down) menufuncx(win,ZTX("next page"));

   if (KBkey == GDK_KEY_Escape) gtk_widget_destroy(win);                   //  Escape = cancel gallery window
   
   if (KBkey == GDK_KEY_F1)
      showz_userguide(zfuncs::F1_help_topic);
   
   return 1;
}


//  Public function - get file position in file list.
//  If Nth position matches file, this is returned.
//  Otherwise the file is searched from position 0.
//  Position 0-last is returned if found, or -1 if not.

int image_gallery_position(cchar *file, int Nth)
{
   using namespace navi;
   
   int      ii;
   
   if (! nfiles) return -1;

   if (Nth >= 0 && Nth < nfiles) ii = Nth;
   else ii = 0;
   
   if (strEqu(file+1,flist[ii]+1)) return ii;                              //  1st flist[ii] char. may be !

   for (ii = 0; ii < nfiles; ii++)
      if (strEqu(file+1,flist[ii]+1)) break;

   if (ii < nfiles) return ii;
   return -1;
}


//  public function
//  determine if a file is a directory or a supported image file type
//  return: 0 = error, 1 = directory, 2 = image file, 3 = other

int image_file_type(cchar *file)
{
   using namespace navi;

   int            err, cc;
   cchar          *pp;
   struct stat    statbuf;

   if (! file) return 0;
   err = stat(file,&statbuf);
   if (err) return 0;

   if (S_ISDIR(statbuf.st_mode)) {                                         //  directory
      cc = strlen(file);
      if (cc > 12) {
         pp = file + cc - 12;
         if (strEqu(pp,"/.thumbnails")) return 3;                          //  .thumbnails
      }
      return 1;
   }

   if (S_ISREG(statbuf.st_mode)) {                                         //  reg. file
      pp = strrchr(file,'.');
      if (! pp) return 3;
      pp = strcasestr(imagetypes,pp);                                      //  supported image type
      if (pp) return 2;
   }
   
   return 3;
}


//  Get thumbnail file for the given image file.
//  If missing or stale, add or update in /.thumbnails directory.
//  Returned filespec is subject for zfree().

char * image_thumbfile(char *imagefile)
{
   GdkPixbuf         *thumbpxb;
   GError            *gerror = 0;
   char              *pfile, *bfile, *thumbfile;
   int               err, ww, hh;
   struct stat       statf, statb;
   PXM               *pxm1, *pxm2;
   float             scale;
   
   err = stat(imagefile,&statf);
   if (err) return 0;
   if (! S_ISREG(statf.st_mode)) return 0;                                 //  not a regular file
   if (image_file_type(imagefile) != 2) return 0;                          //  unsupported image type
   
   pfile = strrchr(imagefile,'/');                                         //  get .../filename.xxx
   if (! pfile) return 0;
   if (pfile > imagefile+12 && strnEqu(pfile-12,"/.thumbnails/",13)) {     //  .../.thumbnails/filename.xxx ??
      thumbfile = strdupz(imagefile,0,"image_thumbfile");                  //  yes, a thumbnail file
      return thumbfile;                                                    //  return same file
   }
   
   thumbfile = strdupz(imagefile,20,"image_thumbfile");                    //  construct thumbnail file
   bfile = thumbfile + (pfile - imagefile);                                //    .../.thumbnails/filename.xxx.png
   strcpy(bfile,"/.thumbnails");
   bfile += 12;
   strcpy(bfile,pfile);
   strcat(bfile,".png");

   err = stat(thumbfile,&statb);                                           //  thumbfile exists, up to date ??
   if (! err && statb.st_mtime >= statf.st_mtime) 
      return thumbfile;                                                    //  yes

   *bfile = 0;                                                             //  thumbfile = .../.thumbnails
   err = stat(thumbfile,&statb);
   if (err) err = mkdir(thumbfile,0751);                                   //  create .thumbnails directory
   if (err) {
      printf("create thumbnail error: %s \n",strerror(errno));
      return 0;
   }

   *bfile = *pfile;                                                        //  .../.thumbnails/filename.xxx.png

   ww = hh = thumbfilesize;                                                //  create thumbnail pixbuf
   thumbpxb = gdk_pixbuf_new_from_file_at_size(imagefile,ww,hh,&gerror);
   if (thumbpxb) {
      gdk_pixbuf_save(thumbpxb,thumbfile,"png",&gerror,null);              //  save to /.thumbnails/ directory
      g_object_unref(thumbpxb);
   }
   if (! gerror) {
      navi::genthumbs++;                                                   //  count generated thumbnails
      return thumbfile;
   }

   pxm1 = f_load(imagefile);                                               //  backup thumbnail creator     v.12.03
   if (! pxm1) return 0;
   ww = pxm1->ww;
   hh = pxm1->hh;
   if (ww > hh) scale = 1.0 * thumbfilesize / ww;
   else scale = 1.0 * thumbfilesize / hh;
   if (scale < 1) {
      ww = scale * ww;
      hh = scale * hh;
      pxm2 = PXM_rescale(pxm1,ww,hh);
      PXM_free(pxm1);
      pxm1 = pxm2;
   }
   PXBwrite(pxm1,thumbfile);
   PXM_free(pxm1);
   navi::genthumbs++;
   return thumbfile;
}


//  Public function
//  Get thumbnail image for given image file, from .thumbnails directory.
//  Add thumbnail file if missing, or update it if older than image file.
//  Returned thumbnail belongs to caller: g_object_unref() is necessary.

GdkPixbuf * image_thumbnail(char *fpath, int size)
{
   GdkPixbuf         *thumbpxb;
   GError            *gerror = 0;
   int               ii, err;
   char              *bpath;
   time_t            mtime;
   struct stat       statf;
   const int         cachesize = thumbnail_cachesize;                      //  shorthand

   static int           nextcache, ftf = 1;
   static int           sizecache[cachesize];
   static time_t        mtimecache[cachesize];
   static char          *fpathcache[cachesize];
   static GdkPixbuf     *pixbufcache[cachesize];
   
   zthreadcrash();                                                         //  thread usage not allowed

   if (ftf) {                                                              //  first call
      for (ii = 0; ii < cachesize; ii++) 
         fpathcache[ii] = 0;                                               //  clear cache
      ftf = 0;
   }
   
   err = stat(fpath,&statf);                                               //  fpath status info
   if (err) return 0;

   if (S_ISDIR(statf.st_mode)) {                                           //  if directory, return folder image
      bpath = zmalloc(500);
      strncatv(bpath,499,zfuncs::zicondir,"/folder256.png",null);
      thumbpxb = gdk_pixbuf_new_from_file_at_size(bpath,size,size,&gerror);
      zfree(bpath);
      return thumbpxb;
   }
   
   mtime = statf.st_mtime;                                                 //  last modification time
   
   if (! size) size = thumbfilesize;                                       //  default thumb size
   
   for (ii = nextcache; ii >= 0; ii--)                                     //  get cached pixbuf if found 
      if (fpathcache[ii] && strEqu(fpath,fpathcache[ii]) &&
          sizecache[ii] == size && mtime == mtimecache[ii]) break;         //  check mtime (bugfix)
   if (ii >= 0) {
      thumbpxb = gdk_pixbuf_copy(pixbufcache[ii]);
      return thumbpxb;
   }
   for (ii = cachesize-1; ii > nextcache; ii--)                            //  continue search
      if (fpathcache[ii] && strEqu(fpath,fpathcache[ii]) &&
          sizecache[ii] == size && mtime == mtimecache[ii]) break;
   if (ii > nextcache) {
      thumbpxb = gdk_pixbuf_copy(pixbufcache[ii]);
      return thumbpxb;
   }
   
   if (size > thumbfilesize) {                                             //  make bigger thumbnail
      thumbpxb = gdk_pixbuf_new_from_file_at_size(fpath,size,size,&gerror);
      if (! thumbpxb) return 0;
      goto addcache;
   }
   
   bpath = image_thumbfile(fpath);                                         //  get thumbnail file
   if (! bpath) return 0;
   thumbpxb = gdk_pixbuf_new_from_file_at_size(bpath,size,size,&gerror);   //  get thumbnail
   zfree(bpath);
   if (! thumbpxb) return 0;

addcache:
   nextcache++;                                                            //  next cache slot (oldest)
   if (nextcache == cachesize) nextcache = 0;
   ii = nextcache;
   if (fpathcache[ii]) {                                                   //  free prior occupant
      zfree(fpathcache[ii]);
      g_object_unref(pixbufcache[ii]);
   }
   fpathcache[ii] = strdupz(fpath,0,"thumbnail_cache");                    //  add new occupant
   pixbufcache[ii] = gdk_pixbuf_copy(thumbpxb);                            //  this memory is not tracked
   sizecache[ii] = size;
   mtimecache[ii] = mtime;

   return thumbpxb;                                                        //  return pixbuf to caller
}


/**************************************************************************

   private function - manage list of image files within a directory

   get an image file in the same directory as given file or directory

   char * navi::image_navigate(cchar *filez, cchar *action, int Nth)

   action:  init:    file list = directories and image files in directory filez
            initF:   file list = filez = list of image files to use (cchar **)
            sort:    sort the file list, directories first, ignore case
            insert:  insert filez into file list at position Nth (0 to last+1)
            delete:  delete Nth file in list
            find:    returns Nth file (0 base) or null if Nth > last

   Nth: file to return for action = find
   
   Returned values:
      find: filespec, else null
      The returned file belongs to caller and is subject for zfree().

***************************************************************************/

char * navi::image_navigate(cchar *filez, cchar *action, int Nth)
{
   using namespace navi;

   char           *buff;
   cchar          *findcommand = "find \"%s\" -maxdepth 1";
   char           *pp, *file2;
   int            err, ii, cc, ftyp, contx = 0, fposn;
   FILE           *fid;
   struct stat    statbuf;
   
   if (! strstr("init initF sort insert delete find",action))
         zappcrash("image_navigate %s",action);

   if (strnEqu(action,"init",4))                                           //  init or initF
   {
      if (flist) {
         for (ii = 0; ii < nfiles; ii++) zfree(flist[ii]);                 //  free prior file list
         zfree(flist);
      }

      if (mdlist) {                                                        //  clear prior metadata if any
         for (ii = 0; ii < nfiles; ii++) 
            if (mdlist[ii]) zfree(mdlist[ii]);
         zfree(mdlist);
         mdlist = 0;
      }

      cc = MAXIMAGES * sizeof(char *);      
      flist = (char **) zmalloc(cc,"navi");                                //  list of file pointers

      nfiles = nimages = 0;                                                //  no files
      fposn = 0;
      gallerytype = 0;                                                     //  v.12.01
   }

   if (strEqu(action,"init"))                                              //  initialize from given directory
   {
      if (! filez) return 0;                                               //  should not happen

      if (galleryname) zfree(galleryname);
      galleryname = strdupz(filez,0,"navi");
      gallerytype = 1;                                                     //  gallery type = directory

      err = stat(galleryname,&statbuf);
      if (err) {
         pp = (char *) strrchr(galleryname,'/');                           //  bad file, check directory part  
         if (! pp) return 0;
         pp[1] = 0;
         err = stat(galleryname,&statbuf);
         if (err) return 0;                                                //  give up, empty file list
      }

      if (S_ISREG(statbuf.st_mode)) {                                      //  if a file, get directory part
         pp = (char *) strrchr(galleryname,'/');
         if (! pp) return 0;
         pp[1] = 0;
      }

      while ((buff = command_output(contx,findcommand,galleryname)))       //  find all files
      {
         if (strEqu(buff,galleryname)) {                                   //  skip self directory
            zfree(buff);
            continue;
         }
         
         if (nfiles == MAXIMAGES) {                                        //  message, not crash 
            zmessageACK(0,Btoomanyfiles,MAXIMAGES);
            break;
         }

         ftyp = image_file_type(buff);

         if (ftyp == 1) {                                                  //  subdirectory
            flist[nfiles] = buff;                                          //  add to file list
            flist[nfiles][0] = '!';                                        //  if directory, make it sort first
            nfiles++;
         }

         else if (ftyp == 2) {                                             //  supported image file type
            flist[nfiles] = buff;                                          //  add to file list
            nfiles++;
            nimages++; 
         }

         else {
            zfree(buff);                                                   //  (.thumbnails not ftyp 1)
            continue;
         }
      }

      if (nfiles > 1) 
         HeapSort(flist,nfiles,image_fcomp);                               //  Heap Sort - pointers to strings

      return 0;
   }

   if (strEqu(action,"initF"))                                             //  initialize from given list
   {
      if (galleryname) zfree(galleryname);
      galleryname = strdupz(filez,0,"navi");
      gallerytype = 2;                                                     //  gallery type = file list
      
      fid = fopen(galleryname,"r");                                        //  open file
      if (! fid) return 0;

      buff = zmalloc(maxfcc);
      
      while (true)                                                         //  read list of files
      {
         if (nfiles == MAXIMAGES) {
            zmessageACK(0,Btoomanyfiles,MAXIMAGES);
            break;
         }
         pp = fgets_trim(buff,maxfcc-1,fid,1);
         if (! pp) break;
         err = stat(pp,&statbuf);                                          //  check file exists
         if (err) continue;
         flist[nfiles] = strdupz(buff,0,"navi");                           //  add files to memory list
         nfiles++;
         nimages++;
      }

      fclose(fid);
      zfree(buff);                                                         //  sort removed
      return 0;
   }
   
   if (strEqu(action,"sort"))                                              //  sort the list from initF 
   {
      if (gallerytype == 3) return 0;                                      //  v.12.01
      if (nfiles < 2) return 0;
      HeapSort(flist,nfiles,image_fcomp);                                  //  Heap Sort - pointers to strings
      return 0;
   }

   if (strEqu(action,"insert"))                                            //  insert new file into list
   {
      if (gallerytype == 3) return 0;                                      //  metadata report     v.12.01
      fposn = Nth;                                                         //  file position from caller
      if (fposn < 0) fposn = 0;                                            //  limit to allowed range
      if (fposn > nfiles) fposn = nfiles;

      if (nfiles == MAXIMAGES-1) {                                         //  no room
         zmessageACK(0,Btoomanyfiles,MAXIMAGES);
         return 0;
      }

      for (ii = nfiles; ii > fposn; ii--)                                  //  create hole is list
         flist[ii] = flist[ii-1];

      flist[fposn] = strdupz(filez,0,"navi");                              //  put new file in hole
      nfiles++;
      nimages++;                                                           //  bugfix
   }

   if (strEqu(action,"delete"))                                            //  delete file from list
   {
      if (gallerytype == 3) return 0;                                      //  metadata report     v.12.01
      fposn = Nth;                                                         //  file position from caller must be OK
      if (fposn < 0 || fposn > nfiles-1) return 0;
      nfiles--;
      if (*flist[fposn] != '!') nimages--;                                 //  not a directory              v.12.01
      zfree(flist[fposn]);                                                 //  remove file from list
      for (ii = fposn; ii < nfiles; ii++) {                                //  close the hole
         if (nfiles < 0) printf("meaningless reference %d",ii);            //  stop g++ optimization bug    ////
         flist[ii] = flist[ii+1];
      }
   }
   
   if (strEqu(action,"find"))
   {
      fposn = Nth;                                                         //  file position from caller must be OK
      if (fposn < 0 || fposn > nfiles-1) return 0;
      file2 = strdupz(flist[fposn],0,"navi");                              //  get Nth file
      file2[0] = '/';                                                      //  restore initial '/'
      err = stat(file2,&statbuf);
      if (! err) return file2;
      zfree(file2);
   }
   
   return 0;
}


//  private function for special file name compare
//  directories sort first and upper/lower case is ignored

int navi::image_fcomp(cchar *file1, cchar *file2)
{
   int      nn;
   nn = strcasecmp(file1,file2);                                           //  compare ignoring case
   if (nn != 0) return nn;
   nn = strcmp(file1,file2);                                               //  if equal, do normal compare
   return nn;
}


/**************************************************************************/

//  Select files from the image gallery window, return list of files selected.
//  The dialog shows the list of files selected and can be edited.
//  The returned file list belongs to caller and is subject for zfree().
//  The file list EOL is marked with null.
//  The image_gallery() callback function is changed and must be restored by caller.

int  imgl_showthumb();
void imgl_insert_file(cchar *imagefile);
void imgl_gallery_file(int Nth, int button);

GtkWidget      *imgl_drawarea = 0;
GtkWidget      *imgl_files = 0;
cchar          *imgl_font = "Monospace 8";
zdialog        *imgl_zd = 0;
int            imgl_fontheight = 14;
int            imgl_cursorpos = 0;


char ** image_gallery_getfiles(char *startdir, GtkWidget *parent)
{
   using namespace navi;

   int  imgl_dialog_event(zdialog *zd, cchar *event);
   int  imgl_mouseclick(GtkWidget *, GdkEventButton *event, void *);

   PangoLanguage           *plang;
   PangoFontDescription    *pfontdesc;
   PangoContext            *pcontext;
   PangoFont               *pfont;
   PangoFontMetrics        *pmetrics;
   GdkCursor               *cursor;
   GdkWindow               *gdkwin;
   GtkTextBuffer           *textBuff;
   GtkTextIter             iter1, iter2;

   int               fontascent, fontdescent;
   int               line, nlines, ii;
   char              *imagefile, **filelist;
   
   zthreadcrash();                                                         //  thread usage not allowed

   image_gallery(startdir,"paint1",0,imgl_gallery_file,parent);            //  activate image gallery window

   imgl_zd = zdialog_new(ZTX("Select Files"),wing,ZTX("done"),ZTX("cancel"),null);
   zdialog_add_widget(imgl_zd,"hbox","hb1","dialog",0,"expand|space=5");
   zdialog_add_widget(imgl_zd,"frame","fr11","hb1",0,"expand");
   zdialog_add_widget(imgl_zd,"scrwin","scrwin","fr11",0,"expand");
   zdialog_add_widget(imgl_zd,"edit","files","scrwin");
   zdialog_add_widget(imgl_zd,"vbox","vb12","hb1");
   zdialog_add_widget(imgl_zd,"frame","fr12","vb12");
   zdialog_add_widget(imgl_zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(imgl_zd,"button","delete","hb2",ZTX("delete"),"space=8");
   zdialog_add_widget(imgl_zd,"button","insert","hb2",ZTX("insert"),"space=8");
   zdialog_add_widget(imgl_zd,"button","addall","hb2",ZTX("add all"),"space=30");

   GtkWidget *textbox = zdialog_widget(imgl_zd,"files");                   //  disable text wrap 
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(textbox),GTK_WRAP_NONE);

   GtkWidget *frame = zdialog_widget(imgl_zd,"fr12");                      //  drawing area for thumbnail image
   imgl_drawarea = gtk_drawing_area_new();
   gtk_widget_set_size_request(imgl_drawarea,256,258);                     //  increased
   gdkwin = gtk_widget_get_window(imgl_drawarea);                          //  gtk3
   gtk_container_add(GTK_CONTAINER(frame),imgl_drawarea);
   
   imgl_files = zdialog_widget(imgl_zd,"files");                           //  activate mouse-clicks for
   gtk_widget_add_events(imgl_files,GDK_BUTTON_PRESS_MASK);                //    file list widget
   G_SIGNAL(imgl_files,"button-press-event",imgl_mouseclick,0);

   pfontdesc = pango_font_description_from_string(imgl_font);              //  set default font for files window
   gtk_widget_modify_font(imgl_files,pfontdesc);

   plang = pango_language_get_default();                                   //  get font metrics (what a mess)
   pcontext = gtk_widget_get_pango_context(imgl_files);
   pfont = pango_context_load_font(pcontext,pfontdesc);
   pmetrics = pango_font_get_metrics(pfont,plang);
   fontascent = pango_font_metrics_get_ascent(pmetrics) / PANGO_SCALE;
   fontdescent = pango_font_metrics_get_descent(pmetrics) / PANGO_SCALE;
   imgl_fontheight = fontascent + fontdescent;                             //  effective line height

   zdialog_resize(imgl_zd,600,0);                                          //  start dialog
   zdialog_run(imgl_zd,imgl_dialog_event);

   cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                            //  arrow cursor for file list widget
   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(imgl_files),TEXTWIN);   //  (must be done after window realized)
   gdk_window_set_cursor(gdkwin,cursor);
   imgl_cursorpos = 0;

   zdialog_wait(imgl_zd);                                                  //  wait for completion

   if (imgl_zd->zstat != 1) {                                              //  cancelled
      zdialog_free(imgl_zd);                                               //  kill dialog
      image_gallery(0,"close");                                            //  close image gallery window
      return 0;
   }

   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   nlines = gtk_text_buffer_get_line_count(textBuff);
   
   filelist = (char **) zmalloc((nlines+1) * sizeof(char *),"imgl.getfiles");

   for (ii = line = 0; line < nlines; line++)                              //  get list of files from dialog
   {
      gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);              //  iter at line start
      iter2 = iter1;
      gtk_text_iter_forward_to_line_end(&iter2);
      imagefile = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);      //  get line of text
      if (imagefile && *imagefile == '/') {
         filelist[ii] = strdupz(imagefile,0,"imgl.getfiles");              //  >> next file in list
         free(imagefile);
         ii++;
      }
   }
   filelist[ii] = 0;                                                       //  mark EOL
   
   zdialog_free(imgl_zd);                                                  //  kill dialog
   image_gallery(0,"close");                                               //  close image gallery window

   if (! ii) {
      zfree(filelist);                                                     //  file list is empty
      return 0;
   }

   return filelist;                                                        //  return file list
}


//  imgl dialog event function

int imgl_dialog_event(zdialog *zd, cchar *event)
{
   GtkTextBuffer  *textBuff;
   GtkTextIter    iter1, iter2;
   char           *ftemp;
   static char    *imagefile = 0;
   int            line, posn;
   
   if (strEqu(event,"delete"))                                             //  delete file at cursor position
   {
      if (imagefile) zfree(imagefile);
      imagefile = 0;

      textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
      line = imgl_cursorpos;
      gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);              //  iter at line start
      iter2 = iter1;
      gtk_text_iter_forward_to_line_end(&iter2);                           //  iter at line end

      ftemp = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);          //  get selected file
      if (! ftemp || *ftemp != '/') {
         if (ftemp) free(ftemp);
         return 0;
      }

      imagefile = strdupz(ftemp,0,"imgl.getfiles");                        //  save deleted file for poss. insert
      free(ftemp);
      
      gtk_text_buffer_delete(textBuff,&iter1,&iter2);                      //  delete file text
      gtk_text_buffer_get_iter_at_line(textBuff,&iter2,line+1);
      gtk_text_buffer_delete(textBuff,&iter1,&iter2);                      //  delete empty line (\n)

      imgl_showthumb();                                                    //  thumbnail = next file
   }

   if (strEqu(event,"insert"))                                             //  insert last deleted file
   {                                                                       //    at current cursor position
      if (! imagefile) return 0;
      imgl_insert_file(imagefile);
      zfree(imagefile);
      imagefile = 0;
   }

   if (strEqu(event,"addall"))                                             //  insert all files in image gallery
   {
      textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
      posn = 0;

      while (true)
      {
         imagefile = image_gallery(0,"find",posn);                         //  get first or next file
         if (! imagefile) break;
         posn++;
         gtk_text_buffer_get_iter_at_line(textBuff,&iter1,imgl_cursorpos);
         gtk_text_buffer_insert(textBuff,&iter1,"\n",1);                   //  insert new blank line
         gtk_text_buffer_get_iter_at_line(textBuff,&iter1,imgl_cursorpos);
         gtk_text_buffer_insert(textBuff,&iter1,imagefile,-1);             //  insert image file
         zfree(imagefile);
         imgl_cursorpos++;                                                 //  advance cursor position
      }
   }

   return 0;
}


//  add image file to list at current cursor position, set thumbnail = file

void imgl_insert_file(cchar *imagefile)
{
   GtkTextIter    iter;
   GtkTextBuffer  *textBuff;
   
   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   gtk_text_buffer_get_iter_at_line(textBuff,&iter,imgl_cursorpos);
   gtk_text_buffer_insert(textBuff,&iter,"\n",1);                          //  insert new blank line
   gtk_text_buffer_get_iter_at_line(textBuff,&iter,imgl_cursorpos);
   gtk_text_buffer_insert(textBuff,&iter,imagefile,-1);                    //  insert image file

   imgl_showthumb();                                                       //  update thumbnail
   imgl_cursorpos++;                                                       //  advance cursor position

   return;
}


//  called from image gallery window when a thumbnail is clicked
//  add image file to list at current cursor position, set thumbnail = file

void imgl_gallery_file(int Nth, int button)
{
   int            ftyp;
   char           *imagefile;
   
   if (Nth == -2) {
      showz_userguide(zfuncs::F1_help_topic);                              //  F1 context help
      return;
   }

   imagefile = image_gallery(0,"find",Nth);                                //  get file at clicked position
   if (! imagefile) return;

   ftyp = image_file_type(imagefile);                                      //  ignore directories
   if (ftyp == 2) imgl_insert_file(imagefile);                             //  insert file at current position
   zfree(imagefile);

   return;
}


//  process mouse click in files window: 
//  set new cursor position and set thumbnail = clicked file

int imgl_mouseclick(GtkWidget *, GdkEventButton *event, void *)
{
   int            mpy;
   GtkWidget      *scrollwin;
   GtkAdjustment  *scrolladj;
   double         scrollpos;

   if (event->type != GDK_BUTTON_PRESS) return 0;
   mpy = int(event->y);
   scrollwin = zdialog_widget(imgl_zd,"scrwin");                           //  window scroll position
   scrolladj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrollwin));
   scrollpos = gtk_adjustment_get_value(scrolladj);
   imgl_cursorpos = (mpy + scrollpos) / imgl_fontheight;                   //  line selected
   imgl_showthumb();                                                       //  show thumbnail image
   return 0;
}


//  show thumbnail for file at current cursor position

int imgl_showthumb()
{
   int            line;
   char           *imagefile;
   GdkWindow      *gdkwin;
   cairo_t        *cr;
   GtkTextBuffer  *textBuff;
   GtkTextIter    iter1, iter2;
   GdkPixbuf      *thumbnail = 0;

   gdkwin = gtk_widget_get_window(imgl_drawarea);                          //  gtk3
   cr = gdk_cairo_create(gdkwin);
   
   line = imgl_cursorpos;
   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);                 //  iter at line start
   iter2 = iter1;
   gtk_text_iter_forward_to_line_end(&iter2);                              //  iter at line end

   imagefile = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);         //  get selected file
   if (*imagefile != '/') {
      free(imagefile);
      return 0;
   }

   thumbnail = image_thumbnail(imagefile,256);                             //  get thumbnail
   free(imagefile);

   if (thumbnail) {
      cairo_set_source_rgb(cr,1,1,1);                                      //  gtk3
      cairo_paint(cr);
      gdk_cairo_set_source_pixbuf(cr,thumbnail,0,0);                       //  gtk3
      cairo_paint(cr);
      g_object_unref(thumbnail);
   }

   return 0;
}


