/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.squarespace.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX                                                                 //  disable extern declarations
#include "fotoxx.h"


//  fotoxx main program

int main(int argc, char *argv[])
{
   char           lang[8] = "", *pp;
   int            Fclone=0, cloxx=0, cloyy=0, cloww=0, clohh=0;
   int            Fsyncfiles = 0;
   int            Ftranslate = 0;
   int            err;
   struct stat    statb;
   
   if (argc > 1 && strEqu(argv[1],"-v")) {
      printf(fotoxxvers "\n");                                             //  print version and exit
      return 0;
   }

   //  initialize externals to default values 
   //  (saved parameters will override some of these)

   Ffirsttime = 1;                                                         //  first time startup
   for (int ii = 0; ii < 8; ii++) wtnx[ii] = ii;                           //  thread numbers 0-7
   strcpy(jpeg_quality,def_jpeg_quality);
   mwgeom[0] = mwgeom[1] = 100;                                            //  default main window geometry
   mwgeom[2] = 800; mwgeom[3] = 500;
   Mscale = 1.0;
   trimsize[0] = 1600;
   trimsize[1] = 1000;
   editresize[0] = 1600;
   editresize[1] = 1200;
   batchresize[0] = 1600;
   batchresize[1] = 1200;
   lens_mm = 35;                                                           //  pano lens parameters
   lens_bow = 0;
   gridcount[0] = gridcount[1] = 5;                                        //  default grid line counts
   ss_funcs[0] = 1;                                                        //  enable slide show arrow key navigation
   tbar_style = "both";                                                    //  default toolbar style
   startdisplay = "blank";                                                 //  start with blank window
   Fwarnoverwrite = 1;                                                     //  warn original file overwrite
   zoomcount = 2;                                                          //  zooms to reach 2x image size    v.12.03
   zoomratio = sqrtf(2);                                                   //  corresp. zoom ratio 

   //  process command line parameters

   gtk_init(&argc,&argv);                                                  //  initz. GTK & strip GTK params

   for (int ii = 1; ii < argc; ii++)                                       //  command line parameters
   {
      pp = argv[ii];
      if (strcmpv(pp,"-debug","-d",null))                                  //  -debug
         Fdebug = 1;
      else if (strcmpv(pp,"-lang","-l",null) && argc > ii+1)               //  -lang lc_RC   (language/region code)
         strncpy0(lang,argv[++ii],7);
      else if (strcmpv(pp,"-clone",null) && argc > ii+4) {                 //  -clone xpos ypos width height   v.12.03
         Fclone = 1;
         cloxx = atoi(argv[ii+1]);                                         //  window position and size
         cloyy = atoi(argv[ii+2]);                                         //    passed from parent instance
         cloww = atoi(argv[ii+3]);
         clohh = atoi(argv[ii+4]);
         ii += 4;
      }
      else if (strcmpv(pp,"-slideshow","-ss",null) && argc > ii+1)         //  -slideshow /image/file.jpg
         SS_imagefile = strdupz(argv[++ii]);
      else if (strcmpv(pp,"-music","-m",null) && argc > ii+1)              //  -music /music/file.ogg
         SS_musicfile = strdupz(argv[++ii]);
      else if (strcmpv(pp,"-recent","-r",null))                            //  recent files gallery
         Frecent = 1;
      else if (strcmpv(pp,"-previous","-prev","-p",null))                  //  open previous file
         Fprev = 1;
      else if (strcmpv(pp,"-blank","-b",null))                             //  start with blank window         v.12.01
         Fblank = 1;
      else if (strcmpv(pp,"-syncfiles",null)) {                            //  spawned sync files process
         Fsyncfiles = 1;
         pp = argv[++ii];
         if (strEqu(pp,"auto")) Fautosync = 1;                             //  set auto or manual sync
         if (strEqu(pp,"manual")) Fmansync = 1;
      }
      else if (strcmpv(pp,"-translate","-t",null))                         //  start online translation
         Ftranslate = 1;
      else  initial_file = strdupz(pp);                                    //  must be initial file or directory
   }

   zinitapp("fotoxx",null);                                                //  get app directories
   ZTXinit(lang);                                                          //  setup locale and translations
   
   //  get files in user directory /home/<user>/.fotoxx/*

   snprintf(syncfiles_lock,199,"%s/syncfiles_lock",get_zuserdir());        //  file sync interlock
   snprintf(exiftool_input,199,"%s/exiftool_input",get_zuserdir());        //  exiftool_server input file
   snprintf(topdirk_file,199,"%s/top_directory",get_zuserdir());           //  top_directories file
   snprintf(search_index_file,199,"%s/search_index",get_zuserdir());       //  search_index file
   snprintf(tags_defined_file,199,"%s/tags_defined",get_zuserdir());       //  defined tags file
   snprintf(recentfiles_file,199,"%s/recent_files",get_zuserdir());        //  recent files file
   snprintf(saved_areas_dirk,199,"%s/saved_areas",get_zuserdir());         //  saved areas directory
   snprintf(collections_dirk,199,"%s/collections",get_zuserdir());         //  collections directory
   snprintf(saved_curves_dirk,199,"%s/saved_curves",get_zuserdir());       //  saved curves directory
   snprintf(annotations_dirk,199,"%s/annotations",get_zuserdir());         //  annotations directory

   err = stat(saved_areas_dirk,&statb);                                    //  create any missing directories
   if (err) mkdir(saved_areas_dirk,0750);
   err = stat(collections_dirk,&statb);
   if (err) mkdir(collections_dirk,0750);
   err = stat(saved_curves_dirk,&statb);
   if (err) mkdir(saved_curves_dirk,0750);
   err = stat(annotations_dirk,&statb);
   if (err) mkdir(annotations_dirk,0750);

   //  restore parameters from last session

   load_params();

   //  spawn file sync process or begin file sync process
   
   if (! Fsyncfiles && ! Fclone) {                                         //  this is the user GUI process
      printf("spawn sync files subprocess \n");
      err = system("fotoxx -syncfiles auto &");                            //  spawn process for sync files
      if (err) {
         printf("error: %s \n",wstrerror(err));                            //  quit if failed
         return 1;
      }
   }

   if (Fsyncfiles) {                                                       //  this is spawned sync process
      g_idle_add(syncfiles_func,0);                                        //  do the sync function only
      gtk_main();                                                          //  quit when done
      return 0;
   }

   //  create main window, menus, toolbar

   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create main window
   gtk_window_set_title(MWIN,fotoxxvers);
   
   mVbox = gtk_vbox_new(0,0);                                              //  add vert. packing box
   gtk_container_add(GTK_CONTAINER(mWin),mVbox);

   mMbar = create_menubar(mVbox);                                          //  add menus / sub-menus
   
   if (Ftranslate) {                                                       //  start online translation
      zfuncs::F1_help_topic = "translate";                                 //  (to process the menu names)
      ZTX_translation_start(mWin);
   }

   GtkWidget *mFile = add_menubar_item(mMbar,ZTX("File"),topmenufunc);
      add_submenu_item(mFile,       ZTX("Image Gallery"),            m_gallery);
      add_submenu_item(mFile,       ZTX("Clone Fotoxx"),             m_clone);
      add_submenu_item(mFile,       ZTX("Open Image File"),          m_open);
      add_submenu_item(mFile,       ZTX("Open Previous File"),       m_previous);
      add_submenu_item(mFile,       ZTX("Open Recent File"),         m_recent);
      add_submenu_item(mFile,       ZTX("Save to Same File"),        m_save);
      add_submenu_item(mFile,       ZTX("Save to New Version"),      m_savevers);
      add_submenu_item(mFile,       ZTX("Save to New File"),         m_saveas);
      add_submenu_item(mFile,       ZTX("Create Blank Image"),       m_create);
      add_submenu_item(mFile,       ZTX("Trash Image File"),         m_trash);
      add_submenu_item(mFile,       ZTX("Rename Image File"),        m_rename);
      add_submenu_item(mFile,       ZTX("Batch Rename Files"),       m_batchrename);
      add_submenu_item(mFile,       ZTX("Print Image File"),         m_print);
      add_submenu_item(mFile,       ZTX("Quit fotoxx"),              m_quit);

   GtkWidget *mTools = add_menubar_item(mMbar,ZTX("Tools"),topmenufunc);
      add_submenu_item(mTools,      ZTX("Manage Collections"),       m_manage_collections);
      add_submenu_item(mTools,      ZTX("Move Collections"),         m_move_collection);
      add_submenu_item(mTools,      ZTX("Batch Convert"),            m_batchconvert);
      add_submenu_item(mTools,      ZTX("Convert RAW files"),        m_conv_raw);
      add_submenu_item(mTools,      ZTX("Slide Show"),               m_slideshow);
      add_submenu_item(mTools,      ZTX("Synchronize Files"),        m_syncfiles);
      add_submenu_item(mTools,      ZTX("Show RGB"),                 m_show_RGB);
      add_submenu_item(mTools,      ZTX("Grid Lines"),               m_gridlines);
      add_submenu_item(mTools,      ZTX("Burn Images to CD/DVD"),    m_burn);
      add_submenu_item(mTools,      ZTX("Check Monitor"),            m_moncheck);
      add_submenu_item(mTools,      ZTX("Monitor Gamma"),            m_mongamma);
      add_submenu_item(mTools,      ZTX("Brightness Distribution"),  m_histogram);
      add_submenu_item(mTools,      ZTX("Change Language"),          m_lang);
      add_submenu_item(mTools,          "Edit Translations",         m_translate);
      add_submenu_item(mTools,      ZTX("Menu and Launcher"),        m_menu_launcher);
      add_submenu_item(mTools,      ZTX("User Settings"),            m_settings);
      add_submenu_item(mTools,      ZTX("Resources"),                m_resources);

   GtkWidget *mInfo = add_menubar_item(mMbar,"Info",topmenufunc);
      add_submenu_item(mInfo,       ZTX("Edit Caption/Comments"),    m_edit_cctext);
      add_submenu_item(mInfo,       ZTX("Edit Tags"),                m_edit_tags);
      add_submenu_item(mInfo,       ZTX("Batch Add Tags"),           m_batchAddTags);
      add_submenu_item(mInfo,       ZTX("Batch Delete Tag"),         m_batchDelTag);
      add_submenu_item(mInfo,       ZTX("View Info (short)"),        m_info_view_short);
      add_submenu_item(mInfo,       ZTX("View Info (long)"),         m_info_view_long);
      add_submenu_item(mInfo,       ZTX("Edit Info"),                m_info_edit);
      add_submenu_item(mInfo,       ZTX("Delete Info"),              m_info_delete);
      add_submenu_item(mInfo,       ZTX("Search Images"),            m_search_images);
      add_submenu_item(mInfo,       ZTX("Search Metadata"),          m_search_metadata);

   GtkWidget *mSelect = add_menubar_item(mMbar,ZTX("Select"),topmenufunc);
      add_submenu_item(mSelect,     ZTX("Select"),                   m_select);
      add_submenu_item(mSelect,     ZTX("Show"),                     m_select_show);
      add_submenu_item(mSelect,     ZTX("Hide"),                     m_select_hide);
      add_submenu_item(mSelect,     ZTX("Enable"),                   m_select_enable);
      add_submenu_item(mSelect,     ZTX("Disable"),                  m_select_disable);
      add_submenu_item(mSelect,     ZTX("Invert"),                   m_select_invert);
      add_submenu_item(mSelect,     ZTX("Unselect"),                 m_select_unselect);
      add_submenu_item(mSelect,     ZTX("Copy"),                     m_select_copy);
      add_submenu_item(mSelect,     ZTX("Paste"),                    m_select_paste);
      add_submenu_item(mSelect,     ZTX("Open"),                     m_select_open);
      add_submenu_item(mSelect,     ZTX("Save"),                     m_select_save);
      add_submenu_item(mSelect,     ZTX("Select Whole Image"),       m_select_whole_image);
      add_submenu_item(mSelect,     ZTX("Select and Edit"),          m_select_edit);

   GtkWidget *mTransf = add_menubar_item(mMbar,ZTX("Transform"),topmenufunc);
      add_submenu_item(mTransf,     ZTX("Rotate Image"),             m_rotate);
      add_submenu_item(mTransf,     ZTX("Trim Image"),               m_trim);
      add_submenu_item(mTransf,     ZTX("Auto-Trim Image"),          m_autotrim);
      add_submenu_item(mTransf,     ZTX("Resize Image"),             m_resize);
      add_submenu_item(mTransf,     ZTX("Annotate Image"),           m_annotate);
      add_submenu_item(mTransf,     ZTX("Flip Image"),               m_flip);
      add_submenu_item(mTransf,     ZTX("Make Negative"),            m_negate);
      add_submenu_item(mTransf,     ZTX("Unbend Image"),             m_unbend);
      add_submenu_item(mTransf,     ZTX("Keystone Correction"),      m_keystone);
      add_submenu_item(mTransf,     ZTX("Warp Image (area)"),        m_warp_area);
      add_submenu_item(mTransf,     ZTX("Warp Image (curved)"),      m_warp_curved);
      add_submenu_item(mTransf,     ZTX("Warp Image (linear)"),      m_warp_linear);
      add_submenu_item(mTransf,     ZTX("Warp Image (affine)"),      m_warp_affine);

   GtkWidget *mRetouch = add_menubar_item(mMbar,ZTX("Retouch"),topmenufunc);
      add_submenu_item(mRetouch,    ZTX("Brightness/Color"),         m_tune);
      add_submenu_item(mRetouch,    ZTX("Gamma Curves"),             m_gamma);
      add_submenu_item(mRetouch,    ZTX("Expand Brightness"),        m_xbrange);
      add_submenu_item(mRetouch,    ZTX("Flatten Brightness"),       m_flatten);
      add_submenu_item(mRetouch,    ZTX("Brightness Ramp"),          m_brightramp);
      add_submenu_item(mRetouch,    ZTX("Tone Mapping"),             m_tonemap);
      add_submenu_item(mRetouch,    ZTX("White Balance"),            m_whitebal);
      add_submenu_item(mRetouch,    ZTX("Match Colors"),             m_match_color);
      add_submenu_item(mRetouch,        "DRGB",                      m_DRGB);
      add_submenu_item(mRetouch,    ZTX("Revise RGB"),               m_revise_RGB);
      add_submenu_item(mRetouch,    ZTX("Red Eyes"),                 m_redeye);
      add_submenu_item(mRetouch,    ZTX("Blur Image"),               m_blur);
      add_submenu_item(mRetouch,    ZTX("Sharpen Image"),            m_sharpen);
      add_submenu_item(mRetouch,    ZTX("Reduce Noise"),             m_denoise);
      add_submenu_item(mRetouch,    ZTX("Smart Erase"),              m_smart_erase);
      add_submenu_item(mRetouch,    ZTX("Remove Dust"),              m_dust);
      add_submenu_item(mRetouch,    ZTX("Fix Stuck Pixels"),         m_stuckpix);
      add_submenu_item(mRetouch,    ZTX("Edit Pixels"),              m_pixedit);

   GtkWidget *mArt = add_menubar_item(mMbar,ZTX("Art"),topmenufunc);
      add_submenu_item(mArt,        ZTX("Color Depth"),              m_colordep);
      add_submenu_item(mArt,        ZTX("Drawing"),                  m_draw);
      add_submenu_item(mArt,        ZTX("Outlines"),                 m_outlines);
      add_submenu_item(mArt,        ZTX("Embossing"),                m_emboss);
      add_submenu_item(mArt,        ZTX("Tiles"),                    m_tiles);
      add_submenu_item(mArt,        ZTX("Dots"),                     m_dots);
      add_submenu_item(mArt,        ZTX("Painting"),                 m_painting);

   GtkWidget *mComb = add_menubar_item(mMbar,ZTX("Combine"),topmenufunc);
      add_submenu_item(mComb,       ZTX("High Dynamic Range"),       m_HDR);
      add_submenu_item(mComb,       ZTX("High Depth of Field"),      m_HDF);
      add_submenu_item(mComb,       ZTX("Stack / Paint"),            m_STP);
      add_submenu_item(mComb,       ZTX("Stack / Noise"),            m_STN);
      add_submenu_item(mComb,       ZTX("Panorama"),                 m_pano);
      add_submenu_item(mComb,       ZTX("Vertical Panorama"),        m_vpano);

   GtkWidget *mPlugins = add_menubar_item(mMbar,"Plugins",topmenufunc);
      add_submenu_item(mPlugins,ZTX("Edit Plugins"),m_edit_plugins);
      for (int ii = 0; ii < Nplugins; ii++) {
         pp = strstr(plugins[ii]," = ");
         if (! pp) continue;
         *pp = 0;
         add_submenu_item(mPlugins,plugins[ii],m_run_plugin);
         *pp = ' ';
      }

   GtkWidget *mHelp = add_menubar_item(mMbar,ZTX("Help"),topmenufunc);
      add_submenu_item(mHelp,       ZTX("About"),                    m_help);
      add_submenu_item(mHelp,       ZTX("User Guide"),               m_help);
      add_submenu_item(mHelp,       ZTX("User Guide Changes"),       m_help);
      add_submenu_item(mHelp,       ZTX("Edit Functions Summary"),   m_help);
      add_submenu_item(mHelp,       ZTX("Keyboard Shortcuts"),       m_help);
      add_submenu_item(mHelp,       "README",                        m_help);
      add_submenu_item(mHelp,       ZTX("Change Log"),               m_help);
      add_submenu_item(mHelp,       ZTX("Translations"),             m_help);
      add_submenu_item(mHelp,       ZTX("Home Page"),                m_help);

   mTbar = create_toolbar(mVbox);                                          //  toolbar buttons
      add_toolbar_button(mTbar,  ZTX("Gallery"),  ZTX("Image Gallery"),        "gallery.png",  m_gallery);
      add_toolbar_button(mTbar,  ZTX("Open"),     ZTX("Open Image File"),      "open.png",     m_open);
      add_toolbar_button(mTbar,  ZTX("Prev"),     ZTX("Open Previous File"),   "prev.png",     m_prev);
      add_toolbar_button(mTbar,  ZTX("Next"),     ZTX("Open Next File"),       "next.png",     m_next);
      add_toolbar_button(mTbar,  "Zoom+",         ZTX("Zoom-in (bigger)"),     "zoom+.png",    m_zoom);
      add_toolbar_button(mTbar,  "Zoom-",         ZTX("Zoom-out (smaller)"),   "zoom-.png",    m_zoom);
      add_toolbar_button(mTbar,  ZTX("Undo"),     ZTX("Undo One Edit"),        "undo.png",     m_undo);
      add_toolbar_button(mTbar,  ZTX("Redo"),     ZTX("Redo One Edit"),        "redo.png",     m_redo);
      add_toolbar_button(mTbar,  "separator",     null,                        null,           null);
      add_toolbar_button(mTbar,  Bcolor,          ZTX("Foreground Color"),    "fg-color.png",  m_fg_color);
      add_toolbar_button(mTbar,  "separator",     null,                        null,           null);
      add_toolbar_button(mTbar,  ZTX("Save"),     ZTX("Save to Same File"),    "save.png",     m_save);
      add_toolbar_button(mTbar,  ZTX("Save+V"),   ZTX("Save to New Version"),  "save+V.png",   m_savevers);
      add_toolbar_button(mTbar,  ZTX("Save+F"),   ZTX("Save to New File"),     "save+F.png",   m_saveas);
      add_toolbar_button(mTbar,  "separator",     null,                        null,           null);
      add_toolbar_button(mTbar,  ZTX("Trash"),    ZTX("Move Image to Trash"),  "trash.png",    m_trash);
      add_toolbar_button(mTbar,  "separator",     null,                        null,           null);
      add_toolbar_button(mTbar,  ZTX("Quit"),     ZTX("Quit fotoxx"),          "quit.png",     m_quit);
      add_toolbar_button(mTbar,  ZTX("Help"),     ZTX("Current Topic"),        "help.png",     m_help);


   if (strEqu(tbar_style,"icons"))                                         //  set toolbar style      v.12.01
      gtk_toolbar_set_style(GTK_TOOLBAR(mTbar),GTK_TOOLBAR_ICONS);
   else if (strEqu(tbar_style,"text"))
      gtk_toolbar_set_style(GTK_TOOLBAR(mTbar),GTK_TOOLBAR_TEXT);
   else gtk_toolbar_set_style(GTK_TOOLBAR(mTbar),GTK_TOOLBAR_BOTH);

   drWin = gtk_drawing_area_new();                                         //  add drawing window
   gtk_box_pack_start(GTK_BOX(mVbox),drWin,1,1,0);

   STbar = create_stbar(mVbox);                                            //  add status bar

   G_SIGNAL(mWin,"delete_event",delete_event,0);                           //  connect signals
   G_SIGNAL(mWin,"destroy",destroy_event,0);
   G_SIGNAL(drWin,"draw",mwpaint1,0);                                      //  gtk3

   gtk_widget_add_events(drWin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events
   gtk_widget_add_events(drWin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(drWin,GDK_BUTTON_MOTION_MASK);                    //  motion with button pressed
   gtk_widget_add_events(drWin,GDK_POINTER_MOTION_MASK);                   //  pointer motion
   gtk_widget_add_events(drWin,GDK_SCROLL_MASK);                           //  scroll wheel              v.12.01
   G_SIGNAL(drWin,"button-press-event",mouse_event,0);
   G_SIGNAL(drWin,"button-release-event",mouse_event,0);
   G_SIGNAL(drWin,"motion-notify-event",mouse_event,0);
   G_SIGNAL(drWin,"scroll-event",mouse_event,0);
   G_SIGNAL(mWin,"window-state-event",state_event,0);

   G_SIGNAL(mWin,"key-press-event",KBpress,0);                             //  connect KB events
   G_SIGNAL(mWin,"key-release-event",KBrelease,0);
   
   drag_drop_connect(drWin,m_open_drag);                                   //  connect drag-drop event

   if (Fclone) {                                                           //  clone: open new window          v.12.03
      gtk_window_move(MWIN,cloxx+20,cloyy+30);                             //    slightly offset from old window
      gtk_window_resize(MWIN,cloww,clohh);
   }
   else {
      gtk_window_move(MWIN,mwgeom[0],mwgeom[1]);                           //  main window geometry
      gtk_window_set_default_size(MWIN,mwgeom[2],mwgeom[3]);               //  (defaults or last session params)
   }

   gtk_widget_show_all(mWin);                                              //  show all widgets
   gdkWin = gtk_widget_get_window(drWin);                                  //  drawing window realized now     gtk3

   black[0] = black[1] = black[2] = 0;                                     //  define RGB colors
   white[0] = white[1] = white[2] = 255;
   red[0] = 255; red[1] = red[2] = 0;
   green[1] = 255; green[0] = green[2] = 0;
   lgray[0] = lgray[1] = lgray[2] = 230;
   dgray[0] = lgray[1] = lgray[2] = 127;

   fg_color = green;                                                       //  initial foreground drawing color

   GDKlgray.red = GDKlgray.green = GDKlgray.blue = 59000;                  //  GDK colors for light gray and black
   GDKblack.red = GDKblack.green = GDKblack.blue = 0;

   arrowcursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                       //  cursor for selection
   dragcursor = gdk_cursor_new(GDK_CROSSHAIR);                             //  cursor for dragging
   drawcursor = gdk_cursor_new(GDK_PENCIL);                                //  cursor for drawing lines

   int initzfunc(void *);
   g_timeout_add(0,initzfunc,0);                                           //  initz. call from gtk_main()     gtk3   
   gtk_main();                                                             //  start processing window events
   return 0;
}


/**************************************************************************/

//  Initial function called from gtk_main() at startup.
//  This function MUST return 0.

int initzfunc(void *)
{
   int            err, flag, npid;
   int            ftype, fdirk = 0;
   char           procfile[20], *ppv;
   cchar          *pp, *pp2;
   struct stat    statb;

   cchar *ftmess = ZTX("First time Fotoxx startup. \n"
                       "File synchronization process may need some time. \n"
                       "See menu: Help > User Guide.");
   menulock(1);                                                            //  block menus until initz. done

   snprintf(PIDstring,11,"%06d",getpid());                                 //  get process PID 6-digits

   printf(fotoxxvers "\n");                                                //  print Fotoxx version
   printf("install location: %s \n",get_zprefix());                        //  install location
   printf("last Fotoxx session: %s \n",last_session);                      //  last session exit time

   if (Ffirsttime) {                                                       //  first time startup
      zmessageACK(mWin,ftmess);                                            //  v.12.03
      printf("%s",ftmess);
      Ffirsttime = 0;
      save_params();
   }

   Nwt = sysconf(_SC_NPROCESSORS_ONLN);                                    //  get SMP CPU count
   if (Nwt <= 0) Nwt = 1;
   if (Nwt > max_threads) Nwt = max_threads;                               //  compile time limit
   printf("using %d threads \n",Nwt);

   err = system("which exiftool");                                         //  check for exiftool
   if (! err) Fexiftool = 1;
   err = system("which xdg-open");                                         //  check for xdg-open
   if (! err) Fxdgopen = 1;
   err = system("which ufraw-batch");                                      //  check for ufraw-batch
   if (! err) Fufraw = 1;
   
   undo_files = zmalloc(200);
   *undo_files = 0;                                                        //  look for orphaned undo files
   strncatv(undo_files,199,get_zuserdir(),"/*_undo_*",null);               //  /home/user/.fotoxx/*_undo_*
   flag = 1;
   while ((pp = SearchWild(undo_files,flag)))
   {
      pp2 = strstr(pp,".fotoxx/");
      if (! pp2) continue;
      npid = atoi(pp2+8);                                                  //  pid of file owner
      snprintf(procfile,19,"/proc/%d",npid);
      err = stat(procfile,&statb);
      if (! err) continue;                                                 //  pid is active, keep file
      printf("orphaned undo file deleted: %s \n",pp);
      err = remove(pp);                                                    //  delete file
   }
   
   exiftool_server(null);                                                  //  kill orphan exiftool process    v.12.03

   *undo_files = 0;                                                        //  setup undo filespec template
   strncatv(undo_files,199,get_zuserdir(),"/",PIDstring,"_undo_nn",null);  //  /home/user/.fotoxx/pppppp_undo_nn

   editlog = pvlist_create(maxedits);                                      //  history log of image edits done
   for (int ii = 0; ii < maxedits; ii++)                                   //  pre-load all pvlist entries
      pvlist_append(editlog,"nothing");

   mutex_init(&Fpixmap_lock,0);                                            //  setup lock for edit pixmaps
   zdialog_positions("load");                                              //  load saved dialog positions
   
   //  set up current file and directory from saved parameters or command line

   if (Ntopdirk) curr_dirk = strdupz(topdirk[0]);                          //  use 1st top direk if defined
   else {
      ppv = getcwd(command,ccc);                                           //  or use current directory
      if (ppv) curr_dirk = strdupz(ppv);
   }
   
   if (initial_file) {                                                     //  from command line or previous session
      ppv = realpath(initial_file,0);                                      //  prepend directory if needed
      if (ppv) {
         zfree(initial_file);
         initial_file = strdupz(ppv);
         free(ppv);
      }
      else {
         printf("invalid file: %s \n",initial_file);                       //  invalid file path
         zfree(initial_file);
         initial_file = 0;
      }
   }

   if (initial_file) {
      ftype = image_file_type(initial_file);
      if (ftype == 0) {                                                    //  non-existent file
         printf("invalid file: %s \n",initial_file);
         zfree(initial_file);
         initial_file = 0;
      }
      else if (ftype == 1) {                                               //  is a directory
         if (curr_dirk) zfree(curr_dirk);
         curr_dirk = initial_file;
         initial_file = 0;
         fdirk = 1;
      }
      else if (ftype == 2) {                                               //  supported image file type
         if (curr_dirk) zfree(curr_dirk);
         curr_dirk = strdupz(initial_file);                                //  set current directory from file
         ppv = strrchr(curr_dirk,'/');
         if (ppv) *ppv = 0;
      }
      else {
         printf("unsupported file type: %s \n",initial_file);              //  unsupported file type
         zfree(initial_file);
         initial_file = 0;
      }
   }

   if (curr_dirk) err = chdir(curr_dirk);                                  //  set current directory

   menulock(0);                                                            //  enable menus

   g_timeout_add(100,gtimefunc,0);                                         //  start periodic function (100 ms)
   
   //  process command line options

   if (SS_imagefile) {                                                     //  start slide show if wanted
      f_open(SS_imagefile,1);
      if (SS_musicfile) {
         sprintf(command,"xdg-open \"%s\" ",SS_musicfile);                 //  start music if wanted
         printf("command: %s \n",command);
         err = system(command);
      }
      m_slideshow(0,0);
      return 0;
   }

   if (initial_file)                                                       //  open initial file
      f_open(initial_file,1);
   
   else if (fdirk)                                                         //  open gallery for initial directory
      m_gallery(0,0);
   
   else if (Fprev)                                                         //  start with previous file
      m_previous(0,0);
   
   else if (Frecent)                                                       //  start with recent file
      m_recent(0,0);
   
   else if (Fblank) return 0;                                              //  start with blank window 
   
   //  if no command line option, get startup option from user settings
   
   else if (strEqu(startdisplay,"recent"))
      m_recent(0,0);

   else if (strEqu(startdisplay,"prev"))
      m_previous(0,0);

   else if (strEqu(startdisplay,"dirk")) {
      if (! startdirk || *startdirk != '/') return 0;
      if (curr_dirk) zfree(curr_dirk);
      curr_dirk = strdupz(startdirk);
      err = chdir(curr_dirk); 
      navi::thumbsize = 128;
      m_gallery(0,0);
   }

   else if (strEqu(startdisplay,"file"))
      f_open(startfile,1);

   return 0;                                                               //  start with blank window
}


/**************************************************************************/

//  Periodic function (100 milliseconds)
//  Avoid any thread usage of gtk/gdk functions.

int gtimefunc(void *arg)
{
   static zdialog    *zdmessage = 0;
   static int        domore = 0;
   char              busymess[20];
   int               syncfd, percent;
   int               px, py;

   if (Wrepaint) {                                                         //  window update needed
      if (mutex_trylock(&Fpixmap_lock) != 0) return 1;                     //  cannot lock pixmaps
      Wrepaint = 0;                                                        //  OK, do protected repaint
      gtk_widget_queue_draw(drWin);
      mutex_unlock(&Fpixmap_lock);
   }
   
   if (--domore > 0) return 1;
   domore = 5;

   if (Ffuncbusy) {                                                        //  something is processing
      if (Fbusy_goal) {
         percent = 100 * Fbusy_done / Fbusy_goal;                          //  construct "BUSY nn%"
         sprintf(busymess," BUSY %d%c ",percent,'%');
      }
      else strcpy(busymess," BUSY ");
      
      px = Iorgx / Mscale;
      py = Iorgy / Mscale;
      draw_text(px,py,busymess,"monospace 11");                            //  write upper left visible image window
   }

   update_statusbar();                                                     //  update status bar information

   if (Fslideshow) slideshow_next("timer");                                //  show next image if time is up

   if (threadmessage && ! zdmessage)                                       //  display message for thread process
      zdmessage = zmessage_post(mWin,0,threadmessage);                     //  (avoid use of GTK in threads)

   if (zdmessage && ! threadmessage)                                       //  message terminated by thread
      zdialog_free(zdmessage);
   
   if (zdmessage && zdmessage->sentinel != zdsentinel) {                   //  popup window closed by user
      zdmessage = 0;
      threadmessage = 0;
   }
   
   syncfd = global_lock(syncfiles_lock);                                   //  check for sync files process
   if (syncfd >= 0) {                                                      //  not busy
      global_unlock(syncfd,syncfiles_lock);
      if (Fsyncbusy) threadmessage = 0;                                    //  was busy, now done
      Fsyncbusy = 0;
   }
   else Fsyncbusy = 1;                                                     //  sync files busy

   return 1;
}


/**************************************************************************/

//  update status bar with image data and status
//  called from timer function

void update_statusbar()
{
   static double  time1 = 0, time2, cpu1 = 0, cpu2, cpuload;
   char           text1[300], text2[100];
   int            ww, hh, scale, bpc;
   double         file_MB = 1.0 / mega * curr_file_size;
   
   *text1 = *text2 = 0;
   
   if (! time1) {
      time1 = get_seconds();
      cpu1 = jobtime();
   }

   time2 = get_seconds();                                                  //  compute process cpu load %
   if (time2 - time1 > 1.0) {                                              //    at 1 second intervals
      cpu2 = jobtime();
      cpuload = 100.0 * (cpu2 - cpu1) / (time2 - time1);
      time1 = time2;
      cpu1 = cpu2;
   }

   sprintf(text1,"CPU %03.0f%c",cpuload,'%');                              //  CPU 023%
   
   if (curr_file)
   {
      if (E3pxm) {
         ww = E3ww;
         hh = E3hh;
      }
      else {
         ww = Fww;
         hh = Fhh;
      }

      bpc = curr_file_bpc;

      snprintf(text2,99,"  %dx%dx%d",ww,hh,bpc);                           //  2345x1234x16 (preview) 1.56MB 45%
      strcat(text1,text2);
      if (Fpreview) strcat(text1," (preview)");
      sprintf(text2," %.2fMB",file_MB);
      strcat(text1,text2);
      scale = Mscale * 100 + 0.5;
      sprintf(text2," %d%c",scale,'%');
      strcat(text1,text2);
   
      if (Pundo) {                                                         //  edit undo stack depth 
         snprintf(text2,99,"  edits: %d",Pundo);
         strcat(text1,text2);
      }
   }

   if (Fmenulock) strcat(text1,"  menu locked");
   if (Fsyncbusy) strcat(text1,"  file sync busy");
   if (sa_stat == 3) strcat(text1,"  area active");
   if (zfuncs::zdialog_busy) strcat(text1,"  dialog open");
   
   if (*SB_text) {                                                         //  application text
      strcat(text1,"  ");
      strcat(text1,SB_text);
   }

   stbar_message(STbar,text1);                                             //  write to status bar

   return;
}


/**************************************************************************/

//  functions for main window event signals

int delete_event()                                                         //  main window closed
{
   printf("main window delete event \n");
   if (mod_keep()) return 1;                                               //  allow user bailout
   Fshutdown++;                                                            //  shutdown in progress
   save_params();                                                          //  save parameters for next session
   zdialog_positions("save");                                              //  save dialog positions too
   free_resources();                                                       //  delete undo files
   return 0;
}

int destroy_event()                                                        //  main window destroyed
{
   printf("main window destroy event \n");
   Fshutdown++;                                                            //  shutdown in progress
   free_resources();                                                       //  delete undo files 
   exiftool_server(0);                                                     //  kill exiftool_server process    v.12.03
   exit(1);                                                                //  instead of gtk_main_quit();
   return 0;
}

int state_event(GtkWidget *, GdkEvent *event)                              //  main window state changed
{
   int state = ((GdkEventWindowState *) event)->new_window_state;          //  track window maximized status   v.12.03
   if (state == GDK_WINDOW_STATE_MAXIMIZED) Wmaximized = 1;
   else Wmaximized = 0;
   return 0;
}


//  process top-level menu entry

void topmenufunc(GtkWidget *, cchar *menu)
{
   topmenu = (char *) menu;                                                //  remember top-level menu in case
   return;                                                                 //    this is needed somewhere
}


/**************************************************************************/

//  GTK3 "draw" function for the main/drawing window.
//  Paint window when created, exposed, resized, or image modified (edited).
//  Update window image if Frefresh flag, scale change or window size change.
//  Draw the image section currently within the visible window. 
//  May NOT be called from threads.

int mwpaint1(GtkWidget *drWin, cairo_t *cr)                                //  overhauled    v.12.03
{
   GdkPixbuf      *pixbuf;
   RGB            *pixels;
   PXM            *pxmtemp;
   char           busymess[20];
   int            incrx, incry;                                            //  mouse drag
   static int     pincrx = 0, pincry = 0;                                  //  prior mouse drag
   static int     pdww = 0, pdhh = 0;                                      //  prior window size
   static float   pscale = 1;                                              //  prior scale
   float          wscale, hscale;
   float          Mx, My;
   int            centerx, centery; 
   int            refresh = 0;                                             //  flag, full image refesh needed
   int            mousex, mousey;                                          //  mouse position after zoom
   int            rs, px, py, percent;
   
   if (Fshutdown) return 0;                                                //  shutdown underway
   if (! Fpxm) return 0;                                                   //  no image

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps 

   Dww = gdk_window_get_width(gdkWin);                                     //  (new) drawing window size
   Dhh = gdk_window_get_height(gdkWin);
   if (Dww < 20 || Dhh < 20) return 0;                                     //  too small

   if (Frefresh) {                                                         //  image was updated
      refresh++;                                                           //  image refresh needed
      Frefresh = 0;
   }

   centerx = (Iorgx + 0.5 * dww) / Mscale;                                 //  center of window, image space
   centery = (Iorgy + 0.5 * dhh) / Mscale;                                 //  (before window or scale change)

   if (Dww != pdww || Dhh != pdhh) {                                       //  window size changed
      refresh++;                                                           //  image refresh needed
      pdww = Dww; 
      pdhh = Dhh;
   }

   if (E3pxm) {                                                            //  get image size
      Iww = E3ww;
      Ihh = E3hh;                                                          //  use edited image size
   }
   else {
      Iww = Fww;                                                           //  use original image
      Ihh = Fhh;
   }
   
   if (Fzoom == 0) {                                                       //  scale to fit window
      wscale = 1.0 * Dww / Iww;
      hscale = 1.0 * Dhh / Ihh;
      if (wscale < hscale) Mscale = wscale;                                //  use greatest ww/hh ratio
      else  Mscale = hscale;
      if (Iww < Dww && Ihh < Dhh && ! Fblowup) Mscale = 1.0;               //  small image 1x unless Fblowup
      zoomx = zoomy = 0;                                                   //  no zoom target
   }
   else Mscale = Fzoom;                                                    //  scale to Fzoom level
   
   iww = Iww * Mscale;                                                     //  scaled image size for window
   ihh = Ihh * Mscale;

   dww = Dww;                                                              //  image fitting inside window
   if (iww < dww) dww = iww;                                               //  image < window size
   dhh = Dhh;
   if (ihh < dhh) dhh = ihh;

   if (Mscale != pscale) {                                                 //  scale changed
      Iorgx = Mscale * centerx - 0.5 * dww;                                //  change origin to keep same center
      Iorgy = Mscale * centery - 0.5 * dhh;                                //  (subject to later rules)
      pscale = Mscale;                                                     //  previous scale = new scale
      refresh++;                                                           //  full image refresh needed
   }

   if (refresh) {
      if (E3pxm) pxmtemp = E3pxm;                                          //  use edited image if there
      else pxmtemp = Fpxm;                                                 //  else use original image
      if (Wpixels) zfree(Wpixels);
      Wpixels = PXM_RGB_rescale(pxmtemp,iww,ihh);                          //  scale and convert to RGB (uint8)
   }

   if ((Mxdrag || Mydrag) && Fmousemain)                                   //  scroll via mouse drag
   {
      incrx = Mxdrag - Mxdown;                                             //  mouse shift, image space
      incry = Mydrag - Mydown;
      if (pincrx > 0 && incrx < 0) incrx = 0;                              //  stop bounce from direction change
      else if (pincrx < 0 && incrx > 0) incrx = 0;
      if (pincry > 0 && incry < 0) incry = 0;
      else if (pincry < 0 && incry > 0) incry = 0;
      pincrx = incrx;                                                      //  save for next pass
      pincry = incry;
      Mx = 1.2 * (iww - dww) / dww;                                        //  drag magnification factor
      My = 1.2 * (ihh - dhh) / dhh;
      Mx = Mx * incrx;                                                     //  image shift, magnified
      My = My * incry;
      Iorgx += Mx * Mscale;                                                //  image origin in window + shift
      Iorgy += My * Mscale;
      Mxdown = Mxdrag + Mx;                                                //  mouse position + shift
      Mydown = Mydrag + My;
      zoomx = zoomy = 0;                                                   //  no zoom target
   }

   if (dww < Dww) Dorgx = 0.5 * (Dww - dww);                               //  if scaled image width < window width
   else Dorgx = 0;                                                         //    center image in window
   
   if (dhh < Dhh) Dorgy = 0.5 * (Dhh - dhh);                               //  same for image height
   else Dorgy = 0;

   if (zoomx || zoomy) {                                                   //  requested zoom center
      Iorgx = Mscale * zoomx - 0.5 * dww;                                  //  corresp. image position
      Iorgy = Mscale * zoomy - 0.5 * dhh;
   }
   
   if (Iorgx < 0) Iorgx = 0;                                               //  keep max. image within window
   if (Iorgy < 0) Iorgy = 0;
   if (Iorgx + dww > iww) Iorgx = iww - dww;
   if (Iorgy + dhh > ihh) Iorgy = ihh - dhh;
   
   if (zoomx || zoomy) {                                                   //  zoom target
      mousex = zoomx * Mscale - Iorgx + Dorgx;
      mousey = zoomy * Mscale - Iorgy + Dorgy;                             //  mouse pointer follows target
      gdk_window_move_pointer(gdkWin,mousex,mousey);
      zoomx = zoomy = 0;                                                   //  reset zoom target
   }

   rs = iww * 3;                                                           //  image/pixbuf row stride
   pixels = Wpixels + (Iorgy * iww + Iorgx) * 3;                           //  image origin in window
   pixbuf = gdk_pixbuf_new_from_data(pixels,GDKRGB,0,8,dww,dhh,rs,0,0);    //  draw image area within window

   mwcr = cr;                                                              //  cairo context for draw funcs below
   gdk_cairo_set_source_pixbuf(mwcr,pixbuf,Dorgx,Dorgy);
   cairo_paint(mwcr);
   g_object_unref(pixbuf);

   cairo_set_line_width(mwcr,1);
   if (Ntoplines) draw_toplines(1);                                        //  draw line overlays
   if (Fgrid) draw_gridlines();                                            //  draw grid lines
   if (Ntoptext) draw_toptext();                                           //  draw text strings
   if (Ntopcircles) draw_topcircles();                                     //  draw circles
   if (Fshowarea) sa_show(1);                                              //  draw select area outline
   
   if (Ffuncbusy) {                                                        //  something is processing
      if (Fbusy_goal) {
         percent = 100 * Fbusy_done / Fbusy_goal;                          //  construct "BUSY nn%"
         sprintf(busymess," BUSY %d%c ",percent,'%');
      }
      else strcpy(busymess," BUSY ");

      px = Iorgx / Mscale;
      py = Iorgy / Mscale;
      draw_text(px,py,busymess,"monospace 11");                            //  write upper left visible image window
   }

   mwcr = 0;                                                               //  cairo context invalid

   mutex_unlock(&Fpixmap_lock);                                            //  unlock pixmaps
   Wpainted++;                                                             //  notify edit function of repaint
   if (winhisto) gtk_widget_queue_draw(winhistoH);                         //  update histogram if active

   return 0;
}


/**************************************************************************/

//  Cause (modified) output image to get repainted soon.
//  Same as mwpaint1() but delayed. MAY be called from threads.

void mwpaint2()
{
   Frefresh++;                                                             //  req. full window refresh
   Wrepaint++;                                                             //  req. repaint by periodic function
   return;
}


//  Update a section of the window image Wpixels from a modified section of E3pxm.
//  px3, py3, ww3, hh3: modified section within E3pxm to be updated and repainted.
//  May NOT be called from threads.

void mwpaint3(int px3, int py3, int ww3, int hh3)                          //  overhauled       v.12.03
{
   PXM_RGB_update(px3,py3,ww3,hh3);                                        //  update window image from E3pxm
   mwpaint4(px3,py3,ww3,hh3);                                              //  update window from window image
   return;
}


//  Repaint a section of the window image in the visible window.
//  px3, py3, ww3, hh3: area to be repainted (in image space).
//  May NOT be called from threads.

void mwpaint4(int px3, int py3, int ww3, int hh3)                          //  overhauled       v.12.03
{
   GdkPixbuf   *pixbuf;
   RGB         *pixels;
   int         px2, py2, ww2, hh2;
   int         crflag = 0;

   px2 = Mscale * px3 - 2;                                                 //  image space to Wpixels space
   py2 = Mscale * py3 - 2;                                                 //  (expanded)
   ww2 = Mscale * ww3 + 2 / Mscale + 4;
   hh2 = Mscale * hh3 + 2 / Mscale + 4;

   if (px2 + ww2 > iww) ww2 = iww - px2;                                   //  avoid overshoot
   if (py2 + hh2 > ihh) hh2 = ihh - py2;
   
   if (px2 < Iorgx) {                                                      //  reduce to currently visible window
      ww2 = ww2 - (Iorgx - px2);
      px2 = Iorgx;
   }

   if (py2 < Iorgy) {
      hh2 = hh2 - (Iorgy - py2);
      py2 = Iorgy;
   }

   if (ww2 <= 0 || hh2 <= 0) return;                                       //  completely outside visible window
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }
   
   pixels = Wpixels + (iww * py2 + px2) * 3;                               //  make pixbuf of updated section
   pixbuf = gdk_pixbuf_new_from_data(pixels,GDKRGB,0,8,ww2,hh2,iww*3,0,0);

   px2 = px2 - Iorgx + Dorgx;                                              //  section position in window
   py2 = py2 - Iorgy + Dorgy;

   gdk_cairo_set_source_pixbuf(mwcr,pixbuf,px2,py2);                       //  draw area to window
   cairo_paint(mwcr);
   g_object_unref(pixbuf);
   
   if (Fshowarea) { 
      px3 = (px2 - Dorgx + Iorgx) / Mscale;                                //  back to image scale, expanded
      py3 = (py2 - Dorgy + Iorgy) / Mscale;
      ww3 = ww2 / Mscale + 2;
      hh3 = hh2 / Mscale + 2;
      sa_show_rect(px3,py3,ww3,hh3);                                       //  refresh select area outline
   }

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  mouse event function - capture buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   void mouse_convert(int &xpos, int &ypos);

   static int     bdtime = 0, butime = 0, mbusy = 0;
   int            button, time, type, scroll;

   type = event->type;
   button = event->button;                                                 //  button, 1/2/3 = left/center/right
   time = event->time;
   Mxposn = event->x;                                                      //  mouse position in window
   Myposn = event->y;
   scroll = ((GdkEventScroll *) event)->direction;                         //  scroll wheel event
   
   mouse_convert(Mxposn,Myposn);                                           //  convert to image space
   
   if (type == GDK_SCROLL) {                                               //  scroll wheel = zoom          v.12.01
      zoomx = Mxposn;                                                      //  zoom center = mouse position (doriano)
      zoomy = Myposn;
      if (scroll == GDK_SCROLL_UP) m_zoom(null,"in");
      if (scroll == GDK_SCROLL_DOWN) m_zoom(null,"out");
      return;
   }

   if (type == GDK_MOTION_NOTIFY) {
      if (mbusy) return;                                                   //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (type == GDK_BUTTON_PRESS) {                                         //  button down
      bdtime = time;                                                       //  time of button down
      Mxdown = Mxposn;                                                     //  position at button down time
      Mydown = Myposn;
      if (button) {
         Mdrag++;                                                          //  possible drag start
         Mbutton = button;
      }
      Mxdrag = Mydrag = 0;
   }

   if (type == GDK_BUTTON_RELEASE) {                                       //  button up
      Mxclick = Myclick  = 0;                                              //  reset click status
      butime = time;                                                       //  time of button up
      if (butime - bdtime < 400) {                                         //  less than 0.4 secs down
         if (Mxposn == Mxdown && Myposn == Mydown) {                       //       and not moving
            if (Mbutton == 1) LMclick++;                                   //  left mouse click
            if (Mbutton == 3) RMclick++;                                   //  right mouse click
            Mxclick = Mxdown;                                              //  click = button down position
            Myclick = Mydown;
            if (button == 2) {                                             //  center button click
               zoomx = Mxposn;                                             //  re-center at mouse (Doriano)
               zoomy = Myposn;
               gtk_widget_queue_draw(drWin);                               //  v.12.03
            }
         }
      }
      Mxdown = Mydown = Mxdrag = Mydrag = Mdrag = Mbutton = 0;             //  forget buttons and drag
   }
   
   if (type == GDK_MOTION_NOTIFY && Mdrag) {                               //  drag underway
      Mxdrag = Mxposn;
      Mydrag = Myposn;
   }
   
   Fmousemain = 1;                                                         //  mouse acts on main window
   if (Mcapture) Fmousemain = 0;
   if (mouseCBfunc) Fmousemain = 0;
   if (KBcontrolkey) Fmousemain = 1;                                       //  Ctrl key >> mouse acts on main window
   
   if (mouseCBfunc && ! Fmousemain) {                                      //  pass to handler function
      if (mbusy) return;
      mbusy++;                                                             //  stop re-entrance
      (* mouseCBfunc)();
      mbusy = 0;
      return;
   }
   
   if (LMclick && Fmousemain) {                                            //  left click = zoom request
      LMclick = 0;
      zoomx = Mxclick;                                                     //  zoom center = mouse
      zoomy = Myclick;
      m_zoom(null,"in");
   }

   if (RMclick && Fmousemain) {                                            //  right click
      RMclick = 0;
      if (Fzoom) {
         zoomx = zoomy = 0;                                                //  reset zoom to fit window
         m_zoom(null,"fit");
      }
      else if (edit_coll_fname && curr_file)                               //  pass file to collection manager
         edit_coll_popmenu(null,curr_file,-1);
   }

   if ((Mxdrag || Mydrag) && Fmousemain)                                   //  drag = scroll
      gtk_widget_queue_draw(drWin);

   return;
}


//  convert mouse position from window space to image space

void mouse_convert(int &xpos, int &ypos)
{
   xpos = (xpos - Dorgx + Iorgx) / Mscale + 0.5;                           //  v.12.03
   ypos = (ypos - Dorgy + Iorgy) / Mscale + 0.5;

   if (xpos < 0) xpos = 0;                                                 //  if outside image put at edge
   if (ypos < 0) ypos = 0;

   if (E3pxm) { 
      if (xpos >= E3ww) xpos = E3ww-1;
      if (ypos >= E3hh) ypos = E3hh-1;
   }
   else {
      if (xpos >= Fww) xpos = Fww-1;
      if (ypos >= Fhh) ypos = Fhh-1;
   }
   
   return;
}


/**************************************************************************/

//  keyboard event function - some toolbar buttons have KB equivalents
//  GDK key symbols: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int KBpress(GtkWidget *win, GdkEventKey *event, void *)                    //  prevent propagation of key-press
{                                                                          //    events to toolbar buttons
   KBkey = event->keyval;

   if (KBkey == GDK_KEY_Control_L || KBkey == GDK_KEY_Control_R)           //  Ctrl key is pressed
       KBcontrolkey = 1;

   if (KBkey == GDK_KEY_Shift_L || KBkey == GDK_KEY_Shift_R)               //  Shift key is pressed
       KBshiftkey = 1;
   
   if (KBkey == GDK_KEY_a || KBkey == GDK_KEY_A)                           //  'A' key is pressed
       KB_A_key = 1;                                                       //  (undo/redo --> undo/redo ALL)

   return 1;
}

int KBrelease(GtkWidget *win, GdkEventKey *event, void *)
{
   int      angle = 0;
   PXM      *pxm;
   
   KBkey = event->keyval;
   
   if (KBcapture) return 1;                                                //  let function handle it

   if (KBkey == GDK_KEY_Control_L   ||                                     //  Ctrl key released
       KBkey == GDK_KEY_Control_R) KBcontrolkey = 0;
   
   if (KBkey == GDK_KEY_Shift_L   ||                                       //  Shift key released
       KBkey == GDK_KEY_Shift_R) KBshiftkey = 0;

   if (KBkey == GDK_KEY_a   ||                                             //  'A' key is released
       KBkey == GDK_KEY_A) KB_A_key = 0;
   
   if (KBcontrolkey) {
      if (KBkey == GDK_KEY_s) m_save(0,0);                                 //  Ctrl-* shortcuts
      if (KBkey == GDK_KEY_S) m_saveas(0,0);
      if (KBkey == GDK_KEY_v) m_savevers(0,0);
      if (KBkey == GDK_KEY_V) m_savevers(0,0);
      if (KBkey == GDK_KEY_q) m_quit(0,0);  
      if (KBkey == GDK_KEY_Q) m_quit(0,0);
   }

   if (KBkey == GDK_KEY_G || KBkey == GDK_KEY_g)                           //  toggle grid lines
      toggle_grid(2);

   if (Fslideshow) {
      if (KBkey == GDK_KEY_Left) slideshow_next("prev");                   //  arrow keys = prev/next image
      if (KBkey == GDK_KEY_Right) slideshow_next("next");
      if (KBkey == GDK_KEY_space) slideshow_next("pause");                 //  spacebar = pause/resume
      if (KBkey == GDK_KEY_Escape) m_slideshow(0,0);                       //  escape = exit slideshow
   }
   else {
      if (KBkey == GDK_KEY_Left) m_prev(0,0);                              //  arrow keys = prev/next image
      if (KBkey == GDK_KEY_Right) m_next(0,0);
   }

   if (KBkey == GDK_KEY_plus) m_zoom(null,"in");                           //  + key  >>  zoom in
   if (KBkey == GDK_KEY_minus) m_zoom(null,"fit");                         //  - key  >>  fit to window
   if (KBkey == GDK_KEY_equal) m_zoom(null,"in");                          //  = key: same as +
   if (KBkey == GDK_KEY_KP_Add) m_zoom(null,"in");                         //  keypad +
   if (KBkey == GDK_KEY_KP_Subtract) m_zoom(null,"fit");                   //  keypad -
   
   if (KBkey == GDK_KEY_Delete) m_trash(0,0);                              //  delete  >>  trash

   if (KBkey == GDK_KEY_F1)                                                //  F1  >>  user guide
      showz_userguide(zfuncs::F1_help_topic);                              //  show topic if there, or page 1

   if (! E1pxm) {                                                          //  if no edit in progress,
      if (KBkey == GDK_KEY_R || KBkey == GDK_KEY_r) angle = 90;            //    allow manual rotations
      if (KBkey == GDK_KEY_L || KBkey == GDK_KEY_l) angle = -90;
      if (angle) {
         mutex_lock(&Fpixmap_lock);
         pxm = PXM_rotate(Fpxm,angle);
         PXM_free(Fpxm);
         Fpxm = pxm;
         Fww = pxm->ww;
         Fhh = pxm->hh;
         mutex_unlock(&Fpixmap_lock);
         mwpaint2();
         m_save(0,"nowarn");                                               //  auto save uprighted image
      }
   }

   if (KBcontrolkey && KBkey == GDK_KEY_m) {                               //  zmalloc() activity log
      KBzmalloclog = 1 - KBzmalloclog;                                     //  Ctrl+M = toggle on/off
      if (KBzmalloclog) zmalloc_log(999999);
      else zmalloc_log(0);
      zmalloc_report();
   }
   
   if (KBkey == GDK_KEY_T || KBkey == GDK_KEY_t) m_trim(0,0);              //  Key T = Trim function
   if (KBkey == GDK_KEY_B || KBkey == GDK_KEY_b) m_tune(0,0);              //  Key B = brightness/color
   if (KBkey == GDK_KEY_N || KBkey == GDK_KEY_n) m_rename(0,0);            //  Key N = rename file
   
   if (KBkey == GDK_KEY_z || KBkey == GDK_KEY_Z) {                         //  Z key
      if (! KBcontrolkey) m_zoom(null,"100");                              //  if no CTRL, toggle zoom 1x
      else {                                                               //  if CTRL:
         if (KBshiftkey) m_redo(0,0);                                      //    with shift: redo edit
         else m_undo(0,0);                                                 //    without shift: undo edit
      }
   }
   
   KBkey = 0;
   return 1;
}


/**************************************************************************/

//  set new image zoom level or magnification
//  zoom:   "in"        zoom in in steps
//          "out"       zoom out in steps
//          "fit"       zoom out to fit window
//          "100"       toggle 100% and fit window

void m_zoom(GtkWidget *, cchar *zoom)
{
   int      iww, ihh;
   float    scalew, scaleh, fitscale, fzoom2, zratio = 1;
   
   if (E3pxm) {
      iww = E3ww;
      ihh = E3hh;
   }
   else  {
      iww = Fww;
      ihh = Fhh;
   }

   if (iww > Dww || ihh > Dhh) {                                           //  get window fit scale
      scalew = 1.0 * Dww / iww;
      scaleh = 1.0 * Dhh / ihh;
      if (scalew < scaleh) fitscale = scalew;
      else fitscale = scaleh;                                              //  window/image, < 1.0
   }
   else fitscale = 1.0;                                                    //  if image < window use 100%
   
   if (strEqu(zoom,"Zoom+")) zoom = "in";                                  //  toolbar button, + = zoom in
   if (strEqu(zoom,"Zoom-")) zoom = "fit";                                 //                  - = fit window

   if (strstr("in out",zoom))                                              //  zoom in or out
   {
      if (! Fzoom) Fzoom = fitscale;                                       //  current zoom scale        v.12.03
      for (fzoom2 = 0.125; fzoom2 < 4.0; fzoom2 *= zoomratio)              //  find nearest natural ratio
         if (Fzoom < fzoom2 * sqrt(zoomratio)) break;
      if (strEqu(zoom,"in")) zratio = zoomratio;                           //  zoom in, make image larger
      if (strEqu(zoom,"out")) zratio = 1.0 / zoomratio;                    //  zoom out, make image smaller
      Fzoom = fzoom2 * zratio;
      if (Fzoom > 0.124 && Fzoom < 0.126) Fzoom = 0.125;                   //  hit these ratios exactly
      if (Fzoom > 0.24  && Fzoom < 0.26)  Fzoom = 0.25;
      if (Fzoom > 0.49  && Fzoom < 0.51)  Fzoom = 0.50;
      if (Fzoom > 0.99  && Fzoom < 1.01)  Fzoom = 1.00;
      if (Fzoom > 1.99  && Fzoom < 2.01)  Fzoom = 2.00;
      if (Fzoom > 3.99) Fzoom = 4.0;                                       //  max. allowed zoom
      if (Fzoom < fitscale) Fzoom = 0;                                     //  image < window
   }
   
   if (strEqu(zoom,"fit")) Fzoom = 0;                                      //  zoom to fit window

   if (strEqu(zoom,"100")) {
      if (Fzoom != 0) Fzoom = 0;                                           //  toggle 100% and fit window
      else  Fzoom = 1;
   }

   if (! Fzoom) zoomx = zoomy = 0;                                         //  no requested zoom center

   mwpaint2();                                                             //  refresh window
   return;
}


/**************************************************************************/

//  draw a pixel using given color.
//  px, py are image space.

void draw_pixel(int px, int py)
{
   int               qx, qy, crflag = 0;
   static int        pqx, pqy;
   static int        ftf = 1;
   static RGB        pixel[12];                                            //  2x2 block of pixels
   static GdkPixbuf  *pixbuf1, *pixbuf4;
   
   if (ftf) {
      ftf = 0;
      pixbuf1 = gdk_pixbuf_new_from_data(pixel,GDKRGB,0,8,1,1,3,0,0);      //  1x1 pixels
      pixbuf4 = gdk_pixbuf_new_from_data(pixel,GDKRGB,0,8,2,2,6,0,0);      //  2x2 pixels
   }

   qx = Mscale * px - Iorgx;                                               //  image to window space
   qy = Mscale * py - Iorgy;

   if (qx == pqx && qy == pqy) return;                                     //  avoid redundant points

   pqx = qx;
   pqy = qy;
   
   if (qx < 0 || qx > dww-2) return;                                       //  keep off image edges
   if (qy < 0 || qy > dhh-2) return;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   if (Mscale < 1.99) {                                                    //  write 1x1 pixels
      pixel[0] = fg_color[0];
      pixel[1] = fg_color[1];
      pixel[2] = fg_color[2];

      gdk_cairo_set_source_pixbuf(mwcr,pixbuf1,qx+Dorgx,qy+Dorgy);
      cairo_paint(mwcr);
   }
   
   else {                                                                  //  write 2x2 fat pixels
      pixel[0] = pixel[3] = pixel[6] = pixel[9] = fg_color[0];
      pixel[1] = pixel[4] = pixel[7] = pixel[10] = fg_color[1];
      pixel[2] = pixel[5] = pixel[8] = pixel[11] = fg_color[2];

      gdk_cairo_set_source_pixbuf(mwcr,pixbuf4,qx+Dorgx,qy+Dorgy);
      cairo_paint(mwcr);
   }

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  erase one drawn pixel - restore from window image Wpixels.
//  px, py are image space.

void erase_pixel(int px, int py)
{
   GdkPixbuf      *pixbuf;
   RGB            *pixel;
   static int     pqx, pqy;
   int            qx, qy, crflag = 0;

   qx = Mscale * px;                                                       //  image to window space
   qy = Mscale * py;

   if (qx == pqx && qy == pqy) return;                                     //  avoid same target pixel

   pqx = qx;
   pqy = qy;
   
   if (qx < 0 || qx > iww-2) return;                                       //  pixel outside scaled image
   if (qy < 0 || qy > ihh-2) return;
   
   if (qx < Iorgx || qx > Iorgx + dww-2) return;                           //  pixel outside window
   if (qy < Iorgy || qy > Iorgy + dhh-2) return;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   pixel = Wpixels + (qy * iww + qx) * 3;                                  //  source pixels for erase
   pixbuf = gdk_pixbuf_new_from_data(pixel,GDKRGB,0,8,2,2,iww*3,0,0);      //  2x2 block
   qx = qx - Iorgx + Dorgx;                                                //  target pixel in window
   qy = qy - Iorgy + Dorgy;
   gdk_cairo_set_source_pixbuf(mwcr,pixbuf,qx,qy);
   cairo_paint(mwcr);
   g_object_unref(pixbuf);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  draw line.
//  coordinates are image space.

void draw_line(int x1, int y1, int x2, int y2)
{
   float    px1, py1, px2, py2;
   int      crflag = 0;

   px1 = Mscale * x1 - Iorgx + Dorgx;                                      //  image to window space
   py1 = Mscale * y1 - Iorgy + Dorgy;
   px2 = Mscale * x2 - Iorgx + Dorgx;
   py2 = Mscale * y2 - Iorgy + Dorgy;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);
   cairo_move_to(mwcr,px1,py1);
   cairo_line_to(mwcr,px2,py2);
   cairo_stroke(mwcr);
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  draw dotted line.
//  coordinates are image space.

void draw_dotline(int x1, int y1, int x2, int y2)
{
   float    px1, py1, px2, py2;
   double   dashes[2] = { 4, 4 };
   int      crflag = 0;
   
   px1 = Mscale * x1 - Iorgx + Dorgx;                                      //  image to window space
   py1 = Mscale * y1 - Iorgy + Dorgy;
   px2 = Mscale * x2 - Iorgx + Dorgx;
   py2 = Mscale * y2 - Iorgy + Dorgy;
   
   if (px1 > Dww-2) px1 = Dww-2;
   if (py1 > Dhh-2) py1 = Dhh-2;
   if (px2 > Dww-2) px2 = Dww-2;
   if (py2 > Dhh-2) py2 = Dhh-2;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }
   else cairo_save(mwcr);

   cairo_set_source_rgb(mwcr,1,1,1);
   cairo_set_dash(mwcr,dashes,2,0);
   cairo_move_to(mwcr,px1,py1);
   cairo_line_to(mwcr,px2,py2);
   cairo_stroke(mwcr);
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }
   else cairo_restore(mwcr);

   return;
}


//  erase line. refresh line path from Wpixels window image.
//  double line width is erased.
//  coordinates are image space.

void erase_line(int x1, int y1, int x2, int y2)
{
   float    pxm, pym, slope;
   int      crflag = 0;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   if (abs(y2 - y1) > abs(x2 - x1)) {
      slope = 1.0 * (x2 - x1) / (y2 - y1);
      for (pym = y1; pym <= y2; pym++) {
         pxm = x1 + slope * (pym - y1);
         erase_pixel(pxm,pym);
      }
   }

   else {
      slope = 1.0 * (y2 - y1) / (x2 - x1);
      for (pxm = x1; pxm <= x2; pxm++) {
         pym = y1 + slope * (pxm - x1);
         erase_pixel(pxm,pym);
      }
   }
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  draw pre-set overlay lines on top of image
//  arg = 1:   paint lines only (because window repainted)
//        2:   erase lines and forget them
//        3:   erase old lines, paint new lines, save new in old

void draw_toplines(int arg)
{
   int      ii, crflag = 0;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   if (arg == 2 || arg == 3)                                               //  erase old lines
      for (ii = 0; ii < Nptoplines; ii++)
         erase_line(ptoplinex1[ii],ptopliney1[ii],ptoplinex2[ii],ptopliney2[ii]);
   
   if (arg == 1 || arg == 3)                                               //  draw new lines
      for (ii = 0; ii < Ntoplines; ii++)
         draw_dotline(toplinex1[ii],topliney1[ii],toplinex2[ii],topliney2[ii]);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   if (arg == 2) {
      Nptoplines = Ntoplines = 0;                                          //  forget lines
      return;
   }

   for (ii = 0; ii < Ntoplines; ii++)                                      //  save for future erase
   {
      ptoplinex1[ii] = toplinex1[ii];
      ptopliney1[ii] = topliney1[ii];
      ptoplinex2[ii] = toplinex2[ii];
      ptopliney2[ii] = topliney2[ii];
   }

   Nptoplines = Ntoplines;

   return;
}


/**************************************************************************/

//  draw a grid of horizontal and vertical lines.
//  grid line spacings are in window space.

void draw_gridlines()
{
   int      px, py, stepx, stepy, startx, starty;
   int      crflag = 0;

   if (! Fpxm) return;                                                     //  no image
   if (! Fgrid) return;                                                    //  grid lines off

   stepx = gridspace[0];                                                   //  space between grid lines
   stepy = gridspace[1];                                                   //  (window space)

   if (gridcount[0]) stepx = dww / (1 + gridcount[0]);                     //  if line counts specified,
   if (gridcount[1]) stepy = dhh / ( 1 + gridcount[1]);                    //    set spacing accordingly

   startx = stepx * gridoffset[0] / 100;                                   //  variable offsets
   if (startx < 0) startx += stepx;

   starty = stepy * gridoffset[1] / 100;
   if (starty < 0) starty += stepy;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);                                       //  white lines

   if (gridon[0] && stepx)
      for (px = Dorgx+startx; px < Dorgx+dww; px += stepx) {
         cairo_move_to(mwcr,px,Dorgy);
         cairo_line_to(mwcr,px,Dorgy+dhh);
      }

   if (gridon[1] && stepy)
      for (py = Dorgy+starty; py < Dorgy+dhh; py += stepy) {
         cairo_move_to(mwcr,Dorgx,py);
         cairo_line_to(mwcr,Dorgx+dww,py);
      }

   cairo_stroke(mwcr);

   cairo_set_source_rgb(mwcr,0,0,0);                                       //  adjacent black lines

   if (gridon[0] && stepx) 
      for (px = Dorgx+startx+1; px < Dorgx+dww; px += stepx) {
         cairo_move_to(mwcr,px,Dorgy);
         cairo_line_to(mwcr,px,Dorgy+dhh);
      }

   if (gridon[1] && stepy)
      for (py = Dorgy+starty+1; py < Dorgy+dhh; py += stepy) {
         cairo_move_to(mwcr,Dorgx,py);
         cairo_line_to(mwcr,Dorgx+dww,py);
      }

   cairo_stroke(mwcr);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  maintain a set of text strings written over the image in the window.
//  add a new text string to the list.
//  multiple text strings can be added with the same ID.
//  px and py are image space.

void add_toptext(int ID, int px, int py, cchar *text, cchar *font)
{
   if (Ntoptext == maxtoptext) {
      printf("maxtoptext exceeded \n");
      return;
   }
   
   int ii = Ntoptext++;
   toptext[ii].ID = ID;
   toptext[ii].px = px;
   toptext[ii].py = py;
   toptext[ii].text = text;
   toptext[ii].font = font;
   
   return;
}


//  draw current text strings over the image in window.
//  called from mwpaint1().

void draw_toptext() 
{
   int      crflag = 0;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   for (int ii = 0; ii < Ntoptext; ii++)
      draw_text(toptext[ii].px,toptext[ii].py,toptext[ii].text,toptext[ii].font);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  delete text strings having the given ID from the list

void erase_toptext(int ID)
{
   int      ii, jj;

   for (ii = jj = 0; ii < Ntoptext; ii++)
   {
      if (toptext[ii].ID == ID) continue;
      else toptext[jj++] = toptext[ii];
   }
   
   Ntoptext = jj;
   return;
}


//  draw text on window, black on white background
//  coordinates are image space

void draw_text(int px, int py, cchar *text, cchar *font)
{
   static PangoFontDescription   *pangofont = null;
   static PangoLayout            *pangolayout = null;
   static char                   priorfont[40] = "";
   int                           ww, hh, crflag = 0;

   px = Mscale * px - Iorgx + Dorgx;                                       //  image to window space
   py = Mscale * py - Iorgy + Dorgy;

   if (strNeq(font,priorfont)) {                                           //  change font
      strncpy0(priorfont,font,40);
      if (pangofont) pango_font_description_free(pangofont);
      if (pangolayout) g_object_unref(pangolayout);
      pangofont = pango_font_description_from_string(font);                //  make pango layout for font
      pangolayout = gtk_widget_create_pango_layout(drWin,0);
      pango_layout_set_font_description(pangolayout,pangofont);
   }
   
   pango_layout_set_text(pangolayout,text,-1);                             //  add text to layout
   pango_layout_get_pixel_size(pangolayout,&ww,&hh);
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);                                       //  draw white background
   cairo_rectangle(mwcr,px,py,ww,hh);
   cairo_fill(mwcr);
   
   cairo_move_to(mwcr,px,py);                                              //  draw layout with text
   cairo_set_source_rgb(mwcr,0,0,0);
   pango_cairo_show_layout(mwcr,pangolayout);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  maintain a set of circles drawn over the image in the window
//  px, py are image space, radius is window space
//  color = 1/2/3/4 = white/black/red/green

void add_topcircle(int px, int py, int radius, int color)
{
   if (Ntopcircles == maxtopcircles) {
      printf("maxtopcircles exceeded \n");
      return;
   }
   
   int ii = Ntopcircles++;
   topcircles[ii].px = px;
   topcircles[ii].py = py;
   topcircles[ii].radius = radius;
   topcircles[ii].color = color;
   
   return;
}


//  draw current circles over the image in the window
//  called from mwpaint1()

void draw_topcircles()
{
   int      ii, px, py, rad, color;
   int      crflag = 0;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   for (ii = 0; ii < Ntopcircles; ii++)
   {
      px = topcircles[ii].px * Mscale - Iorgx + Dorgx;                     //  image to window space
      py = topcircles[ii].py * Mscale - Iorgy + Dorgy;
      rad = topcircles[ii].radius;                                         //  radius is window space
      color = topcircles[ii].color;                                        //  1/2/3 = white/black/red

      if (color == 1) cairo_set_source_rgb(mwcr,1,1,1);                    //  set foreground color
      else if (color == 2) cairo_set_source_rgb(mwcr,0,0,0);
      else if (color == 3) cairo_set_source_rgb(mwcr,1,0,0);
      else cairo_set_source_rgb(mwcr,0,1,0);

      cairo_arc(mwcr,px,py,rad,0,2*pi);                                    //  draw 360 deg. arc
      cairo_stroke(mwcr);
   }
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  erase top circles (next window repaint)

void erase_topcircles()
{
   Ntopcircles = 0;
   return;
}


/**************************************************************************/

//  Draw circle/ellipse) around the mouse pointer.
//  (prior circle/ellipse will be erased first)
//  Used for mouse/brush radius in select/paint functions.
//  cx, cy, ww, hh  are image space coordinates.
//  if cx == -1 then erase previous circle/ellipse only.

void draw_mousearc(int cx, int cy, int ww, int hh)                         //  overhauled   v.12.03
{
   GdkPixbuf      *pixbuf;
   uint8          *ppix, *pixels;
   float          a, b, a2, b2;
   float          x, y, x2, y2;
   float          r2 = 1.42;
   int            cx2, cy2, ww2, hh2;
   int            px, py, crflag = 0;
   static int     pcx = -1, pcy, pww, phh;
   static float   pscale = 0;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }
   
   if (pcx != -1 && pscale == Mscale)                                      //  update window image to erase prior arc
      mwpaint3(pcx-pww/2, pcy-phh/2, pww+1, phh+1);
   pcx = -1;

   if (cx == -1) {                                                         //  erase only requested
      if (crflag) {
         cairo_destroy(mwcr);
         mwcr = 0;
      }
      return;
   }

   pcx = cx;                                                               //  save for subsequent erase
   pcy = cy;
   pww = ww;
   phh = hh;
   pscale = Mscale;

   cx2 = cx * Mscale;                                                      //  window image space: Wpixels
   cy2 = cy * Mscale;
   ww2 = ww * Mscale;
   hh2 = hh * Mscale;

   a = 0.5 * ww2;                                                          //  ellipse constants from
   b = 0.5 * hh2;                                                          //    enclosing rectangle
   a2 = a * a;
   b2 = b * b;

   for (y = -b/r2; y < b/r2; y++)                                          //  step through y values
   {
      y2 = y * y;
      x2 = a2 * (1 - y2 / b2);
      x = sqrt(x2);                                                        //  corresp. x values, + and -
      py = y + cy2 + 0.5;
      if (py < 0 || py > ihh-1) continue;
      px = cx2 - x + 0.5;
      if (px > 0) {
         ppix = Wpixels + (py * iww + px) * 3;                             //  draw directly on window image
         ppix[0] = fg_color[0]; 
         ppix[1] = fg_color[1]; 
         ppix[2] = fg_color[2];
      }
      px = cx2 + x + 0.5;
      if (px < iww-1) {
         ppix = Wpixels + (py * iww + px) * 3;
         ppix[0] = fg_color[0]; 
         ppix[1] = fg_color[1]; 
         ppix[2] = fg_color[2];
      }
   }

   for (x = -a/r2; x < a/r2; x++)                                          //  step through x values
   {
      x2 = x * x;
      y2 = b2 * (1 - x2 / a2);
      y = sqrt(y2);                                                        //  corresp. y values, + and -
      px = cx2 + x + 0.5;
      if (px < 0 || px > iww-1) continue;
      py = cy2 - y + 0.5;
      if (py > 0) {
         ppix = Wpixels + (py * iww + px) * 3;
         ppix[0] = fg_color[0]; 
         ppix[1] = fg_color[1]; 
         ppix[2] = fg_color[2];
      }
      py = cy2 + y + 0.5;
      if (py < ihh-1) {
         ppix = Wpixels + (py * iww + px) * 3;
         ppix[0] = fg_color[0]; 
         ppix[1] = fg_color[1]; 
         ppix[2] = fg_color[2];
      }
   }

   cx2 = cx2 - ww2 / 2;
   cy2 = cy2 - hh2 / 2;
   ww2++;
   hh2++;
   if (cx2 < 0) cx2 = 0;
   if (cy2 < 0) cy2 = 0;
   if (cx2 + ww2 > iww) ww2 = iww - cx2;
   if (cy2 + hh2 > ihh) hh2 = ihh - cy2;

   if (ww2 > 0 && hh2 > 0) {                                               //  update window
      pixels = Wpixels + (cy2 * iww + cx2) * 3;
      pixbuf = gdk_pixbuf_new_from_data(pixels,GDKRGB,0,8,ww2,hh2,iww*3,0,0);
      gdk_cairo_set_source_pixbuf(mwcr,pixbuf,cx2-Iorgx+Dorgx,cy2-Iorgy+Dorgy);
      cairo_paint(mwcr);
      g_object_unref(pixbuf);
   }

   if (Fshowarea) {                                                        //  refresh select area outline
      cx = cx - ww - 1;
      cy = cy - hh - 1;
      ww = ww * 2 + 2;
      hh = hh * 2 + 2;
      sa_show_rect(cx,cy,ww,hh);
   }

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************

   spline curve setup and edit functions

   Support multiple frames with curves in parallel
   (edit curve(s) and image mask curve)

   Usage:
   Add frame widget to dialog or zdialog.
   Set up drawing in frame: 
      sd = curve_init(frame, callback_func)
   Initialize no. of curves in frame (1-10): 
      sd->Nspc = n
   Initialize vert/horz flag for curve spc: 
      sd->vert[spc] = hv
   Initialize anchor points for curve spc: 
      sd->nap[spc], sd->apx[spc][xx], sd->apy[spc][yy]
   Generate data for curve spc:  
      splcurve_generate(sd,spc)
   Curves will now be shown and edited inside the frame when window is realized.
   The callback_func(spc) will be called when curve spc is edited.
   Change curve in program: 
      set anchor points, call splcurve_generate(sd,spc)
   Get y-value (0-1) for curve spc and given x-value (0-1):
      yval = splcurve_yval(sd,spc,xval)
   If faster access to curve is needed (no interpolation):
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      yval = sd->yval[spc][kk];

***/

//  initialize for spline curve editing
//  initial anchor points are pre-loaded into spldat before window is realized

spldat * splcurve_init(GtkWidget *frame, void func(int spc))
{
   int      cc = sizeof(spldat);                                           //  allocate spc curve data area
   spldat * sd = (spldat *) zmalloc(cc,"curvedat");
   memset(sd,0,cc);

   sd->drawarea = gtk_drawing_area_new();                                  //  drawing area for curves
   gtk_container_add(GTK_CONTAINER(frame),sd->drawarea);
   sd->spcfunc = func;                                                     //  user callback function

   gtk_widget_add_events(sd->drawarea,GDK_BUTTON_PRESS_MASK);              //  connect mouse events to drawing area
   gtk_widget_add_events(sd->drawarea,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(sd->drawarea,GDK_BUTTON1_MOTION_MASK); 
   G_SIGNAL(sd->drawarea,"motion-notify-event",splcurve_adjust,sd);
   G_SIGNAL(sd->drawarea,"button-press-event",splcurve_adjust,sd);
   G_SIGNAL(sd->drawarea,"draw",splcurve_draw,sd);
   
   return sd;
}


//  modify anchor points in curve using mouse

int splcurve_adjust(void *, GdkEventButton *event, spldat *sd)
{
   int            ww, hh, kk;
   int            mx, my, button, evtype;
   static int     spc, ap, mbusy = 0, Fdrag = 0;                           //  drag continuation logic
   int            minspc, minap;
   float          mxval, myval, cxval, cyval;
   float          dist2, mindist2 = 0;
   
   mx = event->x;                                                          //  mouse position in drawing area
   my = event->y;
   evtype = event->type;
   button = event->button;

   if (evtype == GDK_MOTION_NOTIFY) {
      if (mbusy) return 0;                                                 //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (evtype == GDK_BUTTON_RELEASE) {
      Fdrag = 0;
      return 0;
   }

   ww = gtk_widget_get_allocated_width(sd->drawarea);                      //  drawing area size
   hh = gtk_widget_get_allocated_height(sd->drawarea);
   
   if (mx < 0) mx = 0;                                                     //  limit edge excursions
   if (mx > ww) mx = ww;
   if (my < 0) my = 0;
   if (my > hh) my = hh;
   
   if (evtype == GDK_BUTTON_PRESS) Fdrag = 0;                              //  left or right click
   
   if (Fdrag)                                                              //  continuation of drag
   {
      if (sd->vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (ap < sd->nap[spc]-1 && sd->apx[spc][ap+1] - mxval < 0.05)        //  disallow < 0.05 from next
            return 0;
      if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;           //    or prior anchor point 
   }
   
   else                                                                    //  mouse click or new drag begin
   {
      minspc = minap = -1;                                                 //  find closest curve/anchor point
      mindist2 = 999999;

      for (spc = 0; spc < sd->Nspc; spc++)                                 //  loop curves
      {
         if (sd->vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  loop anchor points
         {
            cxval = sd->apx[spc][ap];
            cyval = sd->apy[spc][ap];
            dist2 = (mxval-cxval)*(mxval-cxval) 
                  + (myval-cyval)*(myval-cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest anchor point
               minspc = spc;
               minap = ap;
            }
         }
      }

      if (minspc < 0) return 0;                                            //  impossible
      spc = minspc;
      ap = minap;
   }
   
   if (evtype == GDK_BUTTON_PRESS && button == 3)                          //  right click, remove anchor point
   {
      if (sqrt(mindist2) > 0.05) return 0;                                 //  not close enough
      if (sd->nap[spc] < 3) return 0;                                      //  < 2 anchor points would remain
      sd->nap[spc]--;                                                      //  decr. before loop
      for (kk = ap; kk < sd->nap[spc]; kk++) {
         if (! kk) printf("meaningless reference %d",kk);                  //  stop gcc optimization bug   /////
         sd->apx[spc][kk] = sd->apx[spc][kk+1];
         sd->apy[spc][kk] = sd->apy[spc][kk+1];
      }
      splcurve_generate(sd,spc);                                           //  regenerate data for modified curve
      gtk_widget_queue_draw(sd->drawarea);                                 //  gtk3
      sd->spcfunc(spc);                                                    //  call user function
      return 0;
   }

   if (! Fdrag)                                                            //  new drag or left click
   {
      if (sd->vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (sqrt(mindist2) < 0.05)                                           //  existing point close enough,
      {                                                                    //    move this anchor point to mouse
         if (ap < sd->nap[spc]-1 && sd->apx[spc][ap+1] - mxval < 0.05)
               return 0;                                                   //  disallow < 0.05 from next
         if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;        //    or prior anchor point 
      }
   
      else                                                                 //  none close, add new anchor point
      {
         minspc = -1;                                                      //  find closest curve to mouse
         mindist2 = 999999;

         for (spc = 0; spc < sd->Nspc; spc++)                              //  loop curves
         {
            if (sd->vert[spc]) {
               mxval = 1.0 * my / hh;                                      //  mouse position in curve space
               myval = 1.0 * mx / ww;
            }
            else {
               mxval = 1.0 * mx / ww;
               myval = 1.0 * (hh - my) / hh;
            }

            cyval = splcurve_yval(sd,spc,mxval);
            dist2 = fabsf(myval - cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest curve
               minspc = spc;
            }
         }

         if (minspc < 0) return 0;                                         //  impossible
         if (mindist2 > 0.05) return 0;                                    //  not close enough to any curve
         spc = minspc;

         if (sd->nap[spc] > 49) {
            zmessageACK(mWin,ZTX("Exceed 50 anchor points"));
            return 0;
         }

         if (sd->vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  find anchor point with next higher x
            if (mxval <= sd->apx[spc][ap]) break;                          //    (ap may come out 0 or nap)

         if (ap < sd->nap[spc] && sd->apx[spc][ap] - mxval < 0.05)         //  disallow < 0.05 from next
               return 0;                                                   //    or prior anchor point
         if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;

         for (kk = sd->nap[spc]; kk > ap; kk--) {                          //  make hole for new point
            sd->apx[spc][kk] = sd->apx[spc][kk-1];
            sd->apy[spc][kk] = sd->apy[spc][kk-1];
         }

         sd->nap[spc]++;                                                   //  up point count
      }
   }

   sd->apx[spc][ap] = mxval;                                               //  new or moved anchor point
   sd->apy[spc][ap] = myval;                                               //    at mouse position

   splcurve_generate(sd,spc);                                              //  regenerate data for modified curve
   gtk_widget_queue_draw(sd->drawarea);                                    //  gtk3
   sd->spcfunc(spc);                                                       //  call user function
   
   if (evtype == GDK_MOTION_NOTIFY) Fdrag = 1;                             //  remember drag is underway
   return 0;
}


//  for expose event or when a curve is changed
//  draw all curves based on current anchor points

int splcurve_draw(GtkWidget *drawarea, cairo_t *cr, spldat *sd)
{
   int         ww, hh, spc, ap;
   float       xval, yval, px, py, qx, qy;
   
   ww = gtk_widget_get_allocated_width(sd->drawarea);                      //  drawing area size
   hh = gtk_widget_get_allocated_height(sd->drawarea);
   if (ww < 50 || hh < 50) return 0;

   cairo_set_line_width(cr,1);
   cairo_set_source_rgb(cr,0.7,0.7,0.7);

   for (int ii = 0; ii < sd->Nscale; ii++)                                 //  draw y-scale lines if any
   {
      px = ww * sd->xscale[0][ii];
      py = hh - hh * sd->yscale[0][ii];
      qx = ww * sd->xscale[1][ii];
      qy = hh - hh * sd->yscale[1][ii];
      cairo_move_to(cr,px,py);
      cairo_line_to(cr,qx,qy);
   }
   cairo_stroke(cr);

   cairo_set_source_rgb(cr,0,0,0);

   for (spc = 0; spc < sd->Nspc; spc++)                                    //  loop all curves
   {
      if (sd->vert[spc])                                                   //  vert. curve
      {
         for (py = 0; py < hh; py++)                                       //  generate all points for curve
         {
            xval = 1.0 * py / hh;
            yval = splcurve_yval(sd,spc,xval);
            px = ww * yval;
            if (py == 0) cairo_move_to(cr,px,py);
            cairo_line_to(cr,px,py);
         }
         cairo_stroke(cr);
         
         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  draw boxes at anchor points
         {
            xval = sd->apx[spc][ap];
            yval = sd->apy[spc][ap];
            px = ww * yval;
            py = hh * xval;
            cairo_rectangle(cr,px-2,py-2,4,4);
         }
         cairo_fill(cr);
      }
      else                                                                 //  horz. curve
      {
         for (px = 0; px < ww; px++)                                       //  generate all points for curve
         {
            xval = 1.0 * px / ww;
            yval = splcurve_yval(sd,spc,xval);
            py = hh - hh * yval;
            if (px == 0) cairo_move_to(cr,px,py);
            cairo_line_to(cr,px,py);
         }
         cairo_stroke(cr);
         
         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  draw boxes at anchor points
         {
            xval = sd->apx[spc][ap];
            yval = sd->apy[spc][ap];
            px = ww * xval;
            py = hh - hh * yval;
            cairo_rectangle(cr,px-2,py-2,4,4);
         }
         cairo_fill(cr);
      }
   }

   return 0;
}


//  generate all curve data points when anchor points are modified

int splcurve_generate(spldat *sd, int spc)
{
   int      kk, kklo, kkhi;
   float    xval, yvalx;

   spline1(sd->nap[spc],sd->apx[spc],sd->apy[spc]);                        //  compute curve fitting anchor points

   kklo = 1000 * sd->apx[spc][0] - 30;                                     //  xval range = anchor point range
   if (kklo < 0) kklo = 0;                                                 //    + 0.03 extra below/above
   kkhi = 1000 * sd->apx[spc][sd->nap[spc]-1] + 30;
   if (kkhi > 1000) kkhi = 1000;

   for (kk = 0; kk < 1000; kk++)                                           //  generate all points for curve
   {
      xval = 0.001 * kk;                                                   //  remove anchor point limits
      yvalx = spline2(xval);
      if (yvalx < 0) yvalx = 0;                                            //  yval < 0 not allowed, > 1 OK
      sd->yval[spc][kk] = yvalx;
   }

   return 0;
}


//  Retrieve curve data using interpolation of saved table of values

float splcurve_yval(spldat *sd, int spc, float xval)
{
   int      ii;
   float    x1, x2, y1, y2, y3;
   
   if (xval <= 0) return sd->yval[spc][0];
   if (xval >= 0.999) return sd->yval[spc][999];
   
   x2 = 1000.0 * xval;
   ii = x2;
   x1 = ii;
   y1 = sd->yval[spc][ii];
   y2 = sd->yval[spc][ii+1];
   y3 = y1 + (y2 - y1) * (x2 - x1);
   return y3;
}


//  load curve data from a file
//  returns 0 if fail (invalid file data), sd not modified
//  returns 1 if succcess, sd is initialized from file data

int splcurve_load(spldat *sd)
{
   char        *pfile;
   int         nn, ii, jj;
   FILE        *fid = 0;
   int         Nspc, vert[10], nap[10];
   float       apx[10][50], apy[10][50];

   zfuncs::F1_help_topic = "curve_edit";

   pfile = zgetfile1(ZTX("load curve from a file"),"open",saved_curves_dirk);
   if (! pfile) return 0;

   fid = fopen(pfile,"r");
   zfree(pfile);
   if (! fid) goto fail;
   
   nn = fscanf(fid,"%d ",&Nspc);                                           //  no. of curves
   if (nn != 1) goto fail;
   if (Nspc < 1 || Nspc > 10) goto fail;
   if (Nspc != sd->Nspc) goto fail2;

   for (ii = 0; ii < Nspc; ii++)                                           //  loop each curve
   {
      nn = fscanf(fid,"%d  %d ",&vert[ii],&nap[ii]);                       //  vertical flag, no. anchor points
      if (nn != 2) goto fail;
      if (vert[ii] < 0 || vert[ii] > 1) goto fail;
      if (nap[ii] < 2 || nap[ii] > 50) goto fail;

      for (jj = 0; jj < nap[ii]; jj++)                                     //  anchor point values 
      {
         nn = fscanf(fid,"%f/%f  ",&apx[ii][jj],&apy[ii][jj]);
         if (nn != 2) goto fail;
         if (apx[ii][jj] < 0 || apx[ii][jj] > 1) goto fail;
         if (apy[ii][jj] < 0 || apy[ii][jj] > 1) goto fail;
      }
   }
   
   fclose(fid);
   
   sd->Nspc = Nspc;                                                        //  copy curve data to caller's arg

   for (ii = 0; ii < Nspc; ii++)
   {
      sd->vert[ii] = vert[ii];
      sd->nap[ii] = nap[ii];
      
      for (jj = 0; jj < nap[ii]; jj++)
      {
         sd->apx[ii][jj] = apx[ii][jj];
         sd->apy[ii][jj] = apy[ii][jj];
      }
   }

   for (ii = 0; ii < Nspc; ii++)                                           //  generate curve data from anchor points
      splcurve_generate(sd,ii);
   
   if (sd->drawarea)                                                       //  regen and redraw all curves
      gtk_widget_queue_draw(sd->drawarea);                                 //  gtk3

   return 1;
   
fail:
   if (fid) fclose(fid);
   zmessageACK(mWin,ZTX("curve file is invalid"));
   return 0;

fail2:
   if (fid) fclose(fid);
   zmessageACK(mWin,ZTX("curve file has different no. of curves"));
   return 0;
}


//  save curve data to a file

int splcurve_save(spldat *sd) 
{
   char        *pfile, *pp;
   int         ii, jj;
   FILE        *fid;

   zfuncs::F1_help_topic = "curve_edit";

   pp = zgetfile1(ZTX("save curve to a file"),"save",saved_curves_dirk);
   if (! pp) return 0;

   pfile = strdupz(pp,8);
   zfree(pp);

   pp = strrchr(pfile,'/');                                                //  force .curve extension
   if (pp) pp = strrchr(pp,'.');
   if (pp) strcpy(pp,".curve");
   else strcat(pfile,".curve");

   fid = fopen(pfile,"w");
   zfree(pfile);
   if (! fid) return 0;
   
   fprintf(fid,"%d \n",sd->Nspc);                                          //  no. of curves

   for (ii = 0; ii < sd->Nspc; ii++)                                       //  loop each curve
   {
      fprintf(fid,"%d  %d \n",sd->vert[ii],sd->nap[ii]);                   //  vertical flag, no. anchor points
      for (jj = 0; jj < sd->nap[ii]; jj++)                                 //  anchor point values 
         fprintf(fid,"%.4f/%.4f  ",sd->apx[ii][jj],sd->apy[ii][jj]);
      fprintf(fid,"\n");
   }
   
   fclose(fid);
   return 0;
}


/**************************************************************************

   edit transaction and thread support functions

   edit transaction management
      edit_setup()                     start new edit - copy E3 > E1
      edit_cancel()                    cancel edit - E1 > E3, delete E1
      edit_done()                      commit edit - add to undo stack
      edit_undo()                      undo edit - E1 > E3
      edit_redo()                      redo edit - run thread again
      edit_fullsize()                  convert preview to full-size pixmaps

   main level thread management
      start_thread(func,arg)           start thread running
      signal_thread()                  signal thread that work is pending
      wait_thread_idle()               wait for pending work complete
      wrapup_thread(command)           wait for exit or command thread exit

   thread function
      thread_idle_loop()               wait for pending work, exit if commanded
      thread_exit()                    exit thread unconditionally

   thread_status (thread ownership
      0     no thread is running
      1     thread is running and idle (no work)
      2     thread is working
      0     thread has exited

   thread_command (main program ownership)
      0     idle, no work pending
      8     exit when pending work is done
      9     exit now, unconditionally

   thread_pend       work requested counter
   thread_done       work done counter
   thread_hiwater    high water mark
   edit_action       done/cancel/undo/redo in progress

***************************************************************************/

int      thread_command = 0, thread_status = 0;
int      thread_pend = 0, thread_done = 0, thread_hiwater = 0;
int      edit_action = 0;


/**************************************************************************

  Setup for a new edit transaction
  Create E1 (edit input) and E3 (edit output) pixmaps from
  previous edit output (Fpxm) or image file (new Fpxm).

  Fprev      0     edit full-size image
             1     edit preview image unless select area exists

  Farea      0     select_area is invalid and will be deleted (e.g. rotate)
             1     select_area not used but remains valid (e.g. red-eye)
             2     select_area can be used and remains valid (e.g. flatten)

***************************************************************************/

int edit_setup(editfunc &EF)                                               //  support parallel edits
{
   int      yn;

   if (! curr_file) return 0;                                              //  no image file
   if (is_syncbusy()) return 0;                                            //  must wait for file sync 
   
   if (CEF) {                                                              //  prior edit function active
      if (&EF == CEF) return 0;                                            //  2nd instance of one function
      if (EF.Fpara && CEF->Fpara)                                          //  check if parallel edits are OK
         edit_suspend();                                                   //  suspend prior edit
      else {
         zmessageACK(mWin,ZTX("cannot parallel edit"));
         return 0;
      }
   }
   else {                                                                  //  if this is 1st active edit
      if (! menulock(1)) return 0;                                         //    test menu lock is allowed
      menulock(0);                                                         //      but don't lock yet
   }

   if (! Fexiftool && ! Fexifwarned) {
      zmessageACK(mWin,ZTX("exiftool is not installed \n"                  //  warn if starting to edit
                           "edited images will lose EXIF data"));          //    and exiftool is missing
      Fexifwarned = 1;
   }

   if (Pundo > maxedits-2) {                                               //  undo capacity reached
      zmessageACK(mWin,ZTX("Too many edits, please save image"));
      return 0;
   }

   if (EF.Farea == 0 && sa_stat) {                                         //  select area will be lost, warn user
      yn = zmessageYN(mWin,ZTX("Select area cannot be kept.\n"
                               "Continue?"));
      if (! yn) return 0;
      sa_unselect();                                                       //  unselect area
      zdialog_free(zdsela);
   }
   
   if (EF.Farea == 2 && sa_stat && sa_stat != 3) {                         //  select area exists and can be used,
      yn = zmessageYN(mWin,ZTX("Select area not active.\n"                 //    but not active, ask user
                               "Continue?"));
      if (! yn) return 0;
   }

   if (! menulock(1)) return 0;                                            //  can now commit, lock menu
   
   Fpreview = 0;                                                           //  use preview image if supported
   if (EF.Fprev && ! (EF.Farea == 2 && sa_stat == 3)) {                    //    and select area will not be used
      Fpreview = 1;                                                        //  use preview image (smaller)
      sa_show(0);                                                          //  hide area if present
      if (Fzoom) {
         Fzoom = 0;                                                        //  bugfix, mwpaint() req.
         gtk_widget_queue_draw(drWin);                                     //  gtk3
      }
   }

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps

   if (! Ffixblue) PXM_fixblue(Fpxm);                                      //  blue=0 >> blue=1 for vpixel()
   Ffixblue = 1;                                                           //  only for fresh image      v.12.03

   if (Fpreview)                                                           //  edit pixmaps are window-size
      E1pxm = PXM_rescale(Fpxm,dww,dhh);                                   //  E1pxm = Fpxm scaled to window
   else E1pxm = PXM_copy(Fpxm);                                            //  edit pixmaps are full-size
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   
   E1ww = E3ww = E1pxm->ww;                                                //  E1 and E3 pixmap dimensions
   E1hh = E3hh = E1pxm->hh;

   mutex_unlock(&Fpixmap_lock);
   
   if (Pundo == 0) save_undo("initial");                                   //  initial image >> undo stack

   CEF = &EF;                                                              //  set current edit function
   CEF->Fmod = 0;                                                          //  image not modified yet

   thread_command = thread_status = 0;                                     //  no thread running
   thread_pend = thread_done = thread_hiwater = 0;                         //  no work pending or done
   if (EF.threadfunc) start_thread(EF.threadfunc,0);                       //  start edit thread
   
   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(drWin);
   zmainloop();                                                            //  must wait for completion

   return 1;
}


/**************************************************************************/

//  suspend a parallel edit function so it can be resumed
//  edit dialog and curve edit are left active
//  logically the equivalent of edit_done()

void edit_suspend()
{
   wait_thread_idle();                                                     //  wait for thread to finish
   
   if (Fpreview && CEF->Fmod) {                                            //  preview image was edited
      Fzoom = 0;
      edit_fullsize();                                                     //  update full image
   }

   wrapup_thread(8);                                                       //  tell thread to finish and exit

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps

   if (CEF->Fmod) {                                                        //  image was modified
      Fmodified++;                                                         //  overall mod counter
      PXM_free(Fpxm);
      Fpxm = E3pxm;                                                        //  E3pxm >> Fpxm
      E3pxm = 0;
      Fww = Fpxm->ww;
      Fhh = Fpxm->hh;
      Pundo++;                                                             //  save next undo state
      Pumax = Pundo;
      save_undo(CEF->funcname);
   }

   PXM_free(E1pxm);                                                        //  free edit pixmaps
   PXM_free(E3pxm);
   PXM_free(ERpxm);                                                        //  free redo copy
   E1ww = E3ww = ERww = 0;

   mutex_unlock(&Fpixmap_lock);
   
   CEF = 0;                                                                //  no current edit func
   menulock(0);
   mwpaint2();
   return;
}


/**************************************************************************/

//  process edit cancel

void edit_cancel(editfunc &EF)
{
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry
   
   if (&EF == CEF)                                                         //  current edit is canceled
      wrapup_thread(9);                                                    //  tell thread to quit, wait
   
   if (EF.mousefunc == mouseCBfunc)                                        //  if my mouse, free mouse
      freeMouse();
   if (EF.zd) zdialog_free(EF.zd);                                         //  kill dialog
   if (EF.curves) zfree(EF.curves);                                        //  free curves data
   EF.zd = 0;
   EF.curves = 0;

   if (&EF != CEF) {                                                       //  inactive edit canceled
      edit_action = 0;
      return;
   }
   
   mutex_lock(&Fpixmap_lock);
   PXM_free(E1pxm);                                                        //  free edit pixmaps E1, E3
   PXM_free(E3pxm);
   PXM_free(ERpxm);                                                        //  free redo copy 
   E1ww = E3ww = ERww = 0;
   mutex_unlock(&Fpixmap_lock);
   
   Fzoom = 0;
   CEF = 0;                                                                //  no current edit func
   Fpreview = 0;                                                           //  no preview mode
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  refresh window
   edit_action = 0;
   return;
}   


/**************************************************************************/

//  process edit dialog [done]  
//  E3pxm >> Fpxm

void edit_done(editfunc &EF)
{
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   if (&EF == CEF) {                                                       //  current edit is done
      wait_thread_idle();                                                  //  wait for thread to finish
      if (Fpreview && CEF->Fmod) {                                         //  preview image was edited
         Fzoom = 0;
         edit_fullsize();                                                  //  update full image
      }
      wrapup_thread(8);                                                    //  tell thread to finish and exit
   }

   if (EF.mousefunc == mouseCBfunc)                                        //  if my mouse, free mouse
      freeMouse();
   if (EF.zd) zdialog_free(EF.zd);                                         //  kill dialog
   if (EF.curves) zfree(EF.curves);                                        //  free curves data
   EF.zd = 0;
   EF.curves = 0;

   if (&EF != CEF) {                                                       //  inactive edit is done 
      edit_action = 0;
      return;
   }

   mutex_lock(&Fpixmap_lock);

   if (CEF->Fmod) {                                                        //  image was modified
      Fmodified++;                                                         //  overall mod counter
      PXM_free(Fpxm);
      Fpxm = E3pxm;                                                        //  E3pxm >> Fpxm
      E3pxm = 0;
      Fww = Fpxm->ww;
      Fhh = Fpxm->hh;
      Pundo++;
      Pumax = Pundo;
      save_undo(CEF->funcname);                                            //  save next undo state
   }

   PXM_free(E1pxm);                                                        //  free edit pixmaps
   PXM_free(E3pxm);
   PXM_free(ERpxm);                                                        //  free redo copy
   E1ww = E3ww = ERww = 0;

   mutex_unlock(&Fpixmap_lock);
   
   CEF = 0;                                                                //  no current edit func
   Fpreview = 0;                                                           //  no preview mode
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  update window
   edit_action = 0;
   return;
}


/**************************************************************************/

//  Setup for edit dialog event or curve edit.
//  Call this function before performing any image edit
//  (switch current edit function to caller's edit function).

void edit_takeover(editfunc &EF)
{
   static int  busy = 0;
   
   if (busy++) return;                                                     //  KB key bounce
   if (CEF && &EF != CEF) {                                                //  edit function changed
      edit_setup(EF);                                                      //  close old and setup new edit
      if (EF.zd) zdialog_send_event(EF.zd,"reset");                        //  reset dialog controls
      freeMouse();                                                         //  free mouse from old dialog 
   }
   busy = 0;
   return;
}


/**************************************************************************/

//  edit undo, redo, reset functions
//  these apply within an active edit function

void edit_undo()
{
   if (thread_status == 2) return;                                         //  thread busy
   if (! CEF->Fmod) return;                                                //  not modified
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry
   
   mutex_lock(&Fpixmap_lock);
   PXM_free(ERpxm);                                                        //  E3 >> redo copy
   ERpxm = E3pxm;
   ERww = E3ww;
   ERhh = E3hh;
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   E3ww = E1ww;
   E3hh = E1hh;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 0;                                                          //  reset image modified status
   mwpaint2();                                                             //  refresh window
   edit_action = 0;
   return;
}


void edit_redo()
{
   if (thread_status == 2) return;                                         //  thread busy
   if (! ERpxm) return;                                                    //  no prior undo
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  redo copy >> E3
   E3pxm = ERpxm;
   E3ww = ERww;
   E3hh = ERhh;
   ERpxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 1;                                                          //  image modified
   Fmodified++;                                                            //  overall mod counter
   mwpaint2();
   edit_action = 0;
   return;
}


void edit_reset()
{
   if (thread_status == 2) return;                                         //  thread busy
   if (! CEF->Fmod) return;                                                //  not modified
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   mutex_lock(&Fpixmap_lock);
   PXM_free(ERpxm);                                                        //  delete redo copy
   PXM_free(E3pxm);
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   E3ww = E1ww;
   E3hh = E1hh;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 0;                                                          //  reset image modified status
   mwpaint2();                                                             //  refresh window
   edit_action = 0;
   return;
}


void edit_zapredo()
{
   PXM_free(ERpxm);                                                        //  no redo copy
   return;
}


/**************************************************************************/

//  Convert from preview mode (window-size pixmaps) to full-size pixmaps.
//  Can only be used when edit function is in an edit thread with idle_loop
//  This function is called directly from some edit functions

void edit_fullsize()
{
   mutex_lock(&Fpixmap_lock);
   PXM_free(E1pxm);                                                        //  free preview pixmaps
   PXM_free(E3pxm);
   E1pxm = PXM_copy(Fpxm);                                                 //  make full-size pixmaps
   E3pxm = PXM_copy(Fpxm);
   E1ww = E3ww = Fww;
   E1hh = E3hh = Fhh;
   mutex_unlock(&Fpixmap_lock);

   Fpreview = 0;
   Fzoom = 0;

   if (! CEF->Fmod) return;                                                //  no change
   signal_thread();                                                        //  signal thread, repeat edit
   wait_thread_idle();
   return;
}


/**************************************************************************
      undo / redo toolbar buttons
***************************************************************************/

//  [undo] menu function - reinstate previous edit in undo/redo stack

void m_undo(GtkWidget *, cchar *)
{
   if (CEF) {                                                              //  undo active edit
      edit_undo();
      return;
   }

   if (Pundo == 0) return;                                                 //  undo past edit
   if (! menulock(1)) return;
   Pundo--;
   if (KB_A_key) Pundo = 0;                                                //  if KB 'A' key, undo all
   load_undo();
   Fmodified = Pundo;
   menulock(0);
   return;
}


//  [redo] menu function - reinstate next edit in undo/redo stack

void m_redo(GtkWidget *, cchar *)
{
   if (CEF) {                                                              //  redo active edit
      edit_redo();
      return;
   }

   if (Pundo == Pumax) return;
   if (! menulock(1)) return;
   Pundo++;
   if (KB_A_key) Pundo = Pumax;                                            //  if KB 'A' key, redo all
   load_undo();
   Fmodified = Pundo;
   menulock(0);
   return;
}


//  undo all edits of the current image
//  (discard modifications)

void undo_all() 
{
   if (! menulock(1)) return;
   Pundo = 0;
   load_undo();
   Fmodified = 0;
   menulock(0);
   return;
}


//  Save Fpxm to undo/redo file stack
//  stack position = Pundo

void save_undo(cchar *funcname)
{
   char     *pp, buff[24];
   int      fid, cc, cc2;
   
   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_WRONLY|O_CREAT|O_TRUNC,0640);
   if (! fid) zappcrash("undo/redo stack corrupted 2");

   snprintf(buff,24," %05d %05d fotoxx ",Fww,Fhh);
   cc = write(fid,buff,20);
   if (cc != 20) zappcrash("undo/redo stack corrupted 3");
   
   cc = Fww * Fhh * 12;
   cc2 = write(fid,Fpxm->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 4");

   close(fid);

   pvlist_replace(editlog,Pundo,funcname);                                 //  save log of edits done
   return;
}


//  Load Fpxm from undo/redo file stack
//  stack position = Pundo

void load_undo()
{
   char     *pp, buff[24], fotoxx[8];
   int      fid, ww, hh, cc, cc2;

   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_RDONLY);
   if (! fid) zappcrash("undo/redo stack corrupted 2");
   
   *fotoxx = 0;
   cc = read(fid,buff,20);
   sscanf(buff," %d %d %8s ",&ww, &hh, fotoxx);
   if (! strEqu(fotoxx,"fotoxx")) zappcrash("undo/redo stack corrupted 4");

   mutex_lock(&Fpixmap_lock);

   PXM_free(Fpxm);
   Fpxm = PXM_make(ww,hh);
   cc = ww * hh * 12;
   cc2 = read(fid,Fpxm->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 5");
   close(fid);
   
   Fww = ww;
   Fhh = hh;
   
   mutex_unlock(&Fpixmap_lock);

   if (sa_fww != Fww || sa_fhh != Fhh)                                     //  disable area if image size changes
      sa_disable();

   mwpaint2();
   return;
}


/**************************************************************************/

//  ask user if modified image should be kept or discarded
//  returns: 0 = mods discarded,  1 = do not discard mods

int mod_keep()
{
   int      keep = 0, choice = 0;
   cchar    *title = ZTX("Discard edits?");
   cchar    *message = ZTX("This action will discard current edits.\n"     //  more clarity
                           "Continue to discard edits.\n"
                           "Go Back to keep edits.");
   cchar    *continu = ZTX("Continue");
   cchar    *goback = ZTX("Go Back");

   if (Fsaved < Pundo) keep = 1;                                           //  curr. edits not saved
   if (keep == 0) return 0;                                                //  no mods
   choice = zdialog_choose(title,mWin,message,continu,goback,null);
   if (choice == 2) return 1;                                              //  keep mods
   undo_all();                                                             //  discard mods
   return 0;
}


/**************************************************************************/

//  menu lock/unlock - some functions must not run concurrently
//  returns 1 if success, else 0

int menulock(int lock)
{
   if (! lock && ! Fmenulock) zappcrash("menu lock error");
   if (lock && Fmenulock) {
      zmessageACK(mWin,ZTX("prior function still active"));
      return 0;
   }
   if (lock) Fmenulock++;
   else Fmenulock--;
   return 1;
}


/**************************************************************************/

//  check if Fsyncbusy is active, return 0 if not,
//  otherwise message user and return 1

int is_syncbusy()
{
   if (! Fsyncbusy) return 0;
   threadmessage = "Synchronize files is running. Edits are blocked. \n"
                   "You can still navigate and view images normally.";
   return 1;
}


/**************************************************************************/

//  zdialog mouse capture and release

void takeMouse(zdialog *zd, CBfunc func, GdkCursor *cursor)                //  capture mouse for dialog
{
   freeMouse();
   mouseCBfunc = func;
   Mcapture = 1;
   gdk_window_set_cursor(gdkWin,cursor);
   return;
}

void freeMouse()                                                           //  free mouse for main window
{
   if (! Mcapture) return;
   mouseCBfunc = 0;
   Mcapture = 0;
   draw_mousearc(-1,0,0,0);                                                //  remove mouse circle
   gdk_window_set_cursor(gdkWin,0);                                        //  set normal cursor
   return;
}


/**************************************************************************
   functions to manage working threads
***************************************************************************/

int      threadbusy = 0;

//  start thread that does the edit work

void start_thread(threadfunc func, void *arg)
{
   thread_status = 1;                                                      //  thread is running
   thread_command = thread_pend = thread_done = thread_hiwater = 0;        //  nothing pending
   start_detached_thread(func,arg);
   return;
}


//  signal thread that work is pending

void signal_thread()
{
   edit_zapredo();                                                         //  reset redo copy
   if (thread_status > 0) thread_pend++;
   return;
}


//  wait for edit thread to complete pending work and become idle

void wait_thread_idle()
{
   while (thread_status && thread_pend > thread_done)
   {
      zmainloop();
      zsleep(0.01);
   }
   
   return;
}


//  wait for thread exit or command thread exit
//  command = 0    wait for normal completion
//            8    finish pending work and exit
//            9    quit, exit now

void wrapup_thread(int command)
{
   thread_command = command;                                               //  tell thread to quit or finish

   while (thread_status > 0)                                               //  wait for thread to finish
   {                                                                       //    pending work and exit
      zmainloop();
      zsleep(0.01);
   }

   return;
}


//  called only from edit threads
//  idle loop - wait for work request or exit command

void thread_idle_loop()
{
   if (thread_status == 2) threadbusy--;
   thread_status = 1;                                                      //  status = idle
   thread_done = thread_hiwater;                                           //  work done = high-water mark

   while (true)
   {
      if (thread_command == 9) thread_exit();                              //  quit now command
      if (thread_command == 8)                                             //  finish work and exit
         if (thread_pend <= thread_done) thread_exit();
      if (thread_pend > thread_done) break;                                //  wait for work request
      zsleep(0.01);
   }
   
   thread_hiwater = thread_pend;                                           //  set high-water mark
   thread_status = 2;                                                      //  thread is working
   threadbusy++;
   return;                                                                 //  perform edit
}


//  exit thread unconditionally, called only from edit threads

void thread_exit()
{
   if (thread_status == 2) threadbusy--;
   thread_pend = thread_done = thread_hiwater = 0;
   thread_status = 0;
   pthread_exit(0);                                                        //  "return" cannot be used here
}


/**************************************************************************/

//  edit support functions for working threads (per processor core)

void start_wthread(threadfunc func, void *arg)                             //  start and increment busy count
{
   zadd_locked(wthreads_busy,+1);
   zadd_locked(threadbusy,+1);
   start_detached_thread(func,arg);
   return;
}


void exit_wthread()                                                        //  decrement busy count and exit
{
   zadd_locked(threadbusy,-1);
   zadd_locked(wthreads_busy,-1);
   pthread_exit(0);                                                        //  "return" cannot be used here
}


void wait_wthreads()                                                       //  wait for all working threads done
{                                                                          //  caller can be main() or a thread
   while (wthreads_busy) {
      zsleep(0.01);
      zmainloop();
   }
   return;
}


/**************************************************************************
      other support functions
***************************************************************************/

//  help menu function

void m_help(GtkWidget *, cchar *menu)
{
   if (strEqu(menu,ZTX("About"))) 
      zmessageACK(mWin,"%s \n%s \n%s \n%s \n%s \n%s",
        fotoxxvers,flicense,fhomepage,fcontact,fcredits,ftranslators);
      
   if (strEqu(menu,ZTX("User Guide"))) 
      showz_userguide();
   
   if (strEqu(menu,ZTX("User Guide Changes")))
      showz_userguide("changes");

   if (strEqu(menu,"README"))
      showz_doctext("README");

   if (strEqu(menu,ZTX("Edit Functions Summary")))
      showz_doctext("edit-menus");

   if (strEqu(menu,ZTX("Keyboard Shortcuts")))
      showz_doctext("KB-shortcuts");

   if (strEqu(menu,ZTX("Change Log")))
      showz_doctext("changelog");

   if (strEqu(menu,ZTX("Translations")))
      showz_doctext("translations");
      
   if (strEqu(menu,ZTX("Home Page")))
      showz_html(fhomepage);

   if (strEqu(menu,ZTX("Help")))                                           //  toolbar button
      showz_userguide(zfuncs::F1_help_topic);                              //  show topic if there, or page 1

   return;
}


/**************************************************************************/

//  table for loading and saving adjustable parameters between sessions

typedef struct {
   char     name[20];
   char     type[12];
   int      count;
   void     *location;
}  param;

#define Nparms 34

param paramTab[Nparms] = {
//     name                 type        count      location
{     "fotoxx version",    "char",        1,       &pversion               },
{     "last session",      "char",        1,       &last_session           },
{     "first time",        "int",         1,       &Ffirsttime             },
{     "new files found",   "int",         1,       &newfiles               },
{     "window geometry",   "int",         4,       &mwgeom                 },
{     "thumbnail size",    "int",         1,       &navi::thumbsize        },
{     "toolbar style",     "char",        1,       &tbar_style             },
{     "warn overwrite",    "int",         1,       &Fwarnoverwrite         },
{     "zoom count",        "int",         1,       &zoomcount              },
{     "startup display",   "char",        1,       &startdisplay           },
{     "start image file",  "char",        1,       &startfile              },
{     "start directory",   "char",        1,       &startdirk              },
{     "lens mm",           "float",       1,       &lens_mm                },
{     "lens bow",          "float",       1,       &lens_bow               },
{     "trim size",         "int",         2,       &trimsize               },
{     "trim buttons",      "char",        6,       &trimbuttons            },
{     "trim ratios",       "char",        6,       &trimratios             },
{     "edit resize",       "int",         2,       &editresize             },
{     "batch resize",      "int",         2,       &batchresize            },
{     "annotate font",     "char",        1,       &annotate_font          },
{     "annotate text",     "char",        1,       &annotate_text          },
{     "annotate color",    "char",        3,       &annotate_color         },
{     "annotate trans",    "int",         3,       &annotate_trans         },
{     "annotate outline",  "int",         1,       &annotate_outline       },
{     "annotate angle",    "float",       1,       &annotate_angle         },
{     "grid spacing",      "int",         2,       &gridspace              },
{     "grid counts",       "int",         2,       &gridcount              },
{     "rotate grid",       "int",         8,       &rotate_grid            },
{     "unbend grid",       "int",         8,       &unbend_grid            },
{     "warp-curved grid",  "int",         8,       &warpC_grid             },
{     "warp-linear grid",  "int",         8,       &warpL_grid             },
{     "warp-affine grid",  "int",         8,       &warpF_grid             },
{     "slideshow trans",   "int",         SSNF,    &ss_funcs               },
{     "slideshow music",   "char",        1,       &ss_musicfile           }   };


//  save parameters to file /home/<user>/.fotoxx/parameters

void save_params()
{
   FILE        *fid;
   char        buff[1050], text[1000];                                     //  limit for character data cc
   char        *pp, *name, *type;
   int         count;
   void        *location;
   char        **charloc;
   int         *intloc;
   float       *floatloc;
   time_t      currtime;
   
   pversion = strdupz(fotoxxvers);                                         //  this fotoxx version
   
   currtime = time(0);
   last_session = ctime(&currtime) + 4;                                    //  fotoxx session exit time 
   if (last_session[4] == ' ') last_session[4] = '0';                      //  format  mmm dd hh:mm:ss yyyy
   pp = strchr(last_session,'\n');
   if (pp) *pp = 0;

   snprintf(buff,199,"%s/parameters",get_zuserdir());                      //  open output file
   fid = fopen(buff,"w");
   if (! fid) return;
   
   for (int ii = 0; ii < Nparms; ii++)                                     //  write table of state data
   {
      name = paramTab[ii].name;
      type = paramTab[ii].type;
      count = paramTab[ii].count;
      location = paramTab[ii].location;
      charloc = (char **) location;
      intloc = (int *) location;
      floatloc = (float *) location;

      fprintf(fid,"%-20s  %-8s  %02d  ",name,type,count);                  //  write parm name, type, count

      for (int kk = 0; kk < count; kk++)                                   //  write "value" "value" ...
      {
         if (strEqu(type,"char")) {
            if (! *charloc) break;                                         //  missing, discontinue this parameter
            repl_1str(*charloc++,text,"\n","\\n");                         //  replace newlines with "\n"
            fprintf(fid,"  \"%s\"",text);
         }
         if (strEqu(type,"int"))
            fprintf(fid,"  \"%d\"",*intloc++);

         if (strEqu(type,"float"))
            fprintf(fid,"  \"%.2f\"",*floatloc++);
      }

      fprintf(fid,"\n");                                                   //  write EOR
   }
   
   fprintf(fid,"\n");
   fclose(fid);                                                            //  close file

   fid = fopen(recentfiles_file,"w");                                      //  recent files file
   if (! fid) return;
   
   for (int ii = 0; ii < Nrecentfiles; ii++)                               //  save list of recent files
      if (recentfiles[ii])
         fprintf(fid,"%s \n",recentfiles[ii]);
   
   fclose(fid);

   snprintf(buff,199,"%s/plugins",get_zuserdir());                         //  open file for plugins
   fid = fopen(buff,"w");
   if (! fid) return;
   
   for (int ii = 0; ii < Nplugins; ii++)                                   //  save plugins
      if (plugins[ii])
         fprintf(fid,"%s \n",plugins[ii]);
   
   fclose(fid);

   return;
}


//  load parameters from file /home/<user>/.fotoxx/parameters

void load_params()
{
   FILE        *fid;
   int         ii, err, cc, pcount, Idata;
   float       Fdata;
   char        buff[1000], text[1000], *pp;
   char        name[20], type[12], count[8], data[1000];
   void        *location;
   char        **charloc;
   int         *intloc;
   float       *floatloc;
   struct stat statb;
   
   snprintf(buff,499,"%s/parameters",get_zuserdir());                      //  open parameters file
   fid = fopen(buff,"r");
   if (! fid) return;                                                      //  none, defaults are used

   while (true)                                                            //  read parameters
   {
      pp = fgets_trim(buff,999,fid,1);
      if (! pp) break;
      if (*pp == '#') continue;                                            //  comment
      if (strlen(pp) < 40) continue;                                       //  rubbish

      err = 0;

      strncpy0(name,pp,20);                                                //  parm name
      strTrim2(name);
      
      strncpy0(type,pp+22,8);                                              //  parm type
      strTrim2(type);
      
      strncpy0(count,pp+32,4);                                             //  parm count
      strTrim2(count);
      err = convSI(count,pcount);
      
      strncpy0(data,pp+38,1000);                                           //  parm value(s)
      strTrim2(data);
      
      for (ii = 0; ii < Nparms; ii++)                                      //  match file record to param table
      {
         if (strNeq(name,paramTab[ii].name)) continue;                     //  parm name
         if (strNeq(type,paramTab[ii].type)) continue;                     //  parm type
         if (paramTab[ii].count != pcount) continue;                       //  parm count

         location = paramTab[ii].location;
         charloc = (char **) location;
         intloc = (int *) location;
         floatloc = (float *) location;
         
         for (int kk = 1; kk <= pcount; kk++)
         {
            pp = (char *) strField(data,' ',kk);
            if (! pp) break;

            if (strEqu(type,"char")) {
               repl_1str(pp,text,"\\n","\n");                              //  replace "\n" with real newlines
               *charloc++ = strdupz(text,0,"parameters");
            }

            if (strEqu(type,"int")) {
               err = convSI(pp,Idata);
               if (err) continue;
               *intloc++ = Idata;
            }

            if (strEqu(type,"float")) {
               err = convSF(pp,Fdata);
               if (err) continue;
               *floatloc++ = Fdata;
            }
         }
      }
   }

   fclose(fid);
   
   for (ii = 0; ii < Nparms; ii++)                                         //  set null strings to "undefined"
   {
      if (strNeq(paramTab[ii].type,"char")) continue;
      charloc = (char **) paramTab[ii].location;
      pcount = paramTab[ii].count;
      for (int jj = 0; jj < pcount; jj++)
         if (! charloc[jj]) charloc[jj] = strdupz("undefined",20);
   }
   
   if (strNeq(pversion,fotoxxvers))                                        //  fotoxx version change 
      printf("new Fotoxx version %s %s \n",pversion,fotoxxvers);

   Ntopdirk = 0;

   fid = fopen(topdirk_file,"r");                                          //  get top directories from
   if (fid)                                                                //    from top directories file
   {
      for (ii = 0; ii < 10; ii++)                                          //  multiple top directories
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         topdirk[ii] = strdupz(buff,0,"top_dirk");
         strTrim2(topdirk[ii]);                                            //  remove surrounding blanks
         cc = strlen(topdirk[ii]);                                         //  remove trailing '/' if present
         if (topdirk[ii][cc-1] == '/') topdirk[ii][cc-1] = 0;
         err = stat(topdirk[ii],&statb);
         if (err || ! S_ISDIR(statb.st_mode)) {
            printf("invalid top directory: %s \n",topdirk[ii]);            //  if invalid top directory,
            ii = ii - 1;                                                   //    remove from list
         }
      }

      fclose(fid);
      Ntopdirk = ii;
   }
   
   zoomratio = pow( 2.0, 1.0 / zoomcount);                                 //  set zoom ratio from zoom count

   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  recent image file list = empty
      recentfiles[ii] = 0;

   fid = fopen(recentfiles_file,"r");                                      //  open recent files file
   if (fid) {
      for (ii = 0; ii < Nrecentfiles; ii++)                                //  read list of recent files
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         if (*pp == '/') 
            recentfiles[ii] = strdupz(buff,0,"recentfile");
         else ii--; 
      }

      fclose(fid);
   }

   for (ii = 0; ii < maxplugins; ii++)                                     //  plugins list = empty
      plugins[ii] = 0;
   
   plugins[0] = strdupz("Gimp = gimp");                                    //  if empty, default Gimp plugin
   Nplugins = 1;

   snprintf(buff,499,"%s/plugins",get_zuserdir());                         //  open plugins file
   fid = fopen(buff,"r");
   if (fid)
   {   
      for (ii = 0; ii < maxplugins; ii++)                                  //  read list of plugins
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         plugins[ii] = strdupz(buff,0,"plugins");
      }

      fclose(fid);
      Nplugins = ii;
   }

   return;
}


/**************************************************************************/

//  Compute the mean brightness of all pixel neighborhoods,
//  using a Guassian or a flat distribution for the weightings. 
//  If a select area is active, only inside pixels are calculated.
//  The flat method is 10-100x faster.

int         brhood_radius;
float       brhood_kernel[200][200];                                       //  up to radius = 99
float       *brhood_brightness = 0;                                        //  neighborhood brightness per pixel
char        brhood_method;                                                 //  g = gaussian, f = flat distribution


void brhood_calc(int radius, char method)
{
   void * brhood_wthread(void *arg);

   int      rad, rad2, dx, dy, cc, ii;
   float    kern;
   
   brhood_radius = radius;
   brhood_method = method;

   if (brhood_method == 'g')                                               //  compute Gaussian kernel
   {                                                                       //  (not currently used)
      rad = brhood_radius;
      rad2 = rad * rad;

      for (dy = -rad; dy <= rad; dy++)
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx*dx + dy*dy <= rad2)                                        //  cells within radius
            kern = exp( - (dx*dx + dy*dy) / rad2);
         else kern = 0;                                                    //  outside radius
         brhood_kernel[dy+rad][dx+rad] = kern;
      }
   }

   if (brhood_brightness) zfree(brhood_brightness);                        //  allocate memory for pixel map
   cc = E1ww * E1hh * sizeof(float);
   brhood_brightness = (float *) zmalloc(cc,"brhood");
   memset(brhood_brightness,0,cc);

   if (sa_stat == 3) Fbusy_goal = sa_Npixel;                               //  set up progress tracking
   else Fbusy_goal = E1ww * E1hh;
   Fbusy_done = 0;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_wthread(brhood_wthread,&wtnx[ii]);
   wait_wthreads();                                                        //  wait for completion

   Fbusy_goal = 0;
   return;
}


//  worker thread function

void * brhood_wthread(void *arg)
{
   int      index = *((int *) arg);
   int      rad = brhood_radius;
   int      ii, px, py, qx, qy, Fstart;
   float    kern, bsum, bsamp, bmean;
   float    *pixel;

   if (brhood_method == 'g')                                               //  use round gaussian distribution
   {
      for (py = index; py < E1hh; py += Nwt)
      for (px = 0; px < E1ww; px++)
      {
         if (sa_stat == 3 && sa_mode != 7) {                               //  select area, not whole image
            ii = py * E1ww + px;                                           //    use only inside pixels
            if (! sa_pixmap[ii]) continue;
         }
         
         bsum = bsamp = 0;

         for (qy = py-rad; qy <= py+rad; qy++)                             //  computed weighted sum of brightness
         for (qx = px-rad; qx <= px+rad; qx++)                             //    for pixels in neighborhood
         {
            if (qy < 0 || qy > E1hh-1) continue;
            if (qx < 0 || qx > E1ww-1) continue;
            kern = brhood_kernel[qy+rad-py][qx+rad-px];
            pixel = PXMpix(E1pxm,qx,qy);
            bsum += pixbright(pixel) * kern;                               //  sum brightness * weight
            bsamp += kern;                                                 //  sum weights
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1ww + px;
         brhood_brightness[ii] = bmean;                                    //  pixel value

         Fbusy_done++;                                                     //  track progress
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }

   if (brhood_method == 'f')                                               //  use square flat distribution
   {
      Fstart = 1;
      bsum = bsamp = 0;

      for (py = index; py < E1hh; py += Nwt)
      for (px = 0; px < E1ww; px++)
      {
         if (sa_stat == 3 && sa_mode != 7) {                               //  select area, not whole image
            ii = py * E1ww + px;                                           //     compute only inside pixels
            if (! sa_pixmap[ii]) {
               Fstart = 1;
               continue;
            }
         }
         
         if (px == 0) Fstart = 1;
         
         if (Fstart) 
         {
            Fstart = 0;
            bsum = bsamp = 0;

            for (qy = py-rad; qy <= py+rad; qy++)                          //  add up all columns
            for (qx = px-rad; qx <= px+rad; qx++)
            {
               if (qy < 0 || qy > E1hh-1) continue;
               if (qx < 0 || qx > E1ww-1) continue;
               pixel = PXMpix(E1pxm,qx,qy);
               bsum += pixbright(pixel);
               bsamp += 1;
            }
         }
         else
         {
            qx = px-rad-1;                                                 //  subtract first-1 column
            if (qx >= 0) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1hh-1) continue;
                  pixel = PXMpix(E1pxm,qx,qy);
                  bsum -= pixbright(pixel);
                  bsamp -= 1;
               }
            }
            qx = px+rad;                                                   //  add last column
            if (qx < E1ww) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1hh-1) continue;
                  pixel = PXMpix(E1pxm,qx,qy);
                  bsum += pixbright(pixel);
                  bsamp += 1;
               }
            }
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1ww + px;
         brhood_brightness[ii] = bmean;

         Fbusy_done++;                                                     //  track progress
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }
         
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  get the neighborhood brightness for given pixel

float get_brhood(int px, int py)
{
   int ii = py * E1ww + px;
   return brhood_brightness[ii];
}


/**************************************************************************/

//  free all resources associated with the current image file

void free_resources(int fkeepundo)
{
   char        *pp;
   int         err;

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps
   
   if (! fkeepundo) {
      strcpy(command,"rm -f ");                                            //  delete all undo files
      strcat(command,undo_files);
      pp = strstr(command,"_undo_");                                       //  /home/user/.fotoxx/pppppp_undo_*
      strcpy(pp + 6,"*");
      err = system(command);
      if (err) printf("error: %s \n",wstrerror(err));
      Fmodified = Pundo = Pumax = Fsaved = 0;                              //  reset undo/redo stack
   }

   if (Fshutdown) {                                                        //  stop here if shutdown mode
      mutex_unlock(&Fpixmap_lock);
      return;
   }
   
   Ntoplines = Nptoplines;                                                 //  no image overlay lines
   freeMouse();                                                            //  free mouse
   
   if (curr_file) {
      sa_unselect();                                                       //  unselect select area
      zdialog_free(zdsela);                                                //  kill dialogs if active
      zfree(curr_file);                                                    //  free image file
      curr_file = 0;
      Fbusy_goal = 0;
      *SB_text = 0;
   }

   if (brhood_brightness) zfree(brhood_brightness);                        //  free brightness map
   brhood_brightness = 0;

   PXM_free(Fpxm);
   PXM_free(E1pxm);
   PXM_free(E3pxm);
   PXM_free(ERpxm);

   Fww = E1ww = E3ww = Dww = 0;                                            //  make unusable (crash)

   mutex_unlock(&Fpixmap_lock);
   return;
}


//  Get a virtual pixel at location (px,py) (real) in a PXM pixmap.
//  Get the overlapping real pixels and build a composite.

int vpixel(PXM *pxm, float px, float py, float *vpix)
{
   int            ww, hh, px0, py0;
   float          *bmp, *pix0, *pix1, *pix2, *pix3;
   float          f0, f1, f2, f3;
   float          red, green, blue;
   
   ww = pxm->ww;
   hh = pxm->hh;
   bmp = pxm->bmp;

   px0 = px;                                                               //  integer pixel containing (px,py)
   py0 = py;

   if (px0 < 1 || py0 < 1) return 0;                                       //  void edge pixels
   if (px0 > ww-2 || py0 > hh-2) return 0;
   
   pix0 = bmp + (py0 * ww + px0) * 3;                                      //  4 pixels based at (px0,py0)
   pix1 = pix0 + ww * 3;
   pix2 = pix0 + 3;
   pix3 = pix0 + ww * 3 + 3;

   f0 = (px0+1 - px) * (py0+1 - py);                                       //  overlap of (px,py)
   f1 = (px0+1 - px) * (py - py0);                                         //   in each of the 4 pixels
   f2 = (px - px0) * (py0+1 - py);
   f3 = (px - px0) * (py - py0);
   
   red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];      //  sum the weighted inputs
   green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
   blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
   
   vpix[0] = red;
   vpix[1] = green;
   vpix[2] = blue;
   
   if (blue < 0.5) return 0;                                               //  edge or voided area    v.12.03
   return 1;
}


//  compare two floats for significant difference
//  return:  0  difference not significant
//          +1  d1 > d2
//          -1  d1 < d2

int sigdiff(float d1, float d2, float signf)
{
   float diff = fabsf(d1-d2);
   if (diff == 0.0) return 0;
   diff = diff / (fabsf(d1) + fabsf(d2));
   if (diff < signf) return 0;
   if (d1 > d2) return 1;
   else return -1;
}


/**************************************************************************
   file read and write utilities
   PXM pixmap <--> file on disk
***************************************************************************/

//  Load an image file into a PXM pixmap
//  Also sets the following file scope variables:
//    f_load_type = "jpg" "tif" "png" or "other"
//    f_load_bpc = disk file bits per color = 1/8/16
//    f_load_size = disk file size
//  If fack is true, failure will cause a popup ACK dialog

PXM * f_load(cchar *filespec, int fack)
{
   int            err;
   cchar          *pext;
   PXM            *pxm1;
   struct stat    fstat;

   err = stat(filespec,&fstat);
   if (err) return 0;                                                      //  file not found
   if (! S_ISREG(fstat.st_mode)) return 0;                                 //  not a regular file
   if (image_file_type(filespec) != 2) return 0;                           //  not a supported image type

   if (strstr(filespec,"/.thumbnails/")) {                                 //  refuse thumbnail files
      zmessageACK(mWin,ZTX("cannot open thumbnail file"));
      return 0;
   }

   f_load_size = fstat.st_size;                                            //  disk file bytes
   
   pext = strrchr(filespec,'/');
   if (! pext) pext = filespec;
   pext = strrchr(pext,'.');
   if (! pext) pext = "";
   
   if (strstr(".jpg .jpeg .JPG .JPEG",pext)) strcpy(f_load_type,"jpg");
   else if (strstr(".tif .tiff .TIF .TIFF",pext)) strcpy(f_load_type,"tif");
   else if (strstr(".png .PNG",pext)) strcpy(f_load_type,"png");
   
   if (strEqu(f_load_type,"tif"))
      pxm1 = TIFFread(filespec);                                           //  use tiff lib to read tiff file
   else  pxm1 = PXBread(filespec);                                         //  use pixbuf lib for others
                                                                           //  both set f_load_bpc = file bpc
   if (! pxm1 && fack) 
      zmessageACK(mWin,"file error: %s",filespec);                         //  inform user      v.12.03

   return pxm1;
}


/**************************************************************************/

//  save current image to specified disk file (same or new).
//  set f_save_type, f_save_bpc, f_save_size
//  update search index file
//  return 0 if OK, else +N

int f_save(char *outfile, cchar *type, int bpc)                            //  use pixbuf or tiff library
{
   cchar          *exifkey[2] = { exif_orientation_key, iptc_editlog_key };
   cchar          *exifdata[2];
   char           **ppv, funcslist[1000];
   char           *pp, *tempfile, *pext;
   int            nkeys, err = 1, cc1, cc2;
   struct stat    fstat;
   
   Ffuncbusy++;                                                            //  can take many seconds on a slow PC

   pext = strrchr(outfile,'/');                                            //  force compatible file extension
   if (pext) pext = strrchr(pext,'.');                                     //    if not already
   if (! pext) pext = outfile + strlen(outfile);
   if (strEqu(type,"tif") && ! strstr(".tif .TIF .tiff .TIFF",pext))
      strcpy(pext,".tif");
   if (strEqu(type,"png") && ! strstr(".png .PNG",pext))
      strcpy(pext,".png");
   if (strEqu(type,"ico") && ! strstr(".ico .ICO",pext))
      strcpy(pext,".ico");
   if (strEqu(type,"bmp") && ! strstr(".bmp .BMP",pext))
      strcpy(pext,".bmp");
   if (strEqu(type,"jpg") && ! strstr(".jpg .JPG .jpeg .JPEG",pext))
      strcpy(pext,".jpg");

   tempfile = strdupz(get_zuserdir(),24,"f_save");                         //  use temp output file
   strcat(tempfile,"/temp_");                                              //    ~/.fotoxx/temp_ppppp.ext
   strcat(tempfile,PIDstring);
   strcat(tempfile,".");
   strcat(tempfile,type);

   if (strEqu(type,"tif"))                                                 //  save as tif file (bpc = 8 or 16)
      err = TIFFwrite(Fpxm,tempfile,bpc);

   else {                                                                  //  save as .png .ico .bmp .jpg 
      err = PXBwrite(Fpxm,tempfile);                                       //  (bpc = 8 always)
      bpc = 8;
   }

   if (err) {
      remove(tempfile);                                                    //  failure, clean up
      zfree(tempfile);
      Ffuncbusy--;
      return 1;
   }

   exifdata[0] = "";                                                       //  EXIF orientation = upright
   nkeys = 1;                                                              //  remove old width, height

   if (Pundo > 0)                                                          //  if edits were made,
   {                                                                       //    update relevant EXIF key
      *funcslist = 0;
      ppv = exif_get(curr_file,&exifkey[1],1);                             //  get existing list of edits
      if (ppv[0]) {
         strncpy0(funcslist,ppv[0],998);
         zfree(ppv[0]);
      }
      cc1 = strlen(funcslist);                                             //  add blank
      if (cc1 && funcslist[cc1-1] != ' ') {
         strcpy(funcslist+cc1," ");
         cc1++;
      }

      for (int ii = 1; ii <= Pundo; ii++)                                  //  append new list of edits
      {                                                                    //  (omit index 0 = initial)
         pp = pvlist_get(editlog,ii);
         cc2 = strlen(pp);
         if (cc1 + cc2 > 998) break;
         strcpy(funcslist+cc1,pp);
         strcpy(funcslist+cc1+cc2," ");
         cc1 += cc2 + 1;
      }
      
      exifdata[1] = funcslist;                                             //  IPTC edit history key, fotoxx edits done
      nkeys = 2;
   }

   err = stat(curr_file,&fstat);
   if (! err && S_ISREG(fstat.st_mode)) {                                  //  if current file exists,
      err = exif_copy(curr_file,tempfile,exifkey,exifdata,nkeys);          //    copy all EXIF/IPTC data to
      if (err) zmessageACK(mWin,"Unable to copy EXIF/IPTC");               //      temp file with above revisions
   }
   
   snprintf(command,ccc,"cp -f \"%s\" \"%s\" ",tempfile,outfile);          //  copy temp file to output file
   err = system(command);
   if (err) zmessageACK(mWin,"Unable to save image: %s",wstrerror(err));

   remove(tempfile);                                                       //  delete temp file
   zfree(tempfile);
   
   save_fileinfo(outfile);                                                 //  save tag changes if any
   if (! curr_file || strNeq(outfile,curr_file))                           //  if save to new file,
      update_search_index(outfile);                                        //    update search index file

   Fsaved = Pundo;                                                         //  update which mods are saved to disk

   add_recent_file(outfile);                                               //  first in recent files list
   
   strcpy(f_save_type,type);                                               //  update f_save_xxx data
   f_save_bpc = bpc;

   err = stat(outfile,&fstat);
   if (err) f_save_size = 0;
   else f_save_size = fstat.st_size;
   
   Ffuncbusy--;
   mwpaint2();
   return 0;
}


/**************************************************************************/

//  intercept TIFF warning messages (many)

void tiffwarninghandler(cchar *module, cchar *format, va_list ap)
{
   return;                                                                 //  stop flood of crap
   char  message[200];
   vsnprintf(message,199,format,ap);
   printf("TIFF warning: %s %s \n",module,message);
   return;
}


//  Read from TIFF file using TIFF library. 
//  Use native TIFF file bits/pixel >> f_load_bpc

PXM * TIFFread(cchar *filespec)
{
   static int  ftf = 1;
   TIFF        *tiff;
   PXM         *pxm;
   char        *tiffbuff;
   uint8       *tiff8;
   uint16      *tiff16;
   float       *pix;
   uint16      bpc, nch, fmt; 
   int         ww, hh, rps, stb, nst;                                      //  int not uint
   int         tiffstat, row, col, strip, cc;
   
   if (ftf) {
      TIFFSetWarningHandler(tiffwarninghandler);                           //  intercept TIFF warning messages
      ftf = 0;
   }

   tiff = TIFFOpen(filespec,"r");
   if (! tiff) {
      printf("TIFF open failure: %s \n",filespec);                         //  v.12.03
      return 0;
   }

   TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &ww);                            //  width
   TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &hh);                           //  height
   TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bpc);                        //  bits per color, 1/8/16
   TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &nch);                      //  no. channels (colors), 1/3/4
   TIFFGetField(tiff, TIFFTAG_ROWSPERSTRIP, &rps);                         //  rows per strip
   TIFFGetField(tiff, TIFFTAG_SAMPLEFORMAT, &fmt);                         //  data format
   stb = TIFFStripSize(tiff);                                              //  strip size
   nst = TIFFNumberOfStrips(tiff);                                         //  number of strips

// printf("ww %d  hh %d  nch %d  bpc %d  rps %d  stb %d  nst %d  fmt %d \n",
//          ww, hh, nch, bpc, rps, stb, nst, fmt);
   
   if (! (bpc <= 8 || bpc == 16)) {                                        //  check for supported bits/color
      printf("TIFF bits=%d not supported: %s \n",bpc,filespec);            //  v.12.03
      TIFFClose(tiff);
      return 0;
   }
   
   if (ww > 20000 || hh > 20000 || ww * hh > 100000000) {                  //  v.12.03
      printf("image too big, %dx%d: %s \n",ww,hh,filespec);
      TIFFClose(tiff);
      return 0;
   }
   
   f_load_bpc = bpc;                                                       //  file bpc 1/8/16

   if (bpc <= 8)                                                           //  use universal TIFF reader
   {                                                                       //    if bits/color <= 8
      tiffbuff = zmalloc(ww*hh*4,"tiffbuff");
      tiffstat = TIFFReadRGBAImage(tiff, ww, hh, (uint *) tiffbuff, 0);
      TIFFClose(tiff);

      if (tiffstat != 1) {
         printf("TIFF read failure: %s \n",filespec);                      //  v.12.03
         zfree(tiffbuff);
         return 0;
      }

      pxm = PXM_make(ww,hh);                                               //  create PXM

      tiff8 = (uint8 *) tiffbuff;

      for (row = 0; row < hh; row++)                                       //  convert RGBA to RGB
      {
         pix = pxm->bmp;
         pix += (hh-1 - row) * ww * 3;

         for (col = 0; col < ww; col++)
         {
            pix[0] = tiff8[0];                                             //  0 - 255  >>  0.0 - 255.0 
            pix[1] = tiff8[1];                                             //  v.12.03
            pix[2] = tiff8[2];
            if (pix[2] == 0) pix[2] = 1;                                   //  blue=0 reserved     v.12.03
            pix += 3;
            tiff8 += 4;
         }
      }

      zfree(tiffbuff);
      return pxm;
   }

   pxm = PXM_make(ww,hh);                                                  //  16 bits/color
   
   stb += 1000000;                                                         //  reduce risk of crash
   tiffbuff = zmalloc(stb,"tiffbuff");                                     //  read encoded strips 

   for (strip = 0; strip < nst; strip++)
   {
      cc = TIFFReadEncodedStrip(tiff,strip,tiffbuff,stb);
      if (cc < 0) {
         printf("TIFF read failure: %s \n",filespec);                      //  v.12.03
         TIFFClose(tiff);
         zfree(tiffbuff);
         PXM_free(pxm);
         return 0;
      }
      
      if (cc == 0) break;
      
      tiff16 = (uint16 *) tiffbuff;
      row = strip * rps;
      pix = pxm->bmp;
      pix += row * ww * 3;

      while (cc >= 6)
      {
         pix[0] = tiff16[0] / 256.0;                                       //  0 - 65535  >>  0.0 - 255.996
         pix[1] = tiff16[1] / 256.0;                                       //  v.12.03
         pix[2] = tiff16[2] / 256.0;
         if (pix[2] == 0) pix[2] = 1;                                      //  blue=0 reserved     v.12.03
         pix += 3;
         tiff16 += nch;
         cc -= nch * 2;
      }
   }

   TIFFClose(tiff);
   zfree(tiffbuff);
   return pxm;
}


//  Write to TIFF file using TIFF library. 
//  File bpc is 8 or 16.
//  returns 0 if OK, +N if error.

int TIFFwrite(PXM *pxm, cchar *filespec, int bpc)
{
   static int  ftf = 1;
   TIFF        *tiff;
   uint8       *tiff8;
   uint16      *tiff16;
   float       *pix;
   int         tiffstat = 0;
   int         ww, hh, row, col, rowcc;                                    //  int not uint
   int         nch, pm = 2, pc = 1, comp = 5; 
   char        *tiffbuff;
   
   if (ftf) {
      TIFFSetWarningHandler(tiffwarninghandler);                           //  intercept TIFF warning messages
      ftf = 0;
   }

   tiff = TIFFOpen(filespec,"w");
   if (! tiff) {
      zmessageACK(mWin,"TIFF open failure: %s",filespec);                  //  v.12.03
      return 1;
   }
   
   ww = pxm->ww;
   hh = pxm->hh;
   nch = 3;                                                                //  alpha channel removed

   TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, ww);
   TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, bpc);
   TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, nch);
   TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, pm);                            //  RGB
   TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, pc);
   TIFFSetField(tiff, TIFFTAG_COMPRESSION, comp);                          //  LZW
   
   rowcc = TIFFScanlineSize(tiff);
   tiffbuff = (char*) zmalloc(rowcc,"tiffbuff");

   for (row = 0; row < hh; row++)
   {
      if (bpc == 8)
      {
         tiff8 = (uint8 *) tiffbuff;
         pix = pxm->bmp + row * ww * 3;

         for (col = 0; col < ww; col++)
         {
            tiff8[0] = pix[0];                                             //  0.0 - 255.996  >>  0 - 255
            tiff8[1] = pix[1];                                             //  v.12.03
            tiff8[2] = pix[2];
            pix += 3;
            tiff8 += nch;
         }
      }

      if (bpc == 16) 
      {
         tiff16 = (uint16 *) tiffbuff;
         pix = pxm->bmp + row * ww * 3;

         for (col = 0; col < ww; col++)
         {
            tiff16[0] = 256.0 * pix[0];                                    //  0.0 - 255.996  >>  0.0 >> 65535
            tiff16[1] = 256.0 * pix[1];                                    //  v.12.03
            tiff16[2] = 256.0 * pix[2];
            pix += 3;
            tiff16 += nch;
         }
      }

      tiffstat = TIFFWriteScanline(tiff,tiffbuff,row,0);
      if (tiffstat != 1) break;
   }

   TIFFClose(tiff);
   zfree(tiffbuff);
   
   if (tiffstat == 1) return 0;
   zmessageACK(mWin,ZTX("TIFF write failure"));
   return 2;
}


/**************************************************************************/

//  Read from jpeg/png file using pixbuf library. bpc = 8.

PXM * PXBread(cchar *filespec)
{
   GError      *gerror = 0;
   PXB         *pxb;
   PXM         *pxm;
   int         ww, hh, px, py, nch, rowst;
   uint8       *bmp1, *pix1;
   float       *bmp2, *pix2;
   
   pxb = gdk_pixbuf_new_from_file(filespec,&gerror);
   if (! pxb) {
      if (gerror) printf("%s \n",gerror->message);
      printf("file type not supported: %s \n",filespec);                   //  v.12.03
      return 0;
   }

   ww = gdk_pixbuf_get_width(pxb);
   hh = gdk_pixbuf_get_height(pxb);
   nch = gdk_pixbuf_get_n_channels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);
   bmp1 = gdk_pixbuf_get_pixels(pxb);
   
   if (ww > 20000 || hh > 20000 || ww * hh > 100000000) {                  //  v.12.03
      printf("image too big, %dx%d: %s \n",ww,hh,filespec);
      g_object_unref(pxb);
      return 0;
   }

   pxm = PXM_make(ww,hh);
   bmp2 = pxm->bmp;

   for (py = 0; py < hh; py++)
   {
      pix1 = bmp1 + rowst * py;
      pix2 = bmp2 + py * ww * 3;

      for (px = 0; px < ww; px++)
      {
         pix2[0] = pix1[0];                                                //  0 - 255  >>  0.0 - 255.0
         pix2[1] = pix1[1];                                                //  v.12.03
         pix2[2] = pix1[2];
         if (pix2[2] == 0) pix2[2] = 1;                                    //  blue=0 reserved     v.12.03
         pix1 += nch;
         pix2 += 3;
      }
   }

   f_load_bpc = 8;                                                         //  file bits per color
   g_object_unref(pxb);

   return pxm;
}


//  Write to jpeg/png file using pixbuf library. bpc = 8.
//  returns 0 if OK, +N if error.

int PXBwrite(PXM *pxm, cchar *filespec)
{
   int         ww, hh, px, py, rowst;
   float       *bmp1, *pix1;
   uint8       *bmp2, *pix2;
   PXB         *pxb;
   cchar       *pext;
   GError      *gerror = 0;
   int         pxbstat;

   ww = pxm->ww;
   hh = pxm->hh;
   
   pxb = gdk_pixbuf_new(GDKRGB,0,8,ww,hh);
   if (! pxb) zappcrash("pixbuf allocation failure");

   bmp1 = pxm->bmp;
   bmp2 = gdk_pixbuf_get_pixels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);

   for (py = 0; py < hh; py++)
   {
      pix1 = bmp1 + py * ww * 3;
      pix2 = bmp2 + rowst * py;

      for (px = 0; px < ww; px++)
      {
         pix2[0] = pix1[0];                                                //  0.0 - 255.996  >>  0 - 255
         pix2[1] = pix1[1];                                                //  v.12.03
         pix2[2] = pix1[2];
         pix1 += 3;
         pix2 += 3;
      }
   }
   
   pext = strrchr(filespec,'/');
   if (! pext) pext = filespec;
   pext = strrchr(pext,'.');
   if (! pext) pext = "";
   
   if (strstr(".png .PNG",pext))
      pxbstat = gdk_pixbuf_save(pxb,filespec,"png",&gerror,null);

   else if (strstr(".ico .ICO",pext))                                      //  v.12.03
      pxbstat = gdk_pixbuf_save(pxb,filespec,"ico",&gerror,null);

   else if (strstr(".bmp .BMP",pext))                                      //  v.12.03
      pxbstat = gdk_pixbuf_save(pxb,filespec,"bmp",&gerror,null);

   else pxbstat = gdk_pixbuf_save(pxb,filespec,"jpeg",&gerror,"quality",jpeg_quality,null);

   g_object_unref(pxb);
   if (pxbstat) return 0;

   printf("%s \n",gerror->message);
   zmessageACK(mWin,"pixbuf write failure");
   return 1;
}


/**************************************************************************
      PXM pixmap conversion and rescale functions
      v.12.03  PXM pixmaps have float pixel values instead of uint16
***************************************************************************/

//  initialize PXM pixmap - allocate memory

PXM * PXM_make(int ww, int hh)
{
   //  impose reasonableness limits: ww + hh < 20K and ww * hh < 50M
   
   if (ww < 5 || hh < 5 || ww + hh > 20000 || ww * hh > 50000000)
      zappcrash("PXM_make() size: %d %d",ww,hh);
   
   PXM *pxm = (PXM *) zmalloc(sizeof(PXM),"PXM");
   pxm->ww = ww;
   pxm->hh = hh;
   pxm->bmp = (float *) zmalloc(ww*hh*12,"PXM.bmp");                       //  float[3] per pixel
   strcpy(pxm->wmi,"pxmpix");
   return pxm;
}


//  free PXM pixmap

void PXM_free(PXM *&pxm)
{
   if (! pxm) return;
   if (! strEqu(pxm->wmi,"pxmpix")) 
      zappcrash("PXM_free(), bad PXM");
   strcpy(pxm->wmi,"xxxxxx");
   zfree(pxm->bmp);
   zfree(pxm);
   pxm = 0;
   return;
}


//  create a copy of an PXM pixmap

PXM * PXM_copy(PXM *pxm1)
{
   int      cc;
   PXM      *pxm2;

   pxm2 = PXM_make(pxm1->ww, pxm1->hh);
   cc = pxm1->ww * pxm1->hh * 3 * sizeof(float);
   memcpy(pxm2->bmp,pxm1->bmp,cc);
   return pxm2;
}


//  create a copy of a PXM area

PXM * PXM_copy_area(PXM *pxm1, int orgx, int orgy, int ww2, int hh2)
{
   float    *bmp1, *bmp2, *pix1, *pix2;
   PXM      *pxm2 = 0;
   int      ww1, px1, py1, px2, py2;

   ww1 = pxm1->ww;

   pxm2 = PXM_make(ww2,hh2);
   bmp1 = pxm1->bmp;
   bmp2 = pxm2->bmp;
  
   for (py1 = orgy, py2 = 0; py2 < hh2; py1++, py2++) 
   {
      for (px1 = orgx, px2 = 0; px2 < ww2; px1++, px2++)
      {
         pix1 = bmp1 + (py1 * ww1 + px1) * 3;
         pix2 = bmp2 + (py2 * ww2 + px2) * 3;

         pix2[0] = pix1[0];
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
         pix1 += 3;
         pix2 += 3;
      }
   }

   return pxm2;
}


//  replace blue = 0 pixels with blue = 1
//  (blue = 0 reserved for pixels voided by warp or overlay offsets)
//  needs about 0.013 seconds for 10 megapixel image and 3.3 GHz CPU

void PXM_fixblue(PXM *pxm)
{
   int      size;
   float    *pixel, *pixel0, *pixelN;

   size = pxm->ww * pxm->hh * 3;
   pixel0 = pxm->bmp + 2;
   pixelN = pixel0 + size;

   for (pixel = pixel0; pixel < pixelN; pixel += 3)                        //  +3 is +12 bytes
      if (! *pixel) *pixel = 1.0;

   return;
}


/**************************************************************************

   Rescale PXM image to new width and height.
   The scale ratios may be different for width and height.

   Method: 
   The input and output images are overlayed, stretching or shrinking the
   output pixels as needed. The contribution of each input pixel overlapping
   an output pixel is proportional to the area of the output pixel covered by
   the input pixel. The contributions of all overlaping input pixels are added.
   The work is spread among Nwt threads to reduce time on SMP processors.

   Example: if the output image is 40% of the input image, then:
     outpix[0,0] = 0.16 * inpix[0,0] + 0.16 * inpix[1,0] + 0.08 * inpix[2,0]
                 + 0.16 * inpix[0,1] + 0.16 * inpix[1,1] + 0.08 * inpix[2,1]
                 + 0.08 * inpix[0,2] + 0.08 * inpix[1,2] + 0.04 * inpix[2,2]

***************************************************************************/

namespace pxmrescale {
   float    *bmp1;
   float    *bmp2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *px1L;
   int      *py1L;
   float    *pxmap;
   float    *pymap;
   int      maxmapx;
   int      maxmapy;
   int      busy[max_threads];
}

PXM * PXM_rescale(PXM *pxm1, int ww, int hh)
{
   using namespace pxmrescale;

   void * pxm_rescale_thread(void *arg);

   PXM         *pxm2;
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;

   ww1 = pxm1->ww;                                                         //  input PXM
   hh1 = pxm1->hh;
   bmp1 = pxm1->bmp;
   
   pxm2 = PXM_make(ww,hh);                                                 //  output PXM
   ww2 = ww;
   hh2 = hh;
   bmp2 = pxm2->bmp;
   
   memset(bmp2, 0, ww2 * hh2 * 3 * sizeof(float));                         //  clear output pixmap

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = scalex + 2;                                              //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = scaley + 2;
   maxmapy += 1;                                                           //  (extra entry for -1 flag)
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));               //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));               //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int));                              //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int));                              //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = py1a;
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = px1a;
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

   for (ii = 0; ii < Nwt; ii++) {                                          //  start working threads
      busy[ii] = 1;
      start_detached_thread(pxm_rescale_thread,&wtnx[ii]);
   }
   
   for (ii = 0; ii < Nwt; ii++)                                            //  wait for all done
      while (busy[ii]) zsleep(0.004);

   zfree(px1L);
   zfree(py1L);
   zfree(pxmap);
   zfree(pymap);

   return pxm2;
}

void * pxm_rescale_thread(void *arg)                                       //  worker thread function
{
   using namespace pxmrescale;

   int         index = *((int *) arg);
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bmp1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bmp2 + (py2 * ww2 + px2) * 3;                         //  save output pixel
            pixel2[0] = red;
            pixel2[1] = green;
            pixel2[2] = blue;
         }
      }
   }

   busy[index] = 0;
   return 0;
}


//  Rescale PXM pixel map (float) to RGB pixel map (uint8).                //  new v.12.03
//  Retain mapping of input to output pixels for use by PXM_RGB_update().

namespace pxmrgbrescale {
   float    *bmp1;
   RGB      *bmp2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *px1L;
   int      *py1L;
   float    scalex, scaley;
   float    *pxmap;
   float    *pymap;
   int      maxmapx;
   int      maxmapy;
   int      busy[max_threads];
}

RGB * PXM_RGB_rescale(PXM *pxm1, int ww, int hh)
{
   using namespace pxmrgbrescale;

   void * pxm_rgb_rescale_thread(void *arg);

   static int  ffree = 0;
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;

   if (ffree) {                                                            //  free prior maps
      zfree(px1L);
      zfree(py1L);
      zfree(pxmap);
      zfree(pymap);
   }

   ffree = 1;

   ww1 = pxm1->ww;                                                         //  input PXM
   hh1 = pxm1->hh;
   bmp1 = pxm1->bmp;
   
   bmp2 = (RGB *) zmalloc(ww*hh*3,"pxm.rgb");                              //  output RGB
   ww2 = ww;
   hh2 = hh;
   
   memset(bmp2, 0, ww2 * hh2);                                             //  clear output

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = scalex + 2;                                              //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = scaley + 2;
   maxmapy += 1;                                                           //  (extra entry for -1 flag)
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));               //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));               //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int));                              //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int));                              //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = py1a;
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = px1a;
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

   for (ii = 0; ii < Nwt; ii++) {                                          //  start working threads
      busy[ii] = 1;
      start_detached_thread(pxm_rgb_rescale_thread,&wtnx[ii]);
   }
   
   for (ii = 0; ii < Nwt; ii++)                                            //  wait for all done
      while (busy[ii]) zsleep(0.004);

   return bmp2;
}

void * pxm_rgb_rescale_thread(void *arg)                                   //  worker thread function
{
   using namespace pxmrgbrescale;

   int         index = *((int *) arg);
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       *pixel1;
   RGB         *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bmp1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bmp2 + (py2 * ww2 + px2) * 3;                         //  save output pixel
            pixel2[0] = red;                                               //  0.0 - 255.996  >>  0 - 255
            pixel2[1] = green;
            pixel2[2] = blue;
         }
      }
   }

   busy[index] = 0;
   return 0;
}


//  Copy a modified PXM section (e.g. from an image edit function) into the
//  corresponding RGB section made with a prior call to PXM_RGB_rescale.
//  orgx/y and ww/hh reference the PXM source image.

void  PXM_RGB_update(int orgx, int orgy, int ww, int hh)                   //  new v.12.03
{
   using namespace pxmrgbrescale;

   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       *pixel1;
   RGB         *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;
   
   orgx = orgx / scalex;                                                   //  1st output pixel
   orgy = orgy / scaley;

   ww = orgx + ww / scalex + 2;                                            //  last output pixel
   hh = orgy + hh / scaley + 2;

   if (orgx < 0) orgx = 0;
   if (orgy < 0) orgy = 0;

   if (ww > ww2) ww = ww2;
   if (hh > hh2) hh = hh2;

   for (py2 = orgy; py2 < hh; py2++)                                       //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = orgx; px2 < ww; px2++)                                    //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bmp1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bmp2 + (py2 * ww2 + px2) * 3;                         //  save output pixel
            pixel2[0] = red;                                               //  0.0 - 255.996  >>  0 - 255
            pixel2[1] = green;
            pixel2[2] = blue;
         }
      }
   }
   
   return;
}


/**************************************************************************

      PXM *pxm2 = PXM_rotate(PXM *pxm1, float angle)

      Rotate PXM pixmap through an arbitrary angle (degrees).

      The returned image has the same size as the original, but the
      pixmap size is increased to accomodate the rotated image.
      (e.g. a 100x100 image rotated 45 deg. needs a 142x142 pixmap).

      The space added around the rotated image is black (RGB 0,0,0).
      Angle is in degrees. Positive direction is clockwise.
      Speed is about xx million pixels/sec/thread for a 2.4 GHz CPU.       ////  add benchmark
      Loss of resolution is less than 1 pixel.
      
      Work is divided among Nwt threads to gain speed.
      
***************************************************************************/

namespace pxmrotate {
   int      busy = 0;
   float    *bmp1;
   float    *bmp2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   float    angle;
}

PXM * PXM_rotate(PXM *pxm1, float anglex)
{
   using namespace pxmrotate;

   void     *PXM_rotate_thread(void *);

   int      ii;
   PXM      *pxm2;
   
   ww1 = pxm1->ww;                                                         //  input PXM
   hh1 = pxm1->hh;
   bmp1 = pxm1->bmp;
   angle = anglex;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabsf(angle) < 0.001) {                                             //  angle = 0 within my precision
      pxm2 = PXM_copy(pxm1);                                               //  return a copy of the input PXM
      return pxm2;
   }

   ww2 = ww1*fabsf(cosf(angle)) + hh1*fabsf(sinf(angle));                  //  rectangle containing rotated image
   hh2 = ww1*fabsf(sinf(angle)) + hh1*fabsf(cosf(angle));
   
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM
   bmp2 = pxm2->bmp;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_detached_thread(PXM_rotate_thread,&wtnx[ii]);
   zadd_locked(busy,+Nwt);

   while (busy) zsleep(0.004);                                             //  wait for completion
   return pxm2;
}

void * PXM_rotate_thread(void *arg)
{
   using namespace pxmrotate;

   int      index = *((int *) (arg));
   int      px2, py2, px0, py0;
   float    *pix0, *pix1, *pix2, *pix3;
   float    px1, py1;
   float    f0, f1, f2, f3, red, green, blue;
   float    a, b, d, e, ww15, hh15, ww25, hh25;

   ww15 = 0.5 * ww1;
   hh15 = 0.5 * hh1;
   ww25 = 0.5 * ww2;
   hh25 = 0.5 * hh2;

   a = cosf(angle);
   b = sinf(angle);
   d = - sinf(angle);
   e = cosf(angle);
   
   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)
   {
      px1 = a * (px2 - ww25) + b * (py2 - hh25) + ww15;                    //  (px1,py1) = corresponding
      py1 = d * (px2 - ww25) + e * (py2 - hh25) + hh15;                    //    point within input pixels

      px0 = px1;                                                           //  pixel containing (px1,py1)
      py0 = py1;
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = bmp2 + (py2 * ww2 + px2) * 3;                              //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = bmp1 + (py0 * ww1 + px0) * 3;                                 //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = bmp2 + (py2 * ww2 + px2) * 3;                                 //  output pixel
      pix2[0] = red;
      pix2[1] = green;
      pix2[2] = blue;
   }
   
   zadd_locked(busy,-1);
   return 0;
}


